/***************************************************************************
                       dfg1394_driver.c  -  description
                             -------------------
 Copyright (c) 2003 Norbert Wagner <nw@softwarekombinat.de>



    created              : Thu Jun 25 2003 Norbert Wagner <nw@softwarekombinat.de>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include <linux/init.h>
#include <linux/module.h>

#include "dfg1394.h"

#include "ieee1394_core.h"
#include "ieee1394_hotplug.h"
#include "nodemgr.h"



/**
 * @brief the one and only driver object
 */
static dfg1394_driver_t dfg1394_driver;


/* ************************************************************************ */
/** 
 *   @name exported sysfs attributes                                
 */
/* ************************************************************************ */
/* @{                                                                       */

/** @brief shows the @c frames attribute */
ssize_t dfg1394_driver__dev_attr_frames_show(
	struct device* device, char* buf)
{
	dfg1394_device_t* dev = (dfg1394_device_t*)device->driver_data;
	return sprintf(
		buf, "%u %u %u\n", 
		dfg1394_node__get_frames_captured(&dev->node),
		dfg1394_node__get_frames_dropped(&dev->node),
		dfg1394_node__get_frames_lost(&dev->node)
	);
}

/** @brief declares the @frames attribute */
DEVICE_ATTR(frames, 644,
	    dfg1394_driver__dev_attr_frames_show, NULL);

/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_driver: PnP Handling                                  
 */
/* ************************************************************************ */
/* @{                                                                       */

/**
 * @brief Callback: called when if a new device is attached 
 */
int dfg1394_driver__add_device( struct device* device  )
{
	dfg1394_device_t* dev = NULL;
	struct list_head* entry;
	int ret;
	u32 minor = 0;

	PRINT_INFO("DRIVER:PROBE\n");


	/* 
	   try to find a ghost device we can resurrect 
	*/
	list_for_each(entry, &dfg1394_driver.devices) {
		dev = list_entry(entry, dfg1394_device_t, list_entry);
		if (dfg1394_device__is_ghost(dev)) {
			ret = dfg1394_device__resurrect(dev,device);
			if (IS_SUCCESS(ret))
				return SUCCESS;
		}
	}

	/* 
	   => no ghost device found, create a new one
	 */

	// create devices
	dev = dfg1394_device__create( device );
	if ( dev == NULL )
		return -ENOMEM;
	
	/*
	  We find the minor number by simply using the 
	  number of dfg1394 devices. This is stupid, but works
	  well for one device ;-)
	  @todo We should make the minors configurable!
	*/
	minor = dfg1394_driver.device_count++;
	ret = dfg1394_device__register( dev, minor );
	
	if ( IS_FAILURE( ret )) {
		dfg1394_device__destroy( dev );
		dfg1394_driver.device_count--;
	} else {
		list_add_tail( 
			&dev->list_entry, &dfg1394_driver.devices);
	}

	/* create device attributes */
	device_create_file( device, &dev_attr_frames);

	return ret;
}

/**
 * @brief Callback:  Called by the hotplug subsystem when a device is removed
 */
int dfg1394_driver__remove_device( struct device* device )
{
	dfg1394_device_t* dev = (dfg1394_device_t*) device->driver_data;
	int ret;

	/* remove device attributes */
	device_remove_file( device, &dev_attr_frames);

	ret = dfg1394_device__unregister( dev );

	if(IS_SUCCESS(ret)) {
		list_del( &dev->list_entry );
		dfg1394_driver.device_count--;
		dfg1394_device__destroy( dev );
	} else
		dfg1394_device__ghostify( dev );

	return SUCCESS;
}


/**
 * @brief exported device IDs
 */
static struct ieee1394_device_id dfg1394_id_table[] = {
        {
                .match_flags    =  IEEE1394_MATCH_SPECIFIER_ID
         	                 | IEEE1394_MATCH_VERSION,
		.specifier_id   = 0x000748,
                .version        = 0x526f6e,
        },
        {} // end of array
};

/** @brief export the module's ID table */
MODULE_DEVICE_TABLE(ieee1394, dfg1394_id_table);

/** @brief the device name */
#define DFG1394_NAME "dfg1394-1"

const char* dfg1394_name = DFG1394_NAME;

/**
 * @brief the IEEE1394 specific driver object
 */
static struct hpsb_protocol_driver dfg1394_protocol_driver = {
        .name           = "DFG/1394-1 Video Converter Driver",
        .id_table       = dfg1394_id_table,
        .driver         = {
                .name   = DFG1394_NAME,
                .bus    = NULL, /* to be set in driver__create() */
		.probe  = dfg1394_driver__add_device,
		.remove = dfg1394_driver__remove_device,
        },
};


/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_driver: Management                                    
 */
/* ************************************************************************ */
/* @{                                                                       */


/**
 * @brief Creates the driver singleton
 */
int dfg1394_driver__create( void )
{
	/* register protocol driver */
	dfg1394_driver.driver = &dfg1394_protocol_driver;//
	dfg1394_driver.driver->driver.bus = &ieee1394_bus_type;
	dfg1394_driver.device_count = 0;
	INIT_LIST_HEAD(&dfg1394_driver.devices);

	hpsb_register_protocol( dfg1394_driver.driver );
	
	return SUCCESS;
}

/**
 * @brief Destroys the driver singleton
 */
int dfg1394_driver__destroy( void )
{
	dfg1394_device_t* dev;

	while ( !list_empty(&dfg1394_driver.devices) ) {
		dev = list_entry( 
			dfg1394_driver.devices.next,
			dfg1394_device_t, list_entry);
		list_del_init(&dev->list_entry);
		
		if (dfg1394_device__is_ghost(dev)) {
			PRINT_INFO("destroying ghost device.\n");
			dfg1394_device__unregister(dev);
			dfg1394_device__destroy(dev);
		}
	}
		
	hpsb_unregister_protocol( dfg1394_driver.driver );

	return SUCCESS;
}

/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_driver: Module handling                               
 */
/* ************************************************************************ */
/* @{                                                                       */

/** @brief export the module's author */
MODULE_AUTHOR( "Norbert Wagner" );

/** @brief export the module's description */
MODULE_DESCRIPTION("Driver for the DFG-1394-1 framegrabber");

/** @brief export the module's supported device name*/
MODULE_SUPPORTED_DEVICE("DFG-1394-1");

/** @brief export the module's license */
MODULE_LICENSE("GPL");


/**
 * @brief module initialization
 */
int dfg1394_init( void )
{
	int ret = SUCCESS; /* return code */
        
	/* initialize global driver data */
	ret = dfg1394_driver__create();
	if ( IS_FAILURE( ret )) goto error;

	/* initialize firewire layer */
	ret = dfg1394_fw_init();
	if( IS_FAILURE( ret )) goto error;

	PRINT_NOTICE( "loaded.\n" );
	return ret;

 error:
	PRINT_WARNING( "loading failed.\n" );
	return ret;
}

/**
 * @brief module cleanup
 */
void dfg1394_exit( void )
{
	/* shutdown firewire layer */
	dfg1394_fw_exit();

	/* shutdown global driver data */
	dfg1394_driver__destroy();

	PRINT_NOTICE( "unloaded.\n" );
}

/** @brief declare the modules init function */            
module_init( dfg1394_init );

/** @brief declare the modules exit function */            
module_exit( dfg1394_exit );

/*@}*/
