/***************************************************************************
                     dfg1394_firewire.c  -  description
                             -------------------
 Copyright (c) 2003 Norbert Wagner <nw@softwarekombinat.de>



    created              : 2003-03-06 by Norbert Wagner <nw@softwarekombinat.de>
    modified             : 2003-10-07 Arne Caspari <arne_caspari@eu.theimagingsource.com>
                           - changed unit identification to look at spec_id and sw_version instead of uid
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "dfg1394.h"

#include "linux/sched.h"
#include "linux/list.h"
#include <linux/mm.h>

/* ieee1394 includes */
#include "ieee1394_transactions.h"
#include "ieee1394_core.h"
#include "highlevel.h"
#include "hosts.h"
#include "iso.h"



/* --------------------------------------------------------------------------
   --- external callbacks ---------------------------------------------------
   --------------------------------------------------------------------------*/

extern void dfg1394_fw_on_iso_receive( struct hpsb_iso* iso );

/* --------------------------------------------------------------------------
   --- private declarations -------------------------------------------------
   --------------------------------------------------------------------------*/

/**
 * @brief Global data used by the firewire layer
 */
struct dfg1394_firewire
{

	/**
	   @brief Used for fast channel->node lookup
	 */
        dfg1394_node_t* channel_allocation[ 64 ];

	/**
 	* @brief used for fast nodeid->node lookup
	*
 	* @todo This has to be on a per host basis
	*       if we want to support more than
	*       one host.
 	*/
	dfg1394_node_t* node_allocation[ 64 ];
	
        /** @brief handle to the ieee1394 highlevel interface */
	struct hpsb_highlevel* hl_handle;

	/** @brief The IEEE 1394 Host */
	struct hpsb_host* host;

};
typedef struct dfg1394_firewire dfg1394_firewire_t;

/** @brief global firwire data */
static dfg1394_firewire_t dfg1394_firewire;


/**
 * @brief lowlevel IEEE1394 Write
 *
 * @warning This function internally waits on a semaphore 
 * and may therefore not be used while handling interrupts !
 *
 * @return @c SUCCESS or a negative error code 
 *
 */
int dfg1394_fw_write(
	struct hpsb_host *host, u32 generation, nodeid_t nodeid, u64 addr, 
	quadlet_t *buffer, size_t length )
{
	return hpsb_write( host, nodeid, generation, addr, buffer, length );
}

/**
 * @brief lowlevel IEEE1394 Read
 *
 * @warning This function internally waits on a semaphore 
 * and may therefore not be used while handling interrupts !
 *
 * @return @c SUCCESS or a negative error code 
 *
 */
int dfg1394_fw_read(
	struct hpsb_host *host, u32 generation, nodeid_t nodeid, u64 addr, 
	quadlet_t *buffer, size_t length )
{
	return hpsb_read( host, nodeid, generation, addr, buffer, length );
}

/**
 * @brief Reads the node's unique identifier (vendor_id + chip_ids) from the
 * CSR Bus Info Block.
 *
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_read_node_uid( 
	struct hpsb_host *host, 
	int node_idx, 
	u32 generation, 
	dfg1394_uid_t* uid )
{
	int ret = SUCCESS;

	if ( IS_FAILURE( ret = dfg1394_fw_read(
			host, generation, LOCAL_BUS | node_idx, 
			CSR_REGISTER_BASE + CSR_CONFIG_ROM + 0x0C, 
			(quadlet_t*)uid, 4 )))
		return ret;

	if ( IS_FAILURE( ret = dfg1394_fw_read(
			host, generation, LOCAL_BUS | node_idx, 
			CSR_REGISTER_BASE + CSR_CONFIG_ROM + 0x10, 
			((quadlet_t*)uid)+1, 4 )))
		return ret;

	*uid = be64_to_cpu( *uid );

	return ret;
}

/**
 * @brief Reads the node's spec id
 * CSR Bus Info Block.
 *
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_read_node_spec_id( 
	struct hpsb_host *host, 
	int node_idx, 
	u32 generation, 
	u32* spec_id )
{
	int ret = SUCCESS;
	quadlet_t quad;

	if ( IS_FAILURE( ret = dfg1394_fw_read(
			host, generation, LOCAL_BUS | node_idx, 
			CSR_REGISTER_BASE + CSR_CONFIG_ROM + 0x24, 
			&quad, 4 )))
	{
		return ret;
	}

	quad = be32_to_cpu( quad );

	if( quad != 0xd1000001 )
	{
		*spec_id = 0;
		return -1;
	}

	if ( IS_FAILURE( ret = dfg1394_fw_read(
			host, generation, LOCAL_BUS | node_idx, 
			CSR_REGISTER_BASE + CSR_CONFIG_ROM + 0x2c, 
			&quad, 4 )))
		return ret;


	quad = be32_to_cpu( quad );

	if( (quad >> 24) == 0x12 )
	{
		*spec_id = quad & 0xffffff;
	}
	else
	{
		*spec_id = 0;
		return -1;
	}

	return ret;
}

/**
 * @brief Reads the node's sw version
 * CSR Bus Info Block.
 *
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_read_node_sw_version( 
	struct hpsb_host *host, 
	int node_idx, 
	u32 generation, 
	u32* sw_version )
{
	int ret = SUCCESS;
	quadlet_t quad;

	if ( IS_FAILURE( ret = dfg1394_fw_read(
			host, generation, LOCAL_BUS | node_idx, 
			CSR_REGISTER_BASE + CSR_CONFIG_ROM + 0x24, 
			&quad, 4 )))
	{
		return ret;
	}
	
	quad = be32_to_cpu( quad );

	if( quad != 0xd1000001 )
	{
		*sw_version = 0;
		return -1;
	}

	if ( IS_FAILURE( ret = dfg1394_fw_read(
			host, generation, LOCAL_BUS | node_idx, 
			CSR_REGISTER_BASE + CSR_CONFIG_ROM + 0x30, 
			&quad, 4 )))
		return ret;

	quad = be32_to_cpu( quad );

	if( (quad >> 24) == 0x13 )
	{
		*sw_version = quad & 0xffffff;
	}
	else
	{
		*sw_version = 0;
		return -1;
	}

	return ret;
}




/**
 * @brief Finds a free ISO Channel 
 *
 * @todo really search here, don't just return @c 0
 *
 * @return the iso channel
 */
int dfg1394_fw_find_free_iso_channel( struct hpsb_host* host, u32 generation )
{
	return 0;
}

/**
 * @brief Starts iso reception on the node's iso_channel
 *
 * @param node node for which to receive packets
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_start_iso_receive( dfg1394_node_t* node  )
{
	u8 channel = node->iso_channel;

	/* sanity checks */
	if ( channel >= 64 ) {
		PRINT_ERR("channel invalid: %d\n", (u32) channel );
		return -EINVAL;
	}

	if ( dfg1394_firewire.channel_allocation[channel] != NULL ) {
		PRINT_ERR("channel %d already allocated.\n", channel );
		return -EBUSY;
	}
	
	node->iso = hpsb_iso_recv_init( 
		node->host,
		4096*200, 
		200,
		channel,
		80,
		HPSB_ISO_DMA_PACKET_PER_BUFFER,
		dfg1394_fw_on_iso_receive
	);

	if (node->iso == NULL) {
		PRINT_ERR("failed to create iso context\n");
		return -ENOMEM;
	}

	dfg1394_firewire.channel_allocation[channel] = node;
	
	return hpsb_iso_recv_start( node->iso, -1, -1, -1);
}

/**
 * @brief stop to receive iso packets
 *
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_stop_iso_receive( dfg1394_node_t* node )
{
	u8 channel = node->iso_channel;
	if( channel >= 64 
	||  dfg1394_firewire.channel_allocation[channel] == NULL) {
		return -EINVAL;
	}

	dfg1394_firewire.channel_allocation[channel] = NULL;

	if (node->iso != NULL) {
		hpsb_iso_stop( node->iso );
		hpsb_iso_shutdown( node->iso );
		node->iso = NULL;
	}

	return SUCCESS;
}



/**
 * @brief check if the specified node is a DFG/1394-1
 */
inline dfg1394_bool_t dfg1394_fw_is_dfg1394_node( 
	struct hpsb_host *host, int node_idx, u32 generation )
{
	u32 spec_id;
	u32 sw_version;
	
	int ret;

	ret = dfg1394_fw_read_node_spec_id( host, node_idx, generation, &spec_id );
	if( !IS_SUCCESS( ret ) || (spec_id != 0x748 ) )
	{
		return 0;
	}
	ret = dfg1394_fw_read_node_sw_version( host, node_idx, generation, &sw_version );
	if( !IS_SUCCESS( ret ) || (sw_version != 0x526f6e ) )
	{
		return 0;
	}
	

	return 1;
}

/** @brief FCP command base code */
#define FCP_COMMAND 0xE031323300000000LL

/**
 * @brief Sends a FCP command to the specified node
 *
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_send_fcp_command( 
	dfg1394_node_t* node,	u8 cmd, u8 arg1, u8 arg2 )
{
	u64 fcp_cmd; 
	int ret;

	/* create command */	
	fcp_cmd = FCP_COMMAND | (cmd << 24) | (arg1 << 16) | (arg2 << 8);
//	PRINT_TRACE("send FCP command: %016llx\n", fcp_cmd );

	fcp_cmd = cpu_to_be64( fcp_cmd );

	/* send command */
	ret = dfg1394_fw_write(
		node->host, node->generation, node->node | LOCAL_BUS,
		CSR_REGISTER_BASE + CSR_FCP_COMMAND,
		(quadlet_t*) &fcp_cmd, 8 );
	
	return ret;
}

/**
 * @brief Sends a FCP command to the specified node
 *
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_send_fcp3_command( 
	dfg1394_node_t* node,	u8 cmd, u8 arg1, u8 arg2, u8 arg3 )
{
	u64 fcp_cmd; 
	int ret;

	/* create command */	
	fcp_cmd = FCP_COMMAND | (cmd << 24) | (arg1 << 16) | (arg2 << 8) | arg3;
/* 	PRINT_INFO("send FCP command: %016llx\n", fcp_cmd ); */

	fcp_cmd = cpu_to_be64( fcp_cmd );

	/* send command */
	ret = dfg1394_fw_write(
		node->host, node->generation, node->node | LOCAL_BUS,
		CSR_REGISTER_BASE + CSR_FCP_COMMAND,
		(quadlet_t*) &fcp_cmd, 8 );
	
	return ret;
}

/**
 * @brief scans the bus for a DFG/1394-1 device
 *
 * @return @c SUCCESS or @c -ENODEV
 */
int dfg1394_fw_scan_for_device( dfg1394_node_t* dev )
{
	struct hpsb_host *host = dfg1394_firewire.host;
	int node_count = host->node_count;
	int node_idx = 0;
	int generation = 0;

	generation = get_hpsb_generation( host );
	for ( node_idx = 0; node_idx < node_count; ++node_idx ) {
		if ( dfg1394_fw_is_dfg1394_node( host, node_idx, generation ))
		{
			dev->host = host;
			dev->node = node_idx;
			dev->generation = generation;

			dfg1394_firewire.node_allocation[node_idx] = dev;

			return SUCCESS;
		}
	}

	dev->host = NULL;
	dev->node = -1;
	dev->generation = 0;

	return -ENODEV;
}


/* ************************************************************************ */
/**                                                                         
 * @name IEEE1394 highlevel callbacks 
 */
/* ************************************************************************ */
/*@{*/


/**
 * @brief add_host
 *
 * Called on driver registration for each ieee1394 bus attached to the system.
 * Is also called when a new bus gets attached later on ( hotplugging device. )
 */
void dfg1394_fw_hlop_add_host( struct hpsb_host *host )
{
//	PRINT_TRACE1("dfg1394hw_hlop_add_host()\n");

	
	if ( dfg1394_firewire.host != NULL ) {
		PRINT_WARNING("Only one IEEE1394 host supported, one host ignored.\n");
		return;
	}
	dfg1394_firewire.host = host;
}



/**
 * @brief remove_host
 *
 * Called on driver unregistration for each ieee1394 bus attached to 
 * the system.
 *
 * Is also called when a bus gets detached (hotplugging a controllere).
 */
void dfg1394_fw_hlop_remove_host( struct hpsb_host *host )
{
//	PRINT_TRACE1("dfg1394hw_hlop_remove_host()\n");
//	dfg1394_fw_shutdown_host( host );
	int node_count = host->node_count;
	int node_idx = 0;
	u32 generation = 0;

	dfg1394_node_t* node;

	/* ignore hosts we don't support */
	if ( host != dfg1394_firewire.host) return;

	generation = get_hpsb_generation( host );
	for ( node_idx = 0; node_idx < node_count; ++node_idx ) {

		node = dfg1394_firewire.node_allocation[ node_idx ];
		dfg1394_firewire.node_allocation[ node_idx ] = NULL;

		if ( node ) {
			node->host = NULL;
			node->node = -1;
			node->generation = 0;
		}
	}

	dfg1394_firewire.host = NULL;
}

/**
 * @brief host reset
 */
void dfg1394_fw_hlop_host_reset( struct hpsb_host *host )
{
	PRINT_INFO("HOST RESET\n");
}

/**
 * @brief fcp_request
 *
 * stores received response in node's fcp context and wakes up
 * waiting clients 
 *
 */
void dfg1394_fw_hlop_fcp_request(
	struct hpsb_host *host, int nodeid, int is_response, 
	int cts, u8* data, u32 length )
{
	dfg1394_node_t* node = NULL;
	u64 rsp              = 0;
	int node_idx         = nodeid & ~LOCAL_BUS;
  
	if ( length < 8) return;

	rsp = *((u64*)data);
	rsp = be64_to_cpu( rsp );

	if ( !is_response ) {
		PRINT_WARNING("ignoring received fcp command: %016llx\n", rsp );
	}


	node = dfg1394_firewire.node_allocation[ node_idx ];
	if ( node == NULL ) {
		PRINT_WARNING("ignoring received fcp response "
			      "to unknown node/host: %016llx\n", rsp );
		return;
	}
 
	dfg1394_node__process_fcp_response( node, rsp );
 }

/**
 * @brief IEEE 1394 highlevel operations
 */
static struct hpsb_highlevel dfg1394_hl =
{
	name:        "DFG1394-1 Video Converter", 
	add_host:    dfg1394_fw_hlop_add_host,
	remove_host: dfg1394_fw_hlop_remove_host,
	host_reset:  dfg1394_fw_hlop_host_reset,
	fcp_request: dfg1394_fw_hlop_fcp_request,
};

/*@}*/

/**
 * Callback: Called by the IEEE1394 driver if iso packets have been received
 */
void dfg1394_fw_on_iso_receive( struct hpsb_iso* iso )
{
	dfg1394_node_t* node = 
		dfg1394_firewire.channel_allocation[iso->channel];
	struct hpsb_iso_packet_info* info = NULL;
	int n_ready = hpsb_iso_n_ready(iso);
	int pos     = iso->first_packet; 
	int i;


	for (i = 0; i < n_ready; i++ ) {

		info = &iso->infos[pos];
		if ( info->len > 8 ) {

			// let the node process this iso packet
/* 			 dfg1394_node__process_iso_packet(   */
/* 				node,  */
/* 				info->sy,  */
/* 				iso->data_buf.kvirt + info->offset + 8, */
/* 				info->len-8 */
/* 			);   */
			 dfg1394_node__process_iso_packet(  
				node, 
				info->sy, 
				iso->data_buf.kvirt + info->offset,
				info->len
			);  
		}

		if ( ++pos >= iso->buf_packets)
		{
			pos = 0;
		}
	}

	hpsb_iso_recv_release_packets( iso, n_ready );	
}



/* ************************************************************************ *\
 *                                                                          *
 *   Hardware subsystem control                                             *
 *                                                                          *
\* ************************************************************************ */

/**
 * @brief Initializes the firewire layer
 *
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_init( void )
{
	int i;

	/*
	 * initialize firewire structure
	 */
	for ( i=0; i<64; ++i ) {
		dfg1394_firewire.node_allocation[i] = NULL;
		dfg1394_firewire.channel_allocation[i] = NULL;
	}

	/*
	 * register ieee1394 highlevel device
	 */
	hpsb_register_highlevel( &dfg1394_hl );
	dfg1394_firewire.hl_handle =  &dfg1394_hl;

	//! @todo add error handling here?

	if (dfg1394_firewire.hl_handle == NULL) 
		return -ENOMEM;


	return SUCCESS;
}

/**
 * @brief Cleans up the firewire layer
 *
 * @return @c SUCCESS or a negative error code 
 */
int dfg1394_fw_exit(void)
{
	if ( dfg1394_firewire.hl_handle != NULL ) 
		hpsb_unregister_highlevel( dfg1394_firewire.hl_handle );

	dfg1394_firewire.hl_handle = NULL;

	return SUCCESS;
}

