/***************************************************************************
                       dfg1394_node.c  -  description
                             -------------------
 Copyright (c) 2003 Norbert Wagner <nw@softwarekombinat.de>



    created              : 2003-03-06 Norbert Wagner <nw@softwarekombinat.de>
    modified             : 2003-10-07 Arne Caspari
                           - made compatible with firmware version 2.xx
                           2003-11-17 Arne Caspari
                           - added contrast, hue, saturation, brightness, sharpness support
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "dfg1394.h" 

#include "linux/list.h"
#include "linux/errno.h"
#include "linux/sched.h"


/* ************************************************************************ */
/** 
 *   @name dfg1394_node: private FCP related methods       
 */
/* ************************************************************************ */
/* @{                                                                       */

/** @brief FCP command codes */
enum {
	FCP_MSG_READ_I2C_BYTE_MSG =  0x10,
	FCP_MSG_WRITE_I2C_BYTE_MSG = 0x11,
	FCP_MSG_MODE_SELECT =        0x12,
	FCP_MSG_GET_FIRMWARE =       0x13,
	FCP_MSG_INPUT_SELECT =       0x14,
	FCP_MSG_CONTROL_TRANSMIT =   0x16,
	FCP_MSG_FREQUENCY_SELECT =   0x1B
};

#define SAA7112_ADDR 0x42

enum 
{
	I2C_SHARPNESS  = 0x09,
	I2C_BRIGHTNESS = 0x0A,
	I2C_CONTRAST   = 0x0B,
	I2C_SATURATION = 0x0C,
	I2C_HUE        = 0x0D,
};



/**
 * @brief Set FCP_DONE Flag
 */
inline void dfg1394_node__set_FCP_DONE( dfg1394_node_t * node, u64 response )
{
	node->fcp_response = response;
 	set_bit( DFG1394_NODE_FLAG__FCP_DONE, &node->flags );
	wake_up_interruptible( &node->fcp_response_wait );

}

/**
 * @brief Clear FCP_DONE Flag
 */
inline void dfg1394_node__clear_FCP_DONE( dfg1394_node_t* node )
{
 	clear_bit( DFG1394_NODE_FLAG__FCP_DONE, &node->flags );
}

/**
 * @brief Check FCP_DONE Flag
 */
inline int dfg1394_node__is_FCP_DONE( dfg1394_node_t* node )
{
 	return test_bit( DFG1394_NODE_FLAG__FCP_DONE, &node->flags );
}

/**
 * @brief FCP positive response code
 */
#define FCP_RSP_CODE_ACK         0xAA


/**
 * @brief executes the specified FCP command
 */
int dfg1394_node__execute_fcp_command(
	dfg1394_node_t* node, 
	u8 cmd,
	u8 arg1, u8 arg2,
	u8* ret1, u8* ret2 )
{
	int ret = SUCCESS;
	u8 code;

	if ( node->node < 0 || node->node >= 64 )
		return -EINVAL;

	/* lock fcp context to sequence fcp commands 
	   (no parallel fcp commands allowed) */
	if ( down_interruptible( &node->fcp_busy_sema ) )
		return -ERESTARTSYS;

	dfg1394_node__clear_FCP_DONE( node );

	/* let the firewire layer send the command */
	ret = dfg1394_fw_send_fcp_command( node, cmd, arg1, arg2 );
	if ( IS_FAILURE( ret ))
		return ret;
	
	/* wait for the response 
	   the FCP_DONE flag will be set by the 
	   dfg1394_process_fcp_response() handler function
	 */
	if ( wait_event_interruptible(
			node->fcp_response_wait, 
			dfg1394_node__is_FCP_DONE( node ))) {
		ret = -ERESTARTSYS;
		goto exit;
	}
		
	/* parse response */
//	PRINT_TRACE("recv FCP response: %016llx\n", node->fcp_response );

	//! @todo move parsing code into firewire layer
	code = (node->fcp_response >> 16) & 0xff;
	if ( ret1 != NULL ) *ret1 = (node->fcp_response >> 8) & 0xff;
	if ( ret2 != NULL ) *ret2 = (node->fcp_response >> 0) & 0xff;

	if ( code != FCP_RSP_CODE_ACK ) ret = -EIO;


exit:

	/* unlock fcp context */
	up( &node->fcp_busy_sema );
	return ret;	
}

/**
 * @brief executes the specified FCP command
 */
int dfg1394_node__execute_fcp3_command(
	dfg1394_node_t* node, 
	u8 cmd,
	u8 arg1, u8 arg2, u8 arg3,
	u8* ret1, u8* ret2 )
{
	int ret = SUCCESS;
	u8 code;

	if ( node->node < 0 || node->node >= 64 )
		return -EINVAL;

	/* lock fcp context to sequence fcp commands 
	   (no parallel fcp commands allowed) */
	if ( down_interruptible( &node->fcp_busy_sema ) )
		return -ERESTARTSYS;

	dfg1394_node__clear_FCP_DONE( node );

	/* let the firewire layer send the command */
	ret = dfg1394_fw_send_fcp3_command( node, cmd, arg1, arg2, arg3 );
	if ( IS_FAILURE( ret ))
		return ret;
	
	/* wait for the response 
	   the FCP_DONE flag will be set by the 
	   dfg1394_process_fcp_response() handler function
	 */
	if ( wait_event_interruptible(
			node->fcp_response_wait, 
			dfg1394_node__is_FCP_DONE( node ))) {
		ret = -ERESTARTSYS;
		goto exit;
	}
		
	/* parse response */
//	PRINT_TRACE("recv FCP response: %016llx\n", node->fcp_response );

	//! @todo move parsing code into firewire layer
	code = (node->fcp_response >> 16) & 0xff;
	if ( ret1 != NULL ) *ret1 = (node->fcp_response >> 8) & 0xff;
	if ( ret2 != NULL ) *ret2 = (node->fcp_response >> 0) & 0xff;

	if ( code != FCP_RSP_CODE_ACK ) ret = -EIO;


exit:

	/* unlock fcp context */
	up( &node->fcp_busy_sema );
	return ret;	
}

/**
 * @brief Callback: Called by the firewire layer for each received FCP response
 *
 */
void dfg1394_node__process_fcp_response( dfg1394_node_t *node, u64 response )
{
	dfg1394_node__set_FCP_DONE( node, response );
}

/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_node: private ISO related methods 
 */
/* ************************************************************************ */
/* @{                                                                       */

/**
 * @brief Values of the @c sy header field of ieee 1394 isochronous packets.
 */
enum {
	SY_START = 1,
	SY_EVEN  = 0,
	SY_ODD   = 4,
};


/**
 * @brief Callback: Called by the firewire layer for each received ISO packet
 */
void dfg1394_node__process_iso_packet( 
	dfg1394_node_t *node, u8 sy, u8* data, u32 length )
{
	dfg1394_buffer_t* cbuf;
	u8             last_sy;
	dfg1394_bool_t is_new_field;
	dfg1394_bool_t is_new_frame;
	dfg1394_bool_t is_done;
	int            copy = 0;
	u8 current_field = 0;

	// do not process packets not containing any video data
	if ( length == 8 ) 
		return;

	/* -------- lock stream state */
	spin_lock(&node->stream_lock);

	/* make sure this is no zombie packet */
	if (!dfg1394_node__is_transmitting(node)) 
		goto exit;

	/*
	 * Update stream state
	 */ 
	
	last_sy = node->stream_last_sy;
	node->stream_last_sy = sy;

	is_new_field = (sy == SY_START && last_sy != SY_START);

	if( node->firmware_version.major < 2 )
	{
		is_new_frame = (sy == SY_START && last_sy == SY_EVEN );
		// skip cip header -- not used by old firmware
		data += 8;
		length -= 8;
	}
	else
	{
		current_field = *(data + 7);
		
		is_new_frame = (sy == SY_START && current_field == SY_ODD );

		// skip cip header
		data += 8;
		length -= 8;
	}
	
	is_done = FALSE;

	/* check if we are (still) synchronizing */
	if ( is_new_frame ) {
		node->stream_is_syncing = FALSE;
	}

	if ( node->stream_is_syncing )
		goto exit;

	/* update field type */
	if ( is_new_field ) {
		if( node->firmware_version.major < 2 )
		{
			node->stream_field_type = 
				(is_new_frame ? DFG1394_FT_ODD : DFG1394_FT_EVEN );
		}
		else
		{
			node->stream_field_type = current_field == SY_ODD ? DFG1394_FT_ODD : DFG1394_FT_EVEN;
		}
	}

	/*
	 * Update cap_buffer
	 */ 

	cbuf = node->stream_buffer;

	if ( is_new_frame ) {

		if(cbuf != NULL && cbuf->offset != cbuf->frame_size) {
			PRINT_WARNING("invalid frame size, frame lost!\n"); 
				atomic_inc( &node->stat_frames_lost );
		} else { /* commit filled buffer, if we have one */
			if (cbuf != NULL) {
				dfg1394_device__push_filled_buffer(
					(dfg1394_device_t*)node, cbuf,
					node->stream_frame_count );
				atomic_inc( &node->stat_frames_captured );
			} else {
				atomic_inc( &node->stat_frames_dropped );
			}
			
			cbuf = dfg1394_device__pop_empty_buffer(
				(dfg1394_device_t*)node );
			node->stream_buffer = cbuf; 
		}
	}


	/* 
	 * Weave this field into the current frame 
	 */
	if ( cbuf != NULL ) {
		
		/* reset offset if a new field begins */
		if (is_new_field) {
			cbuf->offset = 
				node->stream_field_type == DFG1394_FT_ODD
				? 0 : cbuf->line_size;
			node->stream_line_left = cbuf->line_size;
		}


		while (length > 0) {

			/* if the current line is full, move to the
                           next line of this field by skipping one
                           line  */
			if (node->stream_line_left <= 0) {
				node->stream_line_left = cbuf->line_size;
				cbuf->offset += cbuf->line_size;
			}
			
			/* we copy up to the line end at max. */
			copy    = length < node->stream_line_left
				? length : node->stream_line_left;

			/* check if copying is valid */
			if ( cbuf->frame_size < cbuf->offset + copy ) {
				PRINT_ERR("process_iso_packet(): overflow!\n");
				/* invalidate packet, so it will be counted
                                   as "lost", see above */
				cbuf->offset = 0;
				goto exit;
			}

			/* copy the data and adjust pointers and lengths */
			memcpy( cbuf->data + cbuf->offset, data, copy ); 
			cbuf->offset += copy;
			data += copy;
			length -= copy;
			node->stream_line_left -= copy;
		}
	}

 	if ( is_new_frame ) 
		node->stream_frame_count ++;

 exit:
	spin_unlock(&node->stream_lock);
	/* -------- unlock stream state */

}



/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_node: Management 
 */
/* ************************************************************************ */
/* @{                                                                       */


/**
 * @brief Initializes the specified node
 *
 * @param node      [in]  @c this pointer
 */
void dfg1394_node__initialize( dfg1394_node_t* node )
{

	/* initialize IEEE1395 node related members 
	 */
	node->host = NULL;
	node->node = -1;
	node->generation = 0;
	node->flags = 0;

	/* initialize DFG1394-1 hardware settings
	 */
	node->iso_channel = -1; // => no transmission
	node->input = DFG1394_INPUT_DEFAULT;
	node->pix_mode = DFG1394_PIX_MODE_DEFAULT;
	node->frequency = DFG1394_FREQ_DEFAULT;


	/* initialize FCP handling
	 */
	node->fcp_response = 0;
	init_waitqueue_head( &node->fcp_response_wait );
	sema_init( &node->fcp_busy_sema, 1 );


	/* initialize iso streaming state
	 */
	   
	node->stream_last_sy = -1;
	node->stream_frame_count = 0;
	node->stream_is_syncing = TRUE;
	spin_lock_init( &node->stream_lock );
	spin_lock_init( &node->lock );

	/* initialize statistics 
	 */

	atomic_set( &node->stat_frames_captured, 0);
	atomic_set( &node->stat_frames_dropped, 0);
	atomic_set( &node->stat_frames_lost, 0);

}


/**
 * @brief Starts the node, initializes the node hardware
 *
 * @param node      [in]  @c this pointer
 *
 * The node hardware is initialized by setting the
 * configured input, mode and frequency 
 */
int dfg1394_node__start( dfg1394_node_t* node )
{
	int ret = SUCCESS;

	if (IS_FAILURE(ret = dfg1394_fw_scan_for_device( node ))) {
		PRINT_ERR("No DFG/1394-1 device found.\n"); 
		return ret;
	}

	ret =  dfg1394_node__set_frequency( node, node->frequency );
	if (IS_FAILURE(ret))
		return ret;
			
	ret = dfg1394_node__set_mode( node, node->pix_mode );
 	if (IS_FAILURE(ret))
		return ret;

	ret = dfg1394_node__set_input( node, node->input  );
 	if (IS_FAILURE(ret))
		return ret;
	ret = dfg1394_node__get_firmware_version( node );
	
	return ret;
}

/**
 * @brief Aborts everything the node might be doing
 *
 * @param node      [in]  @c this pointer
 *
 * Any active ISO or FCP communication is killed.
 */
void dfg1394_node__drop_dead( dfg1394_node_t* node )
{
	/* disable the iso_receive_handler */
	clear_bit( DFG1394_NODE_FLAG__TRANSMITTING, &node->flags );

	dfg1394_fw_stop_iso_receive( node );
	
	/* wait for process_iso_packet() to */
	spin_unlock_wait(&node->stream_lock);

	/* fake invalid FCP response to abort pending fcp commands*/
	dfg1394_node__set_FCP_DONE(node, 0x00000000000000000);
	
}

/*@}*/

/* ************************************************************************ */
/** 
 *   @name dfg1394_node: Configuration and Control
 */
/* ************************************************************************ */
/* @{                                                                       */


int dfg1394_node__get_firmware_version( dfg1394_node_t* node)
{
	dfg1394_firmware_version_t vers;
	int ret;

	
	ret = dfg1394_node__execute_fcp_command(
		node,
		FCP_MSG_GET_FIRMWARE,
		0,
		0,
		&vers.major,
		&vers.minor
	);


	if( IS_SUCCESS( ret ) )
	{
		spin_lock( &node->lock);
		node->firmware_version = vers;
		spin_unlock(&node->lock);

		PRINT_INFO( "firmware version major = %d, minor = %d\n", vers.major, vers.minor );
	}
	else
	{
		PRINT_ERR( "failed to read firmware version\n" );
	}

	return ret;
}



/**
 * @brief Gets the current configured input
 *
 * @param node      [in]  @c this pointer
 * @return the configured input 
 */
dfg1394_input_t dfg1394_node__get_input(dfg1394_node_t* node)
{
	dfg1394_input_t input;

	spin_lock( &node->lock);
	input = node->input;
	spin_unlock(&node->lock);

	return input;
}

/**
 * @brief Sets the converter's input 
 *
 * @param node      [in]  @c this pointer
 * @param req_input [in]  The requested input channel (@c 0xff to autoselect)
 * @param input     [out] Contains the selected input
 *                        channel after successful completion (may be @c NULL)
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__set_input( dfg1394_node_t* node, dfg1394_input_t req_input )
{
	int ret = SUCCESS;
	u8 result;

	ret = dfg1394_node__execute_fcp_command(
		node,
		FCP_MSG_INPUT_SELECT,
		req_input,
		0,
		&result,
		NULL
	);

	if ( IS_SUCCESS( ret )) {

		spin_lock(&node->lock);
		node->input = result;
		spin_unlock(&node->lock);

		PRINT_INFO("node: input set to %d\n", result );
	} else {
		PRINT_WARNING(
			"node: setting input to %d FAILED!\n", req_input );
	}

	return ret;
}


/**
 * @brief Gets the converter's current frequency settings
 *
 * @param node      [in]  @c this pointer
 * @return the configured frequency
 */
dfg1394_freq_t dfg1394_node__get_frequency(dfg1394_node_t* node)
{
	dfg1394_freq_t freq;

	spin_lock(&node->lock);
	freq = node->frequency;
	spin_unlock(&node->lock);

	return freq;
}

/**
 * @brief Sets the converter's frequency (PAL or NTSC)
 *
 * @param node      [in]  @c this pointer
 * @param freq      [in]  the frequency
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__set_frequency( 
	dfg1394_node_t* node, dfg1394_freq_t req_freq )
{
	int ret = SUCCESS;
	u8 result;
	
	ret = dfg1394_node__execute_fcp_command(
		node,
		FCP_MSG_FREQUENCY_SELECT,
		req_freq,
		0,
		&result,
		NULL
	);

	if ( IS_SUCCESS( ret )) {

		spin_lock(&node->lock);
		node->frequency = result;
		spin_unlock(&node->lock);

		PRINT_INFO(
			"node: frequency set to %d\n", result );
	} else {
		PRINT_WARNING(
			"node: setting frequency to %d FAILED!\n", req_freq );
	}

	return ret;
}


/**
 * @brief Set the brightness property
 *
 * @param node      [in]  @c this pointer
 * @param value     [in]  @c new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__set_brightness( 
	dfg1394_node_t *node, u8 value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command( 
		node, 
		FCP_MSG_WRITE_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_BRIGHTNESS, 
		value, 
		0, 
		0 );

	return ret;
}

int dfg1394_node__get_brightness(
	dfg1394_node_t *node, u8 *value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command(
		node, 
		FCP_MSG_READ_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_BRIGHTNESS, 
		0, 
		0, 
		value );

	return ret;
}

/**
 * @brief Set the contrast property
 *
 * @param node      [in]  @c this pointer
 * @param value     [in]  @c new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__set_contrast(
	dfg1394_node_t *node, u8 value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command( 
		node, 
		FCP_MSG_WRITE_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_CONTRAST, 
		value, 
		0, 
		0 );

	return ret;
}
	
/**
 * @brief Returns the current contrast setting
 *
 * @param node      [in]  @c this pointer
 * @param value     [out] @c memory are to store new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__get_contrast(
	dfg1394_node_t *node, u8 *value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command(
		node, 
		FCP_MSG_READ_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_CONTRAST, 
		0, 
		0, 
		value );

	return ret;
}

/**
 * @brief Set the hue property
 *
 * @param node      [in]  @c this pointer
 * @param value     [in]  @c new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__set_hue(
	dfg1394_node_t *node, u8 value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command( 
		node, 
		FCP_MSG_WRITE_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_HUE, 
		value, 
		0, 
		0 );

	return ret;
}
	
/**
 * @brief Returns the current hue setting
 *
 * @param node      [in]  @c this pointer
 * @param value     [out] @c memory are to store new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__get_hue(
	dfg1394_node_t *node, u8 *value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command(
		node, 
		FCP_MSG_READ_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_HUE, 
		0, 
		0, 
		value );

	return ret;
}

/**
 * @brief Set the saturation property
 *
 * @param node      [in]  @c this pointer
 * @param value     [in]  @c new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__set_saturation(
	dfg1394_node_t *node, u8 value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command( 
		node, 
		FCP_MSG_WRITE_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_SATURATION, 
		value, 
		0, 
		0 );

	return ret;
}
	
/**
 * @brief Returns the current saturation setting
 *
 * @param node      [in]  @c this pointer
 * @param value     [out] @c memory are to store new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__get_saturation(
	dfg1394_node_t *node, u8 *value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command(
		node, 
		FCP_MSG_READ_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_SATURATION, 
		0, 
		0, 
		value );

	return ret;
}

/**
 * @brief Set the current sharpness property
 *
 * @param node      [in]  @c this pointer
 * @param value     [in]  @c new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__set_sharpness(
	dfg1394_node_t *node, u8 value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command( 
		node, 
		FCP_MSG_WRITE_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_SHARPNESS, 
		value, 
		0, 
		0 );

	return ret;
}
	
/**
 * @brief Returns the current sharpness setting
 *
 * @param node      [in]  @c this pointer
 * @param value     [out] @c memory are to store new value
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__get_sharpness(
	dfg1394_node_t *node, u8 *value )
{
	int ret = SUCCESS;
	
	ret = dfg1394_node__execute_fcp3_command(
		node, 
		FCP_MSG_READ_I2C_BYTE_MSG, 
		SAA7112_ADDR, 
		I2C_SHARPNESS, 
		0, 
		0, 
		value );

	return ret;
}

/**
 * @brief Gets the converter's configured video mode
 *
 * @param node      [in]  @c this pointer
 * @return the configured mode
 */
dfg1394_pix_mode_t dfg1394_node__get_mode(dfg1394_node_t* node)
{
	dfg1394_pix_mode_t mode;
	
	spin_lock(&node->lock);
	mode = node->pix_mode;
	spin_unlock(&node->lock);
	
	return mode;
}

/**
 * @brief Sets the converter's video mode
 *
 * @param node      [in]  @c this pointer
 * @param mode      [in]  the video mode
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__set_mode( dfg1394_node_t* node, dfg1394_pix_mode_t mode )
{
	int ret = SUCCESS;
	static u8 hw_modes[] = { 4, 6, 1, 5, 2, 3, 7, 8, 9 };
	
	//                             
	if ( mode >= sizeof(hw_modes) )
		return -EINVAL;

	ret = dfg1394_node__execute_fcp_command(
		node,
		FCP_MSG_MODE_SELECT,
		hw_modes[mode],
		0,
		NULL,
		NULL
	);



	if ( IS_SUCCESS( ret )) {

		spin_lock(&node->lock);
		node->pix_mode = mode;
		spin_unlock(&node->lock);

		PRINT_INFO(
			"node: mode set to %d (hardware: %d)\n", 
			mode, hw_modes[mode] );
	} else {
		PRINT_WARNING(
			"node: setting mode to %d (hardware: %d) FAILED!\n", 
			mode, hw_modes[mode] );
	}

	return ret;
}

/**
 * @brief Make the converter start transmitting
 *
 * @param node      [in]  @c this pointer
 * @param node      [in]  channel the iso channel
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__start_transmission( dfg1394_node_t* node, u8 channel )
{
	int ret = SUCCESS;
	unsigned long flags;

	/* to disable the iso_receive_handler */
	clear_bit( DFG1394_NODE_FLAG__TRANSMITTING, &node->flags );

	/* make converter transmitting */
	ret = dfg1394_node__execute_fcp_command(
		node,
		FCP_MSG_CONTROL_TRANSMIT,
		channel,
		0x01, /* 01 -> enable */
		NULL,
		NULL
	);

	/* make us receiving */	
	if ( IS_SUCCESS( ret )) {

		node->iso_channel = channel;

		/* sync'ed reset of streaming state */
		spin_lock_irqsave(&node->stream_lock, flags);
		node->stream_is_syncing = TRUE;
		node->stream_frame_count = 0;
		node->stream_last_sy = SY_START;
		node->stream_buffer = NULL;
		node->stream_line_left = 0;
		spin_unlock_irqrestore(&node->stream_lock, flags);

		/* reset statistics */
		atomic_set( &node->stat_frames_captured, 0 );
		atomic_set( &node->stat_frames_lost, 0 );
		atomic_set( &node->stat_frames_dropped, 0 );

		ret = dfg1394_fw_start_iso_receive(  node  );
		
		if (IS_SUCCESS(ret))
			set_bit(DFG1394_NODE_FLAG__TRANSMITTING, 
				&node->flags );
		else
			dfg1394_node__stop_transmission( node );
	}

	/* give feedback */
	if ( IS_SUCCESS( ret )) {
		PRINT_INFO(
			"node: transmission started on channel %d\n", 
			(u32) channel );
	} else {
		PRINT_WARNING(
			"node: starting transmission on channel %d FAILED!\n", 
			(u32) channel );
	}

	return ret;
}

/**
 * @brief Make the converter stop transmitting
 *
 * @param node      [in]  @c this pointer
 * @return @c SUCCESS or a negative error code
 */
int dfg1394_node__stop_transmission( dfg1394_node_t* node )
{
	int ret = SUCCESS;

	PRINT_INFO("node: stopping transmission...\n" );

	clear_bit( DFG1394_NODE_FLAG__TRANSMITTING, &node->flags );

	/* make converter stop transmission */
	ret = dfg1394_node__execute_fcp_command(
		node,
		FCP_MSG_CONTROL_TRANSMIT,
		node->iso_channel,
		0x00, /* 00 -> disable */
		NULL,
		NULL
	);

	/* make us stop receiving */
	PRINT_INFO("node: stopping iso receive...\n" );
	dfg1394_fw_stop_iso_receive( node );

	if ( IS_SUCCESS( ret )) {
		node->iso_channel = -1;
		PRINT_INFO("node: transmission stopped\n" );
	} else {
		PRINT_WARNING("node: stopping transmission FAILED!\n" );
	}

	/* make sure that process_iso_packet() is done, so no further
	   processing takes place (The TRANSMITTING flag is cleared, so
	   new invocations of process_iso_packet() bail out before
	   doing any harm)
	*/
	spin_unlock_wait(&node->stream_lock);

	return ret;
}




/*@}*/
