/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the G100 and G200 video extension
 * registers.  
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Mike Bernson <mike@mlb.org>
 *	Leon van Stuivenberg <leonvs@iae.nl>
 *	Gernot Ziegler <gz@lysator.liu.se>
 */

/* Changes: 000219 - gz: .) Changed the picture parameter range 
                            (like brightness) to signed char.
                         .) Implemented VIDIOCGAUDIO.
			 .) Changed input channel order (to fix xawtv).
			    (_incompatible_ with the current buz_tools !!!!)
*/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/proc_fs.h>

#include <linux/version.h>
#include <linux/spinlock.h>

#include <linux/videodev.h>

#include <asm/uaccess.h>

#include "mgacore.h"
#include "aiee.h"
#include "mgabes.h"
#include "mgagrab.h"
#include "mgacap.h"
#include <linux/video_decoder.h>
#include <linux/video_encoder.h>


#define MIN(a,b) (((a)>(b))?(b):(a))
#define MAX(a,b) (((a)>(b))?(a):(b))


#if VIDEO_MAX_FRAME <= 32
#define   V4L_MAX_FRAME   32
#elif VIDEO_MAX_FRAME <= 64
#define   V4L_MAX_FRAME   64
#else
#error   "Too many video frame buffers to handle"
#endif
#define   V4L_MASK_FRAME   (V4L_MAX_FRAME - 1)

#define V4L_NAME      "V4L-mgavideo"   /* name of the driver */

#define V4L_ERR       KERN_ERR     V4L_NAME
#define V4L_DEBUG     KERN_INFO    V4L_NAME
#define V4L_INFO      KERN_INFO    V4L_NAME
#define V4L_WARNING   KERN_WARNING V4L_NAME

#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */



#define MGA_MAX 4
#define MGA_NAME "mga_v4l1cap"
#define MGA_MAX_WIDTH 720
#define MGA_MIN_WIDTH 8
#define MGA_MAX_HEIGHT 580
#define MGA_MIN_HEIGHT 8 



/* Buffer size is computed for a complete image (2 fields) in RGB 32 format */
/* It is an integer number of pages */
#define MAXV4LBUF 2
#define	MAXV4LSIZE ((((_VINWID * _VINHGT * 2 * 4) + (PAGE_SIZE-1))/PAGE_SIZE) * PAGE_SIZE)

static int v4l_nbufs = MAXV4LBUF;
static int v4l_bufsize = 1024 * 768 * 4;
///static int video_nr = -1;

module_param(debug, int, 0644);
module_param(v4l_nbufs, int, 0644);
///MODULE_PARM(video_nr,"i");
MODULE_LICENSE("GPL");


struct mgacap_v4l1 {
/****	struct video_device	vbi_dev; ******/
	struct mga_dev_s	*mga;
	int			id;
	int			initialized;
	int			user;

	/* Video for Linux parameters */
	struct video_buffer buffer;	/* Current buffer params */
	struct video_window window;	/* Current window params */
	int buffer_set, window_set;	/* Flags if the above structures are set */
	int video_interlace;	/* Image on screen is interlaced */

	WAIT_QUEUE	   captureq;

	int		v4l_grab_frame;	/* Frame number being currently grabbed */
	int		v4l_grab_seq;	/* Number of frames grabbed */

	/* V4L grab queue of frames pending */
	/* Put empty buffers at queue, fill buffers at head */

	/* Main memory buffers (these ones are shared) */
	u8		*v4l1_buffer;
	
	unsigned	v4l_que_head;
	unsigned	v4l_que_tail;
	unsigned	v4l_grab_head;
	unsigned	v4l_que_size;
	int		v4l_pend[V4L_MAX_FRAME+1]; /* +1 -> tell difference between full & empty */
	struct video_mmap v4l_pend_geometry[V4L_MAX_FRAME];
       	spinlock_t	v4l_queuelock; /* Lock for the queue data */
};


int mgacap_open(struct mgacap_v4l1* mv, int flags)
{
	if(flags == 0 && mv->user == 0) {
		dprintk("mgacap: opening device\n");
		mv->user++;

		mv->v4l_que_head = 0;
		mv->v4l_que_tail = 0;
		mv->v4l_grab_head = 0;

		// XXX MOD_INC_USE_COUNT;
		return 0;
	}

	return -EBUSY;
}
EXPORT_SYMBOL(mgacap_open);

void mgacap_close(struct mgacap_v4l1* mv)
{
	dprintk("mgacap: closing device\n");

	mgagrab_preview_enable(mv->mga->grab, 0);

	mv->user--;

	// XXX MOD_DEC_USE_COUNT;
}
EXPORT_SYMBOL(mgacap_close);



static int
mgacap_run_v4l_loop(struct mgacap_v4l1* mv)
{
	int buffer, head;
	const int timeout = HZ;
	int wait = timeout;

	while (mv->v4l_grab_head != mv->v4l_que_head) {
	
		buffer = mv->v4l_pend[ mv->v4l_grab_head ];
		
		/* grab_image can fail if no buffer availabe */
		/* (not the correct frame type, new scaling applied ..) */
			
		if (mgagrab_grab_image(mv->mga,
			&mv->v4l_pend_geometry[buffer],
			&mv->v4l1_buffer[buffer * v4l_bufsize], 
			v4l_bufsize) == 0) {

			head = mv->v4l_grab_head + 1;
			if (head >= mv->v4l_que_size) head = 0;
			mv->v4l_grab_head = head;
			
			wait = timeout;
			
		} else {
			if ( wait ) {
				interruptible_sleep_on_timeout(&mv->captureq, 1);
				if (signal_pending(current))
					return -ERESTART;
				wait--;
			} else {
				/* timeout; insert dummy image */
				memset( &mv->v4l1_buffer[buffer * v4l_bufsize], 0x10, v4l_bufsize );
				
				head = mv->v4l_grab_head + 1;
				if (head >= mv->v4l_que_size) head = 0;
				mv->v4l_grab_head = head;
				
				wait = timeout;
			}
		}
	}
	return 0;
}


int mgacap_ioctl(struct mgacap_v4l1* mv, unsigned int cmd, void *arg, int *handled)
{
	int rv;
	
	*handled = 1;
	
	switch (cmd) {

	case VIDIOCCAPTURE:
		{
			int v;

			if (copy_from_user(&v, arg, sizeof(v))) {
				return -EFAULT;
			}
			dprintk("mga_v4l1 ioctl VIDIOCCAPTURE: %d\n", v);
			mgagrab_preview_enable(mv->mga->grab, v);
			return 0;
		}

	case VIDIOCGWIN:
		{
			struct video_window vw = mv->window;

			dprintk("mga_v4l1 ioctl VIDIOCGWIN x=%d "
				"y=%d width=%d hieght=%d chormakey=%x "
				"flags=%x\n", vw.x, vw.y, vw.width, vw.height,
				vw.chromakey, vw.flags);
			if (copy_to_user(arg, &vw, sizeof(vw))) 
				return -EFAULT;
			
			return 0;
		}

	case VIDIOCSWIN:
		{
			struct video_window vw;

			if (copy_from_user(&vw, arg, sizeof(vw))) {
				return -EFAULT;
			}
			dprintk("mga_v4l1 ioctl VIDIOCSWIN: x=%d y=%d w=%d h=%d chromakey=%x\n", 
			       vw.x, vw.y, vw.width, vw.height, vw.chromakey);

			mv->window = vw;
			mv->window_set = 1;

			mgabes_set_window(mv->mga->bes, vw.x, vw.y, 
				vw.width, vw.height);

			//if ( vw.flags & VIDEO_WINDOW_CHROMAKEY ) ...
			
			mgabes_set_colorkey(mv->mga->bes, vw.chromakey << 8);
			if ((vw.chromakey == -1) ) {//|| (vw.chromakey == 0)) {
				mgabes_set_keymode(mv->mga->bes, 0);
			} 
			return 0;
		}

	case VIDIOCGFBUF:
		{
			dprintk( "mga_v4l1 ioctl VIDIOCGFBUF\n");
			if (copy_to_user(arg, &mv->buffer, sizeof(mv->buffer))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCSFBUF:
		{
			struct video_buffer	v;
			int			width;
			int			height;
			int			depth;
			int			pitch;

			if (copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			
			dprintk("mga_v4l1 ioctl VIDIOCSFBUF: base=%p w=%d h=%d depth=%d bpl=%d\n", 
			       v.base, v.width, v.height, v.depth, v.bytesperline);

			mga_display_size(mv->mga, &width, &height, &depth);
			pitch = mga_display_pitch( mv->mga );

			if (v.depth != depth) {
				printk("mgacap: display depth is %d trying to set %d\n",
				       depth, v.depth);
				v.depth = depth;
			}

			if (v.height != height) {
				printk("mgacap: display height is %d trying to set %d\n",
				       height, v.height);
				v.height = height;
			}

			if (v.width != width) {
				printk("mgacap: display width is %d trying to set %d\n",
				       width, v.width);
				v.width = width;
			}

			if (v.bytesperline != pitch) {
				printk("mgacap: display bytes per line  is %d trying to set %d\n",
				       pitch, v.bytesperline);
				v.bytesperline = pitch;
			}
				
			if (v.base != NULL) 
				mv->buffer.base = (void *)((size_t)v.base & ~3);
			
			mv->buffer.height = v.height;
			mv->buffer.width = v.width;
			mv->buffer.depth = v.depth;
			mv->buffer.bytesperline = v.bytesperline;

			if (mv->buffer.base)
				mv->buffer_set = 1;
			mv->window_set = 0;	/* The user should set new window parameters */
			return 0;
		}

		/* RJ: what is VIDIOCKEY intended to do ??? */
		/* gz:it is there to support colorkeying, I think */


	case VIDIOCSYNC:
		{
			int v;
			int i, b;
			unsigned long flags;
			if (copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			
			dprintk("mga_v4l1 ioctl VIDIOCSYNC %d\n", v);
			
			if (mv->v4l1_buffer == NULL) {
				printk ("VIDIOCSYNC : buffers not mmapped.\n");
				return -EFAULT;
			}

			if (v < 0 || v >= v4l_nbufs) {
				dprintk("mga_v4l1 VIDIOCSYNC buffer out of range\n");
				return -EINVAL;
			}
			
			i = -1;
			while (i == -1) {
				rv = mgacap_run_v4l_loop( mv );
				if (rv)
					return rv;
			
				spin_lock_irqsave(&mv->v4l_queuelock, flags);
				/* lookup index of bufnr v in queue portion with grabbed bufs */
				for (b = mv->v4l_que_tail; b != mv->v4l_grab_head; ) {
					if (mv->v4l_pend[b] == v) {
						i = b;
						break;
					}
					b++;
					if (b >= mv->v4l_que_size) b = 0;
				}
			
				if (i != -1) {
					/* buf found; move queue items if buf not at tail */
					while (i != mv->v4l_que_tail) {
						b = i - 1;
						if (b < 0) b = mv->v4l_que_size-1;
						mv->v4l_pend[ i ] = mv->v4l_pend[ b ];
						i = b;
					}
					/* remove item from queue */
					i++;
					if (i >= mv->v4l_que_size) i = 0;
					mv->v4l_que_tail = i;
				}
				else if ((i == -1) && (b == mv->v4l_que_head)) {
					/* buf not found, and at end of queue */
					spin_unlock_irqrestore(&mv->v4l_queuelock, flags);
					printk ("mgav4l1 VIDIOCSYNC : buffer not queued\n");
					return -EINVAL; /* Buffer not queued */
				}
				spin_unlock_irqrestore(&mv->v4l_queuelock, flags);
			}
			return 0;
		}

	case VIDIOCMCAPTURE:
		{
			struct video_mmap vm;
			struct video_mmap *geom;
			unsigned long flags;
			int b;
			int size;

			if (copy_from_user((void *) &vm, (void *) arg, sizeof(vm))) {
				return -EFAULT;
			}

			if (mv->v4l1_buffer == 0) {
				printk ("VIDIOCMCAPTURE : buffers not mmapped.\n");
				return -EFAULT;
			}
			dprintk("mga_v4l1 ioctl VIDIOCMCAPTURE frame=%d geom=%dx%d fmt=%d\n",
			       vm.frame, vm.width, vm.height, vm.format);

			switch(vm.format) {
				case VIDEO_PALETTE_RGB565:
				case VIDEO_PALETTE_RGB555:
					size = 2;
					break;
				case VIDEO_PALETTE_RGB24:
					size = 3;
					break;
				case VIDEO_PALETTE_RGB32:
					size = 4;
					break;
				case VIDEO_PALETTE_YUYV:
				case VIDEO_PALETTE_YUV422:
				case VIDEO_PALETTE_YUV422P:
				case VIDEO_PALETTE_YUV411P:
				case VIDEO_PALETTE_YUV420P:
					size = 2;  /* between 1 and 2 ... */
					break;
				case VIDEO_PALETTE_GREY:
					size = 1;
					break;
				default:
					dprintk("mga_v4l1 VIDIOCMCAPTURE unknown format\n");
					return -EINVAL;
			}

			if (vm.frame < 0 || vm.frame >= v4l_nbufs) {
				dprintk("mga_v4l1 VIDIOCMCAPTURE bad frame\n");
				return -EINVAL;
			}
		
			/* I disable this for the moment because xawtv requests big images */
			/* (it assumes max grabbed size is equal to max overlay size) */
			/* if (vm.width > 720) vm.width = 720; */
			/* if (vm.width < 64) vm.width = 64; */
			/* if (vm.height < 8) vm.height = 8; */
			/* if (vm.height > _VINHGT*2) vm.height = _VINHGT*2; */
			/* vm.width &= ~3 */
			/* width must be multiple of 4 */

			if (vm.width * vm.height * size > v4l_bufsize) {
				dprintk("mga_v4l1 VIDIOCMCAPTURE image size to large\n");
				return -EINVAL;
			}
			
			/* Check if buffer not already queued */
			spin_lock_irqsave(&mv->v4l_queuelock, flags);
			for (b = mv->v4l_que_tail; b != mv->v4l_que_head; ) {
				if (mv->v4l_pend[b] == vm.frame) {
					/* Buffer already queued */
				
					geom = &mv->v4l_pend_geometry[vm.frame];
					geom->frame = vm.frame;
					geom->width = vm.width;
					geom->height = vm.height;
					geom->format = vm.format;
			
					spin_unlock_irqrestore(&mv->v4l_queuelock, flags);
					dprintk("mga_v4l1 VIDIOCMCAPTURE frame already queued\n");
					/*return -EBUSY; */
					return 0;
				}
				b++;
				if (b >= mv->v4l_que_size) b = 0;
			}
			
			/* Add buffer at head */
			mv->v4l_pend[mv->v4l_que_head] = vm.frame;
			geom = &mv->v4l_pend_geometry[vm.frame];
			geom->frame = vm.frame;
			geom->width = vm.width;
			geom->height = vm.height;
			geom->format = vm.format;
			b = mv->v4l_que_head + 1;
			if (b >= mv->v4l_que_size) b = 0;
			mv->v4l_que_head = b;
			spin_unlock_irqrestore(&mv->v4l_queuelock, flags);

			if (copy_to_user(arg, &vm, sizeof(vm))) {
				return -EFAULT;
			}

			return 0;
		}

	case VIDIOCGMBUF:
		{
			struct video_mbuf vm;
			int i;

			dprintk( "mga_v4l1 ioctl VIDIOCGMBUF\n");
		
			vm.size = v4l_nbufs * v4l_bufsize;
			vm.frames = v4l_nbufs;
			for (i = 0; i < v4l_nbufs; i++) {
				vm.offsets[i] = i * v4l_bufsize;
			}
		
			if (copy_to_user(arg, &vm, sizeof(vm))) {
				return -EFAULT;
			}
			return 0;
		}


	case VIDIOCGCAPTURE:
        	{
			dprintk( "mga_v4l1 ioctl VIDIOCGCAPTURE: unsupported\n");
			return -EINVAL;
        	}

	case VIDIOCSCAPTURE:
        	{
			dprintk("mga_v4l1 ioctl VIDIOCSCAPTURE: unsupported\n");
			return -EINVAL;
        	}

	default:
		*handled = 0;
		return -ENOIOCTLCMD;
	}
	return 0;
}
EXPORT_SYMBOL(mgacap_ioctl);


static int v4lbuf_free(struct mgacap_v4l1 *mv)
{
	if (mv->v4l1_buffer == NULL)
		return 0;

	mga_rvfree(mv->v4l1_buffer, v4l_bufsize * v4l_nbufs);
	mv->v4l1_buffer = NULL;

	return 0;	
}


static int v4lbuf_alloc(struct mgacap_v4l1 *mv)
{
	if (mv->v4l1_buffer)
		return 0;

	mv->v4l1_buffer = (u8 *)mga_rvmalloc(v4l_bufsize * v4l_nbufs);
	if (mv->v4l1_buffer == NULL)
		return -ENOBUFS;

	printk("v4lbuf_alloc: %i x %d KB allocated\n", v4l_nbufs, v4l_bufsize/1024);

	return 0;
}


int mgacap_mmap(struct vm_area_struct *vma, struct mgacap_v4l1* mv
                , const char *adr, unsigned long size)
{
	/* Allocate memory (if not done) */
	if (mv->v4l1_buffer == NULL)
		if (v4lbuf_alloc(mv))
			return -EAGAIN;

	/* Check for size */
	if (size > v4l_nbufs * v4l_bufsize) {
		printk ("mga_mmap : size mismatch (requested %d, max %d)\n",
			(int) size, v4l_nbufs * v4l_bufsize);
		return -EINVAL;
	}

	if (remap_vmalloc_pages(vma, (unsigned long)adr, mv->v4l1_buffer, size))
		return -EAGAIN;

	return 0;
}
EXPORT_SYMBOL(mgacap_mmap);


int mgacap_get_depth(struct mgacap_v4l1* mv)
{
	return mv->buffer.depth;
}
EXPORT_SYMBOL(mgacap_get_depth);


void mgacap_get_unit(struct mgacap_v4l1* mv, struct video_unit *vu)
{
//	vu->vbi = mv->vbi_dev.minor;
}
EXPORT_SYMBOL(mgacap_get_unit);


int mgacap_get_pending(struct mgacap_v4l1* mv)
{
	int p;

	p = mv->v4l_que_head - mv->v4l_que_tail;
	if (p < 0)
		p += mv->v4l_que_size;;
	return p;
}
EXPORT_SYMBOL(mgacap_get_pending);


/* Bottom half handler for V4L grabbing */
static void capture_interrupt (void *v)
{
	struct mgacap_v4l1 *mv = (struct mgacap_v4l1*) v;

	wake_up_interruptible(&mv->captureq);
}



/*********************************

static long vbi_read(struct video_device *v, char *buf, unsigned long count,
		     int nonblock)
{
	return count;
}

static long vbi_write(struct video_device *dev, const char *buf, unsigned long count, int nonblock)
{
	return -EINVAL;
}



static unsigned int vbi_poll(struct video_device *dev, struct file *file,
	poll_table *wait)
{
#ifdef	NEVER
	struct bttv *btv=(struct bttv *)(dev-2);
#endif

	unsigned int mask = 0;

#ifdef	NEVER
	poll_wait(file, &btv->vbiq, wait);

	if (btv->vbip < VBIBUF_SIZE)
		mask |= (POLLIN | POLLRDNORM);
#endif
	return mask;
}

static int vbi_open(struct video_device *dev, int flags)
{
	return 0;   
}

static void vbi_close(struct video_device *dev)
{
}


static int vbi_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	return -EINVAL;
}

static int vbi_init_done(struct video_device *dev)
{
	return 0;
}

static struct video_device vbi_template=
{
	owner:          THIS_MODULE;
	name:		"mga vbi",
	type:		VID_TYPE_CAPTURE|VID_TYPE_TELETEXT,
	hardware:	VID_HARDWARE_BT848,
	open:		vbi_open,
	close:		vbi_close,
	read:		vbi_read,
	write:		vbi_write,
	poll:		vbi_poll,
	ioctl:		vbi_ioctl,
	initialize:	vbi_init_done,
};
*****************/


#include <linux/smp_lock.h>

static void mga_init(struct mgacap_v4l1 *mv, struct mga_dev_s *mga)
{
	mv->mga = mga;
	mv->user = 0;

	mga->cap_v4l1 = mv;

	mgagrab_attach(mv->mga);

	mv->buffer_set = 0;
	mv->buffer.base = NULL;
	mv->buffer.width = 0;
	mv->buffer.height = 0;
	mv->buffer.depth = 0;
	mv->buffer.bytesperline = 0;

	mv->video_interlace = 0;

	mv->v4l_que_head = 0;
	mv->v4l_que_tail = 0;
	mv->v4l_grab_head = 0;
	mv->v4l_que_size = sizeof(mv->v4l_pend) / sizeof(mv->v4l_pend[0]);	
	spin_lock_init(&mv->v4l_queuelock);
	
	init_waitqueue_head(&mv->captureq);

	mgagrab_register_bh(mga->grab, capture_interrupt, mv);

	mv->initialized = 1;
}


int init_module(void)
{
	struct mgacap_v4l1	 *mv;
	struct mga_dev_s *mga;
	int		 i;

	printk(KERN_INFO "Matrox Marvel and Rainbow Runner G Capture Driver (V4L)\n");

	for(i=0; (mga = mga_get_card(i)) != NULL ; i++) {
		if ((mv = kmalloc(sizeof(*mv), GFP_KERNEL)) == NULL) {
			return -ENOMEM;
		}
		memset( mv, 0, sizeof(*mv) );

/******************
		memcpy(&mv->vbi_dev,&vbi_template, sizeof(vbi_template));
		sprintf(mv->vbi_dev.name, "mgavbi--%u", i);

#if LINUX_VERSION_CODE >= 0x020405
		if(video_register_device(&mv->vbi_dev,VFL_TYPE_VBI,video_nr)<0)  {
#else
		if(video_register_device(&mv->vbi_dev,VFL_TYPE_VBI)<0)  {
#endif
                        mga_release_card(mga);
			return -EIO;
		}
********************/

		mga_init(mv, mga);
	}

	if(i == 0) {
		return -ENXIO;
	}
	return 0;
}


void cleanup_module(void)
{
	int i = 0;
	struct mgacap_v4l1* mv;
	struct mga_dev_s *mga;

	for(i=0; (mga = mga_get_card(i)) != NULL; i++) {
	
		mv = mga->cap_v4l1;

		if(mv->initialized) {
		
			mgagrab_deattach(mv->mga->grab);

/*********		video_unregister_device(&mv->vbi_dev); ************/

			/* need to release it 2 times */
			mga_release_card(mv->mga);
			mga_release_card(mv->mga);

			v4lbuf_free(mv);

			mga->cap_v4l1 = NULL;
			kfree(mv);
		}
	}
}

