/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200/G400 and Rainbow Runner-G series
 *
 * This module is an interface to the G100,G200 and G400 video extension
 * registers.  
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Mike Bernson <mike@mlb.org>
 *	Leon van Stuivenberg <leonvs@iae.nl>
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>

#include <linux/slab.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/videodev.h>

#include "aiee.h"
#include "mgacore.h"
#include "mgagrab.h"
#include "mgabes.h"


#define dprintk     if (grab_debug) printk

static int grab_debug = 0; /* insmod parameter */

#if LINUX_VERSION_CODE >= 0x020100
module_param(grab_debug, int, 0644);
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


#define	BUFFLAG_CLIENTLOCK	0x01
#define BUFFLAG_SERVERLOCK	0x02
#define	BUFFLAG_FIELDTYPE	0x04
#define BUFFLAG_GRAB_OK		0x08



struct mga_grab_s {
	struct	mga_dev_s	*mga;

	spinlock_t 		lock;

	long			vsync_active;

        /* address for even and odd frames */
        u32 			vin_addr[4];
        u32 			vbi_addr[4];
	u32			vinctl_value;	//VINCTL0/1
	u32			vin_size;
	u32			vbi_size;
	u32			vinctl;		//VINCTL

	u32			buffer_flags[4];
	u32			buffer_stamp[4];
	int			active_buffer;	/* buffer being grabbed (0..3) */

        int			preview_active;	/* shows if the preview is active in the moment */
	int			grabber_active;	/* refcount for vin grabber activation */

	int			next_active_buffer; /* next buffer being grabbed (0..3) */

	int			grabbed;
	int			vin_active;

	int			interrupt_count;

        /* registered bottom half of isr */
        struct work_struct	tqnode_dpc;

	/* Shall we launch the bottom half handler */
        int 			grab_needed;

	grab_image_func_t	on_grab_image;
};


int mgagrab_read_proc(char *page, char **start, off_t off,
                  int count, int *eof, void *data);

int mgagrab_write_proc(struct file *file, const char *buf,
                          unsigned long count, void *data);


void mgagrab_vsync(struct mga_dev_s *mga);
void mgagrab_set_grabsize (struct mga_grab_s *video,
	int width, int height, int format);


inline static void set_vinctl(struct mga_grab_s *video, u32 on, u32 off)
{
	video->vinctl &= ~off;
	video->vinctl |= on;
	mga_write(video->mga, VINCTL, video->vinctl);
}


static void
mgagrab_disable_grabber(struct mga_grab_s *video)
{
	set_vinctl(video, 0, 0x01); //deassert vinen

	mga_disable_irq(video->mga, MGA_IRQ_VINVSYNC, 
		(mga_interrupt_handler_t)mgagrab_vsync);
}


static void
mgagrab_enable_grabber(struct mga_grab_s *video)
{
	u32 vbi;
	struct mgabus_video_input_s fmt;
	
	if (mga_get_vipformat(video->mga, &fmt) != 0) {
		dprintk( "mgagrab: get_format failed\n" );
		return;
	}

	if (video->mga->mga_type == MGA_G200 || video->mga->mga_type == MGA_G100) {
		
		vbi = ((fmt.vbicnt0 & 0xf)<<4) | (fmt.vbicnt1 & 0xf);
		
		/* 01(rpvaliden)01(vbitasken)a(vbif1cnt)a(vbif0cnt)(01)vinen*/
		set_vinctl(video, 0x01 | (vbi<<8), 0xff<<8); /* for VALIGN=0 */

	} else if (video->mga->mga_type == MGA_G400) {
		/* programming sequence according to G400 spec:
			1) toggle vinen from 0 to 1
			2) set rest of VINCTL, as well as VINHEIGHT and VBICOUNT
			3) enable vsync int    (strange. too early..)
			4) xADDR
			5) VINCTL0-3
		*/
		set_vinctl(video, 0, 0x01); /* default, vinen off */
		set_vinctl(video, 0x01, 0); /* vinen on */
		set_vinctl(video, 0, 0);

		mga_write(video->mga, VINHEIGHT, (626/2) & 0x3FF  );
		
		vbi = ((fmt.vbicnt1&0x3f)<<24) | ((fmt.vblcnt1&0x3f)<<16) |
			((fmt.vbicnt0&0x3f)<<8) | (fmt.vblcnt0&0x3f);

		/* spec:  PAL:0x11181119, NTSC:0x0a130b14 */
		/*        field0                 field1             */
		/* 0a(vbicnt1)13(vinvbcnt1)0b(vbicnt0)14(vinvbcnt0) */
		/* mga_write(video->mga, VBICOUNT, 0x0a120a12 );  for VALIGN=1, VSE=1 */
		/* mga_write(video->mga, VBICOUNT, 0x0a060a12 );  for VALIGN=1, VSE=0 */
		mga_write(video->mga, VBICOUNT, vbi );
	}
	else {
		printk("vbi not set for unknown Gx00 type\n" );
	}

	mga_write(video->mga, VINADDR0, video->vin_addr[0]);
	mga_write(video->mga, VBIADDR0, video->vbi_addr[0]);

	mga_write(video->mga, VINADDR1, video->vin_addr[1]);
	mga_write(video->mga, VBIADDR1, video->vbi_addr[1]);

	if (video->mga->mga_type == MGA_G400) {
		mga_write(video->mga, VINADDR2, video->vin_addr[2]);
		mga_write(video->mga, VINADDR3, video->vin_addr[3]);
	}

	mga_write(video->mga, VINCTL0,  video->vinctl_value);
	mga_write(video->mga, VINCTL1,  video->vinctl_value);

	if (video->mga->mga_type == MGA_G400) {
		mga_write(video->mga, VINCTL2,  video->vinctl_value);
		mga_write(video->mga, VINCTL3,  video->vinctl_value);
	}
	
	mga_enable_irq(video->mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)mgagrab_vsync);
}


/* Reset grabbing and preview parameters for a new video format */
void
mgagrab_init_videoformat_internal(struct mga_grab_s *video)
{
	struct buf_data *buf = &(video->mga->v4lbuf);
	int i;
	unsigned long flags;
	struct mga_video_area_s *crop;
	int besflags;

	struct mgabus_video_input_s fmt;
	
	if (mga_get_vipformat(video->mga, &fmt) != 0) {
		printk( KERN_ERR "mgagrab: get_format failed\n" );
		memset( &fmt, 0, sizeof(fmt)); /*FIXME*/
	}
	
	/* Critical section */
	spin_lock_irqsave(&video->lock, flags);
	
	for (i = 0; i < 4; i++) {
		video->buffer_flags[i] = 0;
		video->buffer_stamp[i] = 0;
	}

	crop = &fmt.crop;
		
	/* End of critical section */	
	spin_unlock_irqrestore(&video->lock, flags);

	if (video->preview_active) {
		besflags = BES_DEINTERLACE | BES_HFILTER | BES_VFILTER | BES_DITHER | BES_WAITFIELD2;
//		if ( pclk >= 135000 )
//			besflags |= BES_ACC2XZOOM;

		/* (bm) BES_WAITFIELD2 indicates that the last field is the bottom field */
		mgabes_set_mode(video->mga->bes, besflags);
	
		mgabes_set_buffers(video->mga->bes, video->vin_addr[0], 
			video->vin_addr[2], _VINWID, _VINHGT_PAL, VIDEO_PALETTE_YUV422);

		mgabes_set_source(video->mga->bes, crop->x, crop->y,
			crop->width, crop->height);
	}

	/* Critical section */
	spin_lock_irqsave(&video->lock, flags);

	buf->image_width = _VINWID;
	buf->real_width = _VINWID;
	buf->image_height = crop->height/2;
	buf->fields = 1;
	buf->grab_width = crop->width;
	buf->grab_height = crop->height;
	buf->subimage_x = crop->x;
	buf->subimage_y = crop->y;
	buf->subimage_w = crop->width;
	buf->subimage_h = crop->height;

	buf->decoder_hscale = 1;

	buf->fields_wanted = FIELD_TOP + FIELD_BOTTOM;
	video->active_buffer = -1; /* Buffer being grabbed must be thrashed. */

	/* End of critical section */	
	spin_unlock_irqrestore(&video->lock, flags);

	/* Recompute scaling and bes parameters */
	mgagrab_set_grabsize(video, buf->grab_width, buf->grab_height,
		buf->grab_format);
}


void
mgagrab_init_videoformat(struct mga_grab_s *video)
{
	if ( video->grabber_active > 0 ) {
		mgagrab_disable_grabber( video );
	}
	
	mgagrab_init_videoformat_internal( video );

	if ( video->grabber_active > 0 ) {
		mgagrab_enable_grabber( video );
	}
}


void
mgagrab_activate(struct mga_grab_s *video)
{
	video->grabber_active++;
	dprintk( "mgagrab active %i\n", video->grabber_active );
	if ( video->grabber_active == 1 ) {
		mgagrab_init_videoformat_internal( video );
		mgagrab_enable_grabber( video );
	}
}


void
mgagrab_deactivate(struct mga_grab_s *video)
{
	if ( video->grabber_active == 0 ) {
		dprintk( "mgagrab: unbalanced deactivate\n" );
		return;
	}
	if ( video->grabber_active == 1 ) {
		mgagrab_disable_grabber( video );
	}
	video->grabber_active--;
	dprintk( "mgagrab d-active %i\n", video->grabber_active );
}


void
mgagrab_vipnotify(struct mga_grab_s *video)
{
	if (video->grabber_active != 0) {
	
		mgagrab_disable_grabber( video );
		
		mgagrab_init_videoformat_internal( video ); /*fixme: check if really safe to call */
		
		mgagrab_enable_grabber( video );
	}
}


void
mgagrab_preview_enable(struct mga_grab_s *video, int flag)
{
	int enable = flag ? 1 : 0;
	
	if ( video->preview_active != enable ) {
		video->preview_active = enable;
		if (enable) {
			mgagrab_activate(video);
			mgabes_activate(video->mga->bes);
		} else {
			mgabes_desactivate(video->mga->bes);
			mgagrab_deactivate(video);
		}
	}
}


/* Image format conversion */
/* Uncompressed image scaling / conversion */
/* Constraint : npixels must be even, first pixel must be even (yu) */

static inline void convert_line_yuyv (u8 *src, u8 *dest, int npixels)
{
	memcpy(dest, src, npixels << 1);
}


static inline void convert_line_yuv422p(u8 *src, u8 *dest, u8 *uplane, u8 *vplane, int npixels)
{
	int i;
	for (i = (npixels >> 1); i; i--) {
		*dest++ = *src++;
		*uplane++ = *src++;
		*dest++ = *src++;
		*vplane++ = *src++;
	}
}


static inline void convert_line_yuv411p(u8 *src, u8 *dest, u8 *uplane, u8 *vplane, int npixels)
{
	int i;
	for (i = (npixels >> 2); i; i--) {
		*dest++ = *src++;
		*uplane++ = *src++;
		*dest++ = *src++;
		*vplane++ = *src++;
		*dest++ = *src++;
		src++; /* skip u. should this one be skipped? could both be used? (avg) */
		*dest++ = *src++;
		src++; /* skip v */
	}
}


static inline void convert_line_yuv420p(u8 *src, u8 *dest, u8 *uplane, u8 *vplane, int npixels, int do_uv)
{
	int i;
	if (do_uv) {
		for (i = (npixels >> 1); i; i--) {
			*dest++ = *src++;
			*uplane++ = *src++;
			*dest++ = *src++;
			*vplane++ = *src++;
		}
	} else {
		for (i = (npixels >> 1); i; i--) {
			*dest++ = *src++;
			src++;
			*dest++ = *src++;
			src++;
		}
	}
}


/*
 y     = [16;235]
 cb,cr = [16;240]

 Y  = (y-16)*1.168949772 -> [0;256] (fp16 76608.29)
 Cb = (cb-128)/0.875	-> [-128;+128]
 Cr = (cr-128)/0.875	-> [-128;+128]

 R = Y                + 1.402   * Cr	(105007)
 G = Y - 0.34414 * Cb - 0.71414 * Cr	(25775,53488)
 B = Y + 1.772   * Cb			(132720)

 See also: http://www.thedirks.org/v4l2/v4l2fmt.htm
*/

#define YUV_2_RGB(y,u,v,r,g,b,rs,rm,gs,gm,bs,bm) \
	y = (y - 16) * 76608;				\
	u -= 128;					\
	v -= 128;					\
	r = (y + v * 105007) >> (rs);			\
	if (r < 0) r = 0; else if (r > (rm)) r = (rm);	\
	g = (y - u * 25775 - v * 53488) >> (gs);	\
	if (g < 0) g = 0; else if (g > (gm)) g = (gm);	\
	b = (y + u * 132720) >> (bs);			\
	if (b < 0) b = 0; else if (b > (bm)) b = (bm);

/*
 following snippet found on the web:
 

> There's been some confusion how the RGB data is suspposed to come. All
> of the V4L programs do it differently. This problem you're seeing with
> the colors is the RGB data being interpreted as BGR. I'm still trying to
> figure out exactly how it's supposed to be interpreted.

I discussed this on the v4l list. The 'correct' way is:

- for read: rgb is just rgb
- for mmap: rgb24/32 is actually bgr24/32, rgb15/16 is rgb

There is no reasoning for the above except history and misunderstanding by
the programmers. I suggest you stick to this for the simple reason that
one would break either all mmap-based applications or all read-based
application to do it right. For this very reason there is no intention
to change this from the v4l-guys side. For v4l2 things are different. You
have the correct palettes there (PALETTE_RGB and PALETTE_BGR).

*/


static inline void yuv_rgb24(int y, int u, int v, u8 *dest)
{
	int r, g, b;
	YUV_2_RGB(y,u,v,r,g,b, 16,255,16,255,16,255);
	/* store as BGR */
	dest[0] = b;
	dest[1] = g;
	dest[2] = r;
}


static inline void yuv_rgb32(int y, int u, int v, u8 *dest)
{
	int r, g, b;
	YUV_2_RGB(y,u,v,r,g,b, 16,255,16,255,16,255);
	/* store as BGR */
	dest[0] = b;
	dest[1] = g;
	dest[2] = r;
	dest[3] = 0x00;
}


static inline void yuv_rgb16(int y, int u, int v, u16 *dest)
{
	int r, g, b;
	YUV_2_RGB(y,u,v,r,g,b, 16+3,31,16+2,63,16+3,31);
	/* store as RGB */
	*dest = (r<<11) + (g<<5) + (b);
}


static inline void yuv_rgb15(int y, int u, int v, u16 *dest)
{
	int r, g, b;
	YUV_2_RGB(y,u,v,r,g,b, 16+3,31,16+3,31,16+3,31);
	/* store as RGB */
	*dest = (r<<10) + (g<<5) + (b);
}


static inline u8 yuv_grey8(int y, int u, int v)
{
	y = (y - 16) * 76608;
	return (y>>16);
}

	
u32 convert_line_grey8(u8 *src, u8 *dest, int npixels)
{
	int y, cb, cr;
	int i;
	u32 val;

	for (i = (npixels >> 1); i; i--) {
		val =  *((u32*)src);
		y = val & 0xff;
		cb = (val >> 8) & 0xff;
		cr = (val >> 24) & 0xff;
		*(u8*)dest = yuv_grey8(y, cb, cr);
		y = (val >> 16) & 0xff;
		*(u8*)(dest+1) = yuv_grey8(y, cb, cr);
		dest += 2;
		src += 4;
	}
	return npixels;
}


u32 convert_line_rgb15(u8 *src, u8 *dest, int npixels)
{
	int y, cb, cr;
	int i;
	u32 val;

	for (i = (npixels >> 1); i; i--) {
		val =  *((u32*)src);
		y = val & 0xff;
		cb = (val >> 8) & 0xff;
		cr = (val >> 24) & 0xff;
		yuv_rgb15(y, cb, cr, (u16*) dest);
		y = (val >> 16) & 0xff;
		yuv_rgb15(y, cb, cr, (u16*) (dest + 2));
		dest += 4;
		src += 4;
	}
	return npixels << 1;
}


u32 convert_line_rgb16(u8 *src, u8 *dest, int npixels)
{
	int y, cb, cr;
	int i;
	u32 val;

	for (i = (npixels >> 1); i; i--) {
		val =  *((u32*)src);
		y = val & 0xff;
		cb = (val >> 8) & 0xff;
		cr = (val >> 24) & 0xff;
		yuv_rgb16(y, cb, cr, (u16*) dest);
		y = (val >> 16) & 0xff;
		yuv_rgb16(y, cb, cr, (u16*) (dest + 2));
		dest += 4;
		src += 4;
	}
	return npixels << 1;
}


u32 convert_line_rgb24(u8 *src, u8 *dest, int npixels)
{
	int y, cb, cr;
	int i;
	u32 val;

	for (i = (npixels >> 1); i; i--) {
		val =  *((u32*)src);
		y = val & 0xff;
		cb = (val >> 8) & 0xff;
		cr = (val >> 24) & 0xff;
		yuv_rgb24(y, cb, cr, dest);
		y = (val >> 16) & 0xff;
		yuv_rgb24(y, cb, cr, dest + 3);
		dest += 6;
		src += 4;
	}
	return npixels * 3;
}


u32 convert_line_rgb32(u8 *src, u8 *dest, int npixels)
{
	int y, cb, cr;
	int i;
	u32 val;

	for (i = (npixels >> 1); i; i--) {
		val =  *((u32*)src);
		y = val & 0xff;
		cb = (val >> 8) & 0xff;
		cr = (val >> 24) & 0xff;
		yuv_rgb32(y, cb, cr, dest);
		y = (val >> 16) & 0xff;
		yuv_rgb32(y, cb, cr, dest + 4);
		dest += 8;
		src += 4;
	}
	return npixels << 2;
}


#define YUV422P_BLACKNESS(n) \
			memset(dest, 0x10, n); \
			memset(uplane, 0x80, n/2); \
			memset(vplane, 0x80, n/2); \
			dest += n; \
			uplane += n/2; \
			vplane += n/2;

#define YUV411P_BLACKNESS(n) \
			memset(dest, 0x10, n); \
			memset(uplane, 0x80, n/4); \
			memset(vplane, 0x80, n/4); \
			dest += n; \
			uplane += n/4; \
			vplane += n/4;

#define YUV420P_BLACKNESS(n,uv) \
			memset(dest, 0x10, n); \
			dest += n; \
			if (uv) { \
				memset(uplane, 0x80, n/2); \
				memset(vplane, 0x80, n/2); \
				uplane += n/2; \
				vplane += n/2; \
			}
			
	
/* For the moment there is a little problem :
the requested width and height can be more than the
maximum allowed. So we have the width and height demanded
by the application in "width" and "height",
and we have the actual geometry of the buffers in "geom" */

void convert_img(u8 *src_even, u8 *src_odd, u32 src_pitch,
	int width, int height, struct buf_data *geom,
	u8 *buffer, int size)
{
	u8 *src;
	u8 *dest;
	u8 *uplane = NULL;
	u8 *vplane = NULL;
	int y = 0, srcline = 0, bytes_per_pixel;
	int x_origin, drop_count;
	int pixel_fill_1;
	int pixel_fill_2;
	int pixel_fill;
	int i;
	int width2, do_uv;

//dprintk( "ci w %i h %i s %i, %p %p %x\n", width, height, size, src_even, src_odd, src_pitch );
//dprintk( "ci gw %i gh %i\n", geom->grab_width, geom->grab_height );
//dprintk( "ci cx %i fmt %i\n", geom->grab_crop_x, geom->grab_format );
//dprintk( "ci sub %i,%i %i,%i\n", geom->subimage_x, geom->subimage_y,  geom->subimage_w, geom->subimage_h );
//dprintk( "ci iw %i ih %i rw %i\n", geom->image_width, geom->image_height, geom->real_width );

	/* Add offset for sub_image cropping */
	x_origin = geom->image_width - geom->real_width - geom->grab_crop_x;
	x_origin += geom->subimage_x * geom->grab_width / geom->subimage_w;
	x_origin &= (~1); /* x_origin must be even */
	src_even += x_origin * 2 + src_pitch * (geom->subimage_y / 2);
	src_odd += x_origin * 2 + src_pitch * (geom->subimage_y / 2);

	/* Scaling factor is 1.14 fixed point */
	switch(geom->grab_format) {
	case VIDEO_PALETTE_RGB555 : bytes_per_pixel = 2; break;
	case VIDEO_PALETTE_RGB565 : bytes_per_pixel = 2; break;
	case VIDEO_PALETTE_RGB24  : bytes_per_pixel = 3; break;
	case VIDEO_PALETTE_RGB32  : bytes_per_pixel = 4; break;
	case VIDEO_PALETTE_YUYV   : bytes_per_pixel = 2; break;
	case VIDEO_PALETTE_YUV422 : bytes_per_pixel = 2; break;
	case VIDEO_PALETTE_GREY   : bytes_per_pixel = 1; break;
	
	case VIDEO_PALETTE_YUV422P:
		bytes_per_pixel = 1;
		uplane = buffer + width*height;
		vplane = uplane + width*height/2;
		break;
	case VIDEO_PALETTE_YUV411P:
	case VIDEO_PALETTE_YUV420P:
		bytes_per_pixel = 1;
		uplane = buffer + width*height;
		vplane = uplane + width*height/4;
		break;
		
	default:
		printk("unknown video format\n");
		return;
	}	
	dest = buffer;

	if (width > geom->real_width) {
		pixel_fill_1 = width - geom->real_width;
		pixel_fill_2 = pixel_fill_1/2;
		pixel_fill_1 -= pixel_fill_2;
		width = geom->real_width;
	} else {
		pixel_fill_1 = 0;
		pixel_fill_2 = 0;
	}

	if (geom->grab_format != VIDEO_PALETTE_YUYV && 
		geom->grab_format != VIDEO_PALETTE_YUV422) {
	}

	if (geom->subimage_h < height) {
		pixel_fill = width + pixel_fill_1 + pixel_fill_2;
		pixel_fill = ((height - geom->subimage_h)/2) * 
				pixel_fill * bytes_per_pixel;

		switch (geom->grab_format) {
		case VIDEO_PALETTE_RGB555:
		case VIDEO_PALETTE_RGB565:
		case VIDEO_PALETTE_RGB24:
		case VIDEO_PALETTE_RGB32:
		case VIDEO_PALETTE_GREY:
			memset(dest, 0, pixel_fill);
			dest += pixel_fill;
			break;

		case VIDEO_PALETTE_YUYV:
		case VIDEO_PALETTE_YUV422:
			pixel_fill = pixel_fill / 2;
			for(i=0; i < pixel_fill; dest += 2, i++) {
				*(short *)dest = 0x0080;
			}
			break;
			
		case VIDEO_PALETTE_YUV422P:
			YUV422P_BLACKNESS(pixel_fill);
			break;
			
		case VIDEO_PALETTE_YUV411P:
			YUV411P_BLACKNESS(pixel_fill);
			break;

		case VIDEO_PALETTE_YUV420P:
			/* will only work with even lines */
			if ( ((height - geom->subimage_h)/2) & 1) {
				printk( "unsupported height in yuv420p conversion\n" );
				return;
			}
			YUV411P_BLACKNESS(pixel_fill);
			break;
		}
	}

	drop_count = 0;
	do {
		if ((srcline & 1) && (src_odd)) {
			src = src_odd;
			src_odd += src_pitch;
		} else {
			src = src_even;
			src_even += src_pitch;
		}
		srcline++;
		/* Shall we drop this line ? */
		drop_count -= geom->grab_height;
		if (drop_count >= 0)
			continue;
		drop_count += geom->subimage_h;

		switch (geom->grab_format) {

		case VIDEO_PALETTE_RGB565:
			if (pixel_fill_1) {
				memset(dest, 0, pixel_fill_1 * 2);
				dest += pixel_fill_1 << 1;
			}

			convert_line_rgb16(src, dest, width);
			dest += width << 1;

			if (pixel_fill_2) {
				memset(dest, 0, pixel_fill_2 * 2);
				dest += pixel_fill_2 << 1;
			}

			break;

		case VIDEO_PALETTE_RGB555:
			if (pixel_fill_1) {
				memset(dest, 0, pixel_fill_1 * 2);
				dest += pixel_fill_1 << 1;
			}

			convert_line_rgb15(src, dest, width);
			dest += width << 1;

			if (pixel_fill_2) {
				memset(dest, 0, pixel_fill_2 * 2);
				dest += pixel_fill_2 << 1;
			}

			break;

		case VIDEO_PALETTE_RGB24:
			if (pixel_fill_1) {
				memset(dest, 0, pixel_fill_1 * 3);
				dest += pixel_fill_1 *3;
			}

			convert_line_rgb24(src, dest, width);
			dest += width * 3;

			if (pixel_fill_2) {
				memset(dest, 0, pixel_fill_2 * 3);
				dest += pixel_fill_2 *3;
			}

			break;

		case VIDEO_PALETTE_RGB32:
			if (pixel_fill_1) {
				memset(dest, 0, pixel_fill_1 * 4);
				dest += pixel_fill_1 * 4;
			}

			convert_line_rgb32(src, dest, width);
			dest += width << 2;

			if (pixel_fill_2) {
				memset(dest, 0, pixel_fill_2 * 4);
				dest += pixel_fill_2 * 4;
			}

			break;

		case VIDEO_PALETTE_GREY:
			if (pixel_fill_1) {
				memset(dest, 0, pixel_fill_1);
				dest += pixel_fill_1;
			}

			convert_line_grey8(src, dest, width);
			dest += width;

			if (pixel_fill_2) {
				memset(dest, 0, pixel_fill_2);
				dest += pixel_fill_2;
			}

			break;


		case VIDEO_PALETTE_YUYV:
		case VIDEO_PALETTE_YUV422:
			if (pixel_fill_1) {
				for(i=0; i < pixel_fill_1; dest += 2, i++) {
					*(short *)dest = 0x0080;
				}
			}

			convert_line_yuyv(src, dest, width);
			dest += width << 1;

			if (pixel_fill_2) {
				for(i=0; i < pixel_fill_2; dest += 2, i++) {
					*(short *)dest = 0x0080;
				}
			}
			break;
			
		case VIDEO_PALETTE_YUV422P:
			pixel_fill = pixel_fill_1 & -2;
			if (pixel_fill) {
				YUV422P_BLACKNESS(pixel_fill);
			}
			
			width2 = width & -2;
			convert_line_yuv422p(src, dest, uplane, vplane, width2);
			dest += width2;
			uplane += width2/2;
			vplane += width2/2;
			
			pixel_fill = (pixel_fill_1-pixel_fill + width-width2 + pixel_fill_2);
			if (pixel_fill & 1) {
				printk( "unsupported width in yuv422p conversion\n" );
				return;
			}
			if (pixel_fill) {
				YUV422P_BLACKNESS(pixel_fill);
			}
			break;
			
		case VIDEO_PALETTE_YUV411P:
			pixel_fill = pixel_fill_1 & -4;
			if (pixel_fill) {
				YUV411P_BLACKNESS(pixel_fill);
			}
			
			width2 = width & -4;
			convert_line_yuv411p(src, dest, uplane, vplane, width2);
			dest += width2;
			uplane += width2/4;
			vplane += width2/4;
			
			pixel_fill = (pixel_fill_1-pixel_fill + width-width2 + pixel_fill_2);
			if (pixel_fill & 3) {
				printk( "unsupported width in yuv411p conversion\n" );
				return;
			}
			if (pixel_fill) {
				YUV411P_BLACKNESS(pixel_fill);
			}
			break;
			
		case VIDEO_PALETTE_YUV420P:
			do_uv = (y&1)==0;
			pixel_fill = pixel_fill_1 & -2;
			if (pixel_fill) {
				YUV420P_BLACKNESS(pixel_fill,do_uv);
			}

			width2 = width & -2;
			convert_line_yuv420p(src, dest, uplane, vplane, width2, do_uv);
			dest += width2;
			if (do_uv) {
				uplane += width2/2;
				vplane += width2/2;
			}
		
			pixel_fill = (pixel_fill_1-pixel_fill + width-width2 + pixel_fill_2);
			if (pixel_fill & 1) {
				printk( "unsupported width in yuv420p conversion\n" );
				return;
			}
			if (pixel_fill) {
				YUV420P_BLACKNESS(pixel_fill,do_uv);
			}
			break;
		}
		y++;
	} while (y < height && srcline < geom->subimage_h);

	if (srcline <= height) {
		pixel_fill = width + pixel_fill_1 + pixel_fill_2;
		pixel_fill = ((height - geom->subimage_h) -
			      ((height - geom->subimage_h)/2));

		switch (geom->grab_format) {
		case VIDEO_PALETTE_RGB555:
		case VIDEO_PALETTE_RGB565:
		case VIDEO_PALETTE_RGB24:
		case VIDEO_PALETTE_RGB32:
		case VIDEO_PALETTE_GREY:
			memset(dest, 0, pixel_fill * bytes_per_pixel);
			break;

		case VIDEO_PALETTE_YUYV:
		case VIDEO_PALETTE_YUV422:
			for(i=0; i < pixel_fill; dest += 2, i++) {
				*(short *)dest = 0x0080;
			}
			break;

		case VIDEO_PALETTE_YUV422P:
			YUV422P_BLACKNESS(pixel_fill);
			break;
			
		case VIDEO_PALETTE_YUV411P:
			YUV411P_BLACKNESS(pixel_fill);
			break;
		
		case VIDEO_PALETTE_YUV420P:
			/*fixme: need even count of vert lines */
			YUV411P_BLACKNESS(pixel_fill);
			break;
		}
	}
}

/****************************************************************************
* interrupt processing
****************************************************************************/
void 
mgagrab_vsync(struct mga_dev_s *mga)
{
	int	field_buffer_match;
	int	i;
	int	program_window;
	struct	mga_grab_s	*video;
	video = mga->grab;


	if (test_and_set_bit(0, &video->vsync_active)) {
		printk("mgavideo: vsync already active\n");
		return;
	}

	video->interrupt_count++;


	/* Update the preview window */
        if (video->preview_active && video->active_buffer >= 0)
		mgabes_quick_set_frame(mga->bes, video->active_buffer,
			video->vin_addr[video->active_buffer]);

        spin_lock(&video->lock);

        /* Update data for the buffer which has been grabbed */
        if (video->active_buffer >= 0) {
		video->buffer_flags[video->active_buffer] &= ~(BUFFLAG_SERVERLOCK + BUFFLAG_FIELDTYPE);
		video->buffer_flags[video->active_buffer] |= BUFFLAG_GRAB_OK;
		if ((mga->status & VS_FIELD) == VS_BOTFIELD)
			video->buffer_flags[video->active_buffer] |= BUFFLAG_FIELDTYPE;
		video->grabbed++;
		video->buffer_stamp[video->active_buffer] = video->grabbed;
	}
	
        /* Test if the fields come in the correct buffers */
	if (((video->active_buffer & 1) == 0 && (mga->status & VS_FIELD) == VS_TOPFIELD)
			|| ((video->active_buffer & 1) != 0 && (mga->status & VS_FIELD) != VS_TOPFIELD))
		field_buffer_match = 1;
	else
		field_buffer_match = 0;

        video->active_buffer = video->next_active_buffer;
        video->next_active_buffer = -1;

        /* Find which buffer to use next. Rules depend on the preview mode. */
	if (1) {
		/* Default choice, use the circular order. */
		video->next_active_buffer = (video->active_buffer + 1) & 3;
		/* If the field-type doesn't match with the buffer number "oddity" then take another */
		/* one. Do the test for first fields only (the next field it being grabbed and is */
		/* also wrong, so don't exchange next time). */
		if ((mga->status & VS_FIELD) == VS_TOPFIELD)
			if (! field_buffer_match)
				video->next_active_buffer = (video->active_buffer + 2) & 3;
		/* If this one is locked, preview window sacrificed : use the normal criteria. */
		if (video->buffer_flags[video->next_active_buffer] & (BUFFLAG_CLIENTLOCK+BUFFLAG_SERVERLOCK))
			video->next_active_buffer = -1;
	}

	if (video->next_active_buffer == -1) {
		/* Best grabbing politics applied (keep wanted buffers as long as possible) : */
		/* First choice : oldest unwanted buffer */
		/* Second choice : oldest unlocked buffer */
		int unwanted_found = 0;
		u32 oldeststamp = 0xffffffff;
		video->next_active_buffer = 0;
		for (i = 0; i < 4; i++) {
			if ((video->buffer_flags[i] & (BUFFLAG_CLIENTLOCK | BUFFLAG_SERVERLOCK)) == 0) {
				int field_type, wanted;
				field_type =
					((video->buffer_flags[i] & BUFFLAG_FIELDTYPE)?FIELD_BOTTOM:FIELD_TOP);
				wanted = field_type & mga->v4lbuf.fields_wanted;
				if (wanted) {
					if (!unwanted_found && oldeststamp > video->buffer_stamp[i]) {
						oldeststamp = video->buffer_stamp[i];
						video->next_active_buffer = i;
					}
				} else if (!unwanted_found || oldeststamp > video->buffer_stamp[i]) {
					oldeststamp = video->buffer_stamp[i];
					video->next_active_buffer = i;
					unwanted_found = 1;
				}
			}
		}
	}
	/* Lock the chosen buffer */
	video->buffer_flags[video->next_active_buffer] |= BUFFLAG_SERVERLOCK;
	video->buffer_flags[video->next_active_buffer] &= ~BUFFLAG_GRAB_OK;

        /* End of critical section */
        spin_unlock(&video->lock);

        /* Program the VIN windows to point to the chosen buffer */
        /* On the g200, we remember which window we programmed last time. */
        /* On the g400, we cannot predict in which window the data will be put */
        /* (it depends on the field type). So we make all windows point to the */
        /* same buffer. Here X1 mode is assumed. */                	
        if (mga->mga_type == MGA_G200 || mga->mga_type == MGA_G100) {
		video->vin_active = (video->vin_active+1) & 1;
	      	mga_write(mga, VINNEXTWIN, video->vin_active);
	       	program_window = (video->vin_active == 0) ? 1 : 2;
	} else {
		program_window = 3;
	}
	if (program_window & 1) {
		mga_write(mga, VINADDR0, video->vin_addr[video->next_active_buffer]);
		mga_write(mga, VBIADDR0, video->vbi_addr[0]);
		mga_write(mga, VINCTL0,  video->vinctl_value);
	}
	if (program_window & 2) {
		mga_write(mga, VINADDR1, video->vin_addr[video->next_active_buffer]);
		mga_write(mga, VBIADDR1, video->vbi_addr[1]);
		mga_write(mga, VINCTL1,  video->vinctl_value);
	}		

	/* Launch the bottom half if needed */
	if (video->grab_needed)
		schedule_work(&video->tqnode_dpc);

	video->vsync_active = 0;
}

/* Converts an image from the framebuffer memory to main memory. */
/* If no complete image was available, returns -1. */
/* If the buffer has been filled with a complete image, returns 0. */

void
mgagrab_grab_needed(struct mga_grab_s *video, int state)
{
	video->grab_needed = state;
	if (state == 0) {
		mgagrab_deactivate(video);
		video->mga->v4lbuf.grab_width = -1;
	} else {
		mgagrab_activate(video);
	}
}


void
mgagrab_set_ongrab(struct mga_grab_s *grab, grab_image_func_t func)
{
	grab->on_grab_image = func;
}


int mgagrab_grab_image(struct mga_dev_s *mga,
	struct video_mmap *geometry, u8 *buffer, int size)
{
	unsigned long flags;
	int buf1, buf2;
	u32 stamp;
	int i;
	u8 *src_even, *src_odd;
	struct mga_grab_s *video;

	video = mga->grab;

	mgagrab_set_grabsize(video, geometry->width, geometry->height, geometry->format);

	if (video->on_grab_image != NULL) {
		i = video->on_grab_image(mga, geometry, buffer, size);
		if (i) return i;
	}
	
	spin_lock_irqsave(&video->lock, flags);

	dprintk ("%x %6d %x %6d %x %6d %x %6d, hs %6d\n",
		video->buffer_flags[0], video->buffer_stamp[0],
		video->buffer_flags[1], video->buffer_stamp[1],
		video->buffer_flags[2], video->buffer_stamp[2],
		video->buffer_flags[3], video->buffer_stamp[3],
		mga->v4lbuf.scale_change_stamp
	);
//	dprintk( "fields wanted %i\n", mga->v4lbuf.fields_wanted );
	
	/* Current politics : always choose the most recent buffers. */
	stamp = 0;
	buf1 = buf2 = -1;

	if (mga->v4lbuf.fields_wanted & FIELD_BOTTOM) {
		/* Find the most recent bottom field */
		for (i = 0; i < 4; i++) {
			if ((video->buffer_flags[i] & BUFFLAG_GRAB_OK)
				&& (video->buffer_flags[i] & BUFFLAG_FIELDTYPE)
				&& (video->buffer_stamp[i] > stamp)
				&& (video->buffer_stamp[i] > mga->v4lbuf.scale_change_stamp + 2)) {
				buf1 = i;
				stamp = video->buffer_stamp[i];
			}
		}
	}
	if (mga->v4lbuf.fields_wanted & FIELD_TOP) {
		if (buf1 >= 0) {
			/* Both fields wanted : find the other one */
			buf2 = buf1;
			buf1 = -1;
			for (i = 0; i < 4; i++) {
				if ((video->buffer_flags[i] & BUFFLAG_GRAB_OK)
					&& ! (video->buffer_flags[i] & BUFFLAG_FIELDTYPE)
					&& (video->buffer_stamp[i] == stamp-1)) {
					buf1 = i;
				}
			}
		} else if ((mga->v4lbuf.fields_wanted & FIELD_BOTTOM) == 0) {
			/* Find the most recent top field */			
			for (i = 0; i < 4; i++) {
				if ((video->buffer_flags[i] & BUFFLAG_GRAB_OK)
					&& ! (video->buffer_flags[i] & BUFFLAG_FIELDTYPE)
					&& (video->buffer_stamp[i] > stamp)
					&& (video->buffer_stamp[i] > mga->v4lbuf.scale_change_stamp + 2)) {
					buf1 = i;
					stamp = video->buffer_stamp[i];
				}
			}
		}
	}
	
	/* Buffers founds ? then lock them, an destroy old buffers */
	/* (otherwise, we could catch an older buffer in the next grab) */
	if (buf1 >= 0) {
		video->buffer_flags[buf1] |= BUFFLAG_CLIENTLOCK;
		if (buf2 >= 0) video->buffer_flags[buf2] |= BUFFLAG_CLIENTLOCK;
		for (i = 0; i < 4; i++) {
			if ((video->buffer_flags[i] & BUFFLAG_GRAB_OK)
				&& (video->buffer_stamp[i] <= stamp)) {
				video->buffer_flags[i] &= ~BUFFLAG_GRAB_OK;
			}
		}
	}
	spin_unlock_irqrestore(&video->lock,flags);
	
	dprintk ("grab_image buf1=%d buf2=%d\n", buf1, buf2);
	
	if (buf1 < 0) {
		dprintk( "mgagrab_grab_image: no bufs\n" );
		return -1;
	}

	src_even = mga->fb + video->vin_addr[buf1];
	if (buf2 >= 0)
		src_odd = mga->fb + video->vin_addr[buf2];
	else
		src_odd = src_even;

	convert_img (src_even, src_odd, _VINWID * 4,
		geometry->width, geometry->height, &(mga->v4lbuf), buffer, size);

	/* Unlock the frame-buffer */
	/* No need to lock, the buffer is CLIENTLOCKed */
	video->buffer_flags[buf1] &= ~BUFFLAG_CLIENTLOCK;
	if (buf2 >= 0) video->buffer_flags[buf2] &= ~BUFFLAG_CLIENTLOCK;
	
	return 0;
}

/* register a routine to be called for each frame */
void mgagrab_register_bh(struct mga_grab_s *video, void(*routine)(void*), void* data)
{
	INIT_WORK(&video->tqnode_dpc, routine, data);
}


/* V4L buffers routines */
/* Design issues :
- ks0127 sends a complete image (720 x active_lines before scaling).
- it scales horizontaly to match the demanded v4l grab size ( not the preview size )
- the bes is programmed to display the demanded portion of the image
  (user can independently specify a preview subimage and a grab subimage)
- v4lgrabbing uses hardware horizontal scaling by the ks0127 and
  line dropping for vertical scaling. It can only down-scale, not upscale.
*/

void mgagrab_set_grabsize (struct mga_grab_s *video,
	int width, int height, int format)
{
	struct buf_data *buf;
	struct mga_dev_s *mga;
	int	scale_width;

	mga = video->mga;
	buf = &(mga->v4lbuf);

	scale_width = width;
	
	/* We can only downscale */
	if (width < 2) scale_width = 2;
	if (width > buf->subimage_w) scale_width = buf->subimage_w;

	if (width == buf->grab_width && height == buf->grab_height) return ;

	buf->grab_width = width;
	buf->grab_height = height;
	buf->grab_format = format;

	if (buf->grab_height > buf->subimage_h / 2)
		buf->fields_wanted = FIELD_TOP + FIELD_BOTTOM;
	else
		buf->fields_wanted = FIELD_TOP;
}


void mgagrab_set_grabsubimage (struct mga_grab_s *video,
	int x, int y, int w, int h)
{
	struct buf_data *buf;
	struct mga_dev_s *mga;


	mga = video->mga;
	buf = &(mga->v4lbuf);
	
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (w < 2) w = 2;
	if (h < 2) h = 2;
	if (w > buf->image_width) w = buf->image_width;
	if (h > buf->image_height) h = buf->image_height;
	if (x + w > buf->image_width) x = buf->image_width - w;
	if (y + h > buf->image_height) y = buf->image_height - h;
	
	buf->subimage_x = x;
	buf->subimage_y = y;
	buf->subimage_w = w;
	buf->subimage_h = h;

	/* Recompute scaling and bes parameters */
	mgagrab_set_grabsize (video, buf->grab_width, buf->grab_height,
		buf->grab_format);
}


/****************************************************************************
* initialization
****************************************************************************/

void mgagrab_attach(struct mga_dev_s *mga)
{
	struct mga_grab_s *video;
	u32		vbi_align;
	u32		vin_align;
	u32		field_bytes;
	u32		vinpitch;
	size_t		b;

	mga->grab = video = kmalloc(sizeof(*video), GFP_KERNEL);
	if (video == NULL)
		return;
	memset(video, 0, sizeof(*video));

	video->mga = mga;
	video->grabber_active = 0;
	video->on_grab_image = NULL;
	
	if (mga->mga_type==MGA_G200 || mga->mga_type==MGA_G100)
		video->vinctl = 0x0101ba00;  /* RP check, vbi-task, default vbi's */
	else
		video->vinctl = 0x01010030; /* RP check, yuy2, video-ints on both fields, 1xMode */
	
	spin_lock_init(&video->lock);
	
	vinpitch = _VINWID;

	if ((mga->mga_type == MGA_G200) || (mga->mga_type == MGA_G100)) {
		video->vinctl_value = (((_VINWID * 2) & (~7)) << 1) | VIN_ON;
		vbi_align = 512;
		vin_align = 8;
		field_bytes = _VIN_WINDOW_G200;
	} else { /* G400 or better */
        	video->vinctl_value = ((_VINWID * 2) & (~7))  | VIN_ON;
		vbi_align = 16;
		vin_align = 16;
		field_bytes = _VIN_WINDOW;
	}

	video->vin_size = b = (field_bytes * 2);
	video->vin_addr[0] = mga_alloc_buffer(mga, b, vin_align);
	video->vin_addr[1] = video->vin_addr[0] + vinpitch * 2;

	video->vin_addr[2] = mga_alloc_buffer(mga, b, vin_align);
	video->vin_addr[3] = video->vin_addr[2] + vinpitch * 2;

	video->vbi_size = b = _VBI_WINDOW;
	video->vbi_addr[0] =
	video->vbi_addr[1] =
	video->vbi_addr[2] =
	video->vbi_addr[3] = mga_alloc_buffer(mga, b, vbi_align );

	mga_create_proc_entry(video->mga, "grab", mgagrab_read_proc, 
		mgagrab_write_proc, video);

	// XXX MOD_INC_USE_COUNT;
}


void mgagrab_deattach(struct mga_grab_s *video)
{
	mgagrab_disable_grabber( video );

	mga_remove_proc_entry(video->mga, "grab");
	
	mga_free_buffer( video->mga, video->vbi_size );
	mga_free_buffer( video->mga, video->vin_size );
	mga_free_buffer( video->mga, video->vin_size );

	video->mga->grab = NULL;
	kfree(video);

	// XXX MOD_DEC_USE_COUNT;
}



int mgagrab_read_proc(char *page, char **start, off_t off,
                  int count, int *eof, void *data)
{
	struct mga_grab_s *video = (struct mga_grab_s*)data;
	char *s = page;
	int len;

	s = s + sprintf(s, "Vsync interrupt count: %d\n", video->interrupt_count);
	
	len = s - page;

	return len;
}


int mgagrab_write_proc(struct file *file, const char *buf,
                          unsigned long count, void *data)
{
	struct mga_grab_s *video = (struct mga_grab_s *)data;

	(void)video;

	return count;		
}


int
mgagrab_get_stamp(struct mga_grab_s *video)
{
	return video->grabbed;
}


EXPORT_SYMBOL(mgagrab_attach);
EXPORT_SYMBOL(mgagrab_deattach);
EXPORT_SYMBOL(mgagrab_init_videoformat);
EXPORT_SYMBOL(mgagrab_preview_enable);
EXPORT_SYMBOL(mgagrab_grab_needed);
EXPORT_SYMBOL(mgagrab_grab_image);
EXPORT_SYMBOL(mgagrab_set_grabsize);
EXPORT_SYMBOL(mgagrab_set_grabsubimage);
EXPORT_SYMBOL(mgagrab_register_bh);
EXPORT_SYMBOL(mgagrab_activate);
EXPORT_SYMBOL(mgagrab_deactivate);
EXPORT_SYMBOL(mgagrab_get_stamp);
EXPORT_SYMBOL(mgagrab_set_ongrab);

