/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200/G400 and Rainbow Runner-G series
 *
 * This module is an interface to the G100,G200 and G400 video extension
 * registers.  
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 *****************************************************************************
 *
 * Modified and extended by
 *	Mike Bernson <mike@mlb.org>
 *	Niels de Carpentier
 *	Gerard v.d. Horst
 *	Leon van Stuivenberg <leonvs@iae.nl>
 *	Gernot Ziegler <gz@lysator.liu.se>
 *
 * Version History:
 * V1.0 Ryan Drake         Initial version by Ryan Drake
 * V1.1 Gerard v.d. Horst  Activated the backend scaler
 * V1.2 Niels de Carpentier Added G400 support
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>

#include <linux/slab.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/videodev.h>

#include "aiee.h"

#include "mgacore.h"
#include "mgavideo.h"
#include "mgagrab.h"


#include "media/tuner.h"
#include "media/audiochip.h"
#include "ks0127.h"

#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */

static int default_norm  = KS_STD_NTSC;


#if LINUX_VERSION_CODE >= 0x020100
module_param(debug, int, 0644);
module_param(default_norm, int, 0644);
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif


struct mga_video_input_s {
	int	norm;
        int	brightness;
	int	hue;
	int	contrast;
	int	color;
	unsigned long	freq;
	struct mga_video_area_s crop[2]; /* 0 ntsc 1 pal */
};

struct mga_video_s {
	struct	mga_dev_s	*mga;

	spinlock_t lock;

	int	switch_active;

	WAIT_QUEUE	sleepq;

	int	norm;		/* current video format for ks0127 */
	int	source;		/* current video source */

	unsigned long	tuner_freq;	/* current video frequence */
	int	tuner_norm;	/* current video format */

	struct  mga_video_input_s input[4];

	int	interrupt_count;
	int	interrupt_count_last;
	
	u32	ks_scaling;
};


#define VALID_INPUT(i) (((i)>=0) && ((i)<sizeof(video->input)/sizeof(video->input[0])))



int mgavideo_read_proc(char *page, char **start, off_t off,
                  int count, int *eof, void *data);

int mgavideo_write_proc(struct file *file, const char *buf,
                          unsigned long count, void *data);



static int
get_ks_format(struct mga_video_s *video, int input, struct mgabus_video_input_s *params)
{
	struct mga_video_area_s *crop;
	
	memset(params, 0, sizeof(*params));

	params->norm = video->input[ input ].norm;
	params->brightness = video->input[ input ].brightness;
	params->hue = video->input[ input ].hue;
	params->contrast = video->input[ input ].contrast;
	params->color = video->input[ input ].color;
	
	switch(video->input[ input ].norm) {

	case KS_STD_NTSC:
		crop = &video->input[input].crop[0];
		break;

	case KS_STD_PAL:
		crop = &video->input[input].crop[1];
		break;

	case KS_STD_SECAM:
		crop = &video->input[input].crop[1];
		break;

	default:
		crop = &video->input[input].crop[0];
		printk("mgavideo_init_videoformat: unknown format %x\n", video->input[ input ].norm);
	}
	params->crop = *crop;
	params->active = *crop; //fixme..
	
	if (video->mga->mga_type == MGA_G200 || video->mga->mga_type == MGA_G100) {
	
		params->vblcnt0 = 0;
		params->vblcnt1 = 0;
	
		switch(video->input[ input ].norm) {
		case KS_STD_NTSC:
		default:
			params->vbicnt0 = 0x0b;
			params->vbicnt1 = 0x0a;
			break;
		case KS_STD_PAL:
			params->vbicnt0 = 0x0f; /* limited by reg. resolution */
			params->vbicnt1 = 0x0e;
			break;
		case KS_STD_SECAM: //?
			params->vbicnt0 = 0x0f; /* limited by reg. resolution */
			params->vbicnt1 = 0x0e;
			break;
		}
		
		if (video->source == KS_INPUT_YUV656) {
			/* mjpeg playback with valign=1, make adjustment */
			params->vbicnt1 += 1;
		}
	} else if (video->mga->mga_type == MGA_G400) {
	
		if (video->mga->options & MGAVIDEO_HAS_MARVEL) {
			/* fixme: assumed: marvel with ks0127B */
			
			switch(video->input[ input ].norm) {
			case KS_STD_NTSC:
			default:
				/* might need to shave off a few lines still.. */
				params->vbicnt0 = 0x00;
				params->vblcnt0 = 0x0f;
				params->vbicnt1 = 0x00;
				params->vblcnt1 = 0x04;
				break;
			case KS_STD_PAL:
				/* there appears to be a big difference between
					odd and even field, wrt blanking */
				params->vbicnt0 = 0x00;
				params->vblcnt0 = 0x0f;
				params->vbicnt1 = 0x00;
				params->vblcnt1 = 0x04;
				break;
			case KS_STD_SECAM:
				params->vbicnt0 = 0x00;
				params->vblcnt0 = 0x0f;
				params->vbicnt1 = 0x00;
				params->vblcnt1 = 0x04;
				break;
			}
			if (video->source == KS_INPUT_YUV656) {
				/* mjpeg playback with valign=1, make adjustment */
				params->vblcnt1 += 1;
			}
		} else {
			/* fixme: assumed: g400/RRG with ks0127 */
			switch(video->input[ input ].norm) {
			case KS_STD_NTSC:
			default:
				/* unchecked */
				params->vbicnt0 = 0x0b;
				params->vblcnt0 = 0x13;
				params->vbicnt1 = 0x0a;
				params->vblcnt1 = 0x12;
				break;
			case KS_STD_PAL:
				/* checked */
				params->vbicnt0 = 0x07;
				params->vblcnt0 = 0x15;
				params->vbicnt1 = 0x06;
				params->vblcnt1 = 0x14;
				break;
			case KS_STD_SECAM:
				/* unchecked */
				params->vbicnt0 = 0x07;
				params->vblcnt0 = 0x15;
				params->vbicnt1 = 0x06;
				params->vblcnt1 = 0x14;
				break;
			}
			if (video->source == KS_INPUT_YUV656) {
				/* mjpeg playback with valign=1, make adjustment */
				// need what?
				params->vbicnt0 = 0x00;
				params->vblcnt0 = 0x00;
				params->vbicnt1 = 0x00;
				params->vblcnt1 = 0x00;
			}
		}
	} else {
		printk("vbi not set for unknown Gx00 type\n" );
		return 1;
	}
	return 0;
}



/* Reset grabbing and preview parameters for a new video format */
static void
mgavideo_init_videoformat(struct mga_video_s *video)
{
	struct buf_data *buf = &(video->mga->v4lbuf);

	mgagrab_init_videoformat( video->mga->grab );

	/* Recompute scaling and bes parameters */
	mgavideo_set_grabsize(video, buf->grab_width, buf->grab_height,
		buf->grab_format);
}


/****************************************************************************
* mgavideo api : public interface to the mga's video subsystem
****************************************************************************/
/* component device control */

void
mgavideo_disable_grabber(struct mga_video_s *video)
{
	mgagrab_deactivate( video->mga->grab );

	mga_disable_irq(video->mga, MGA_IRQ_VINVSYNC, 
		(mga_interrupt_handler_t)mgavideo_vsync);

        if (video->mga->options & MGAVIDEO_HAS_AUDIO) {
		if (video->source == KS_INPUT_TUNER) {
			/*
			__u16 s = AUDIO_MUTE;
			mga_i2c_control(video->mga,
				I2C_DRIVERID_MSP3400, 
				AUDC_SET_INPUT, &s);
//				AUDC_SWITCH_MUTE, 0);
			*/
		}
	}
}


void
mgavideo_enable_grabber(struct mga_video_s *video)
{
	mgagrab_activate( video->mga->grab );
	
	mga_enable_irq(video->mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)mgavideo_vsync);
}


void
mgavideo_audio(struct mga_video_s *video, unsigned cmd, void *data)
{
        if (video->mga->options & MGAVIDEO_HAS_AUDIO) {
		mga_i2c_control(video->mga, I2C_DRIVERID_MSP3400, 
			cmd, data);
	}
}


void
mgavideo_set_active(struct mga_video_s *video, int input, int norm)
{
	int	change;
	int	tuner;
	int	i;
	int	start;
	int	status;
	int	retry;
 	struct mgabus_video_input_s fmt;

	change = 0;
	
	if (!VALID_INPUT(input)) {
		dprintk( "mgavideo_set_active: wrong input %i\n", input );
		return;
	}
	
	if (video->source != input) {
		video->source = input;
		change++;
	}

	if (video->norm != norm) {
		video->norm = norm;
		change++;
	}

	if (video->source == KS_INPUT_TUNER) {
		if (video->tuner_norm != video->input[KS_INPUT_TUNER].norm ||
		    video->tuner_freq != video->input[KS_INPUT_TUNER].freq) {

			change++;
		}
	}
	video->input[input].norm = norm;

	if (change && (video->mga->options & MGAVIDEO_HAS_DECODER)) {
	
		int reset_ok = 0;
	
		video->interrupt_count_last = video->interrupt_count;
		
		start = jiffies;
		for(retry = 0; retry < 5 && video->interrupt_count ==
			video->interrupt_count_last; retry++) {

			mgavideo_disable_grabber(video);
			start = jiffies;

			if (input == KS_INPUT_TUNER && 
			    (video->mga->options & MGAVIDEO_HAS_TUNER)) {
				struct video_channel v;
			    	v.channel = video->input[KS_INPUT_TUNER].freq;
				v.norm = video->input[KS_INPUT_TUNER].norm;
				mga_i2c_control(video->mga,
					I2C_DRIVERID_TUNER, VIDIOCSCHAN, &v);
				mga_i2c_control(video->mga,
					I2C_DRIVERID_TUNER, VIDIOCSFREQ, &(v.channel));

				video->tuner_freq = video->input[KS_INPUT_TUNER].freq;
				video->tuner_norm = video->input[KS_INPUT_TUNER].norm;
			}
		        if (input == KS_INPUT_TUNER &&
			    (video->mga->options & MGAVIDEO_HAS_AUDIO)) {
				struct video_channel vc;
				vc.norm = video->input[input].norm;
				mga_i2c_control(video->mga, I2C_DRIVERID_MSP3400, 
					VIDIOCSCHAN, &vc);
				mga_i2c_control(video->mga, I2C_DRIVERID_MSP3400, 
					VIDIOCSFREQ, 0);
			}

			if (video->mga->options & MGAVIDEO_HAS_DECODER) {
				if ( reset_ok ) {
					/* Reset device and set value back to
					 * good default values.
					 */
					mga_i2c_control(video->mga,
						I2C_DRIVERID_KS0127, 
						KS0127_RESET, 0);
					reset_ok = 0;
				}
				mga_i2c_control(video->mga,
					I2C_DRIVERID_KS0127, 
					KS0127_SET_STANDARD, &video->norm);

				mga_i2c_control(video->mga,
					I2C_DRIVERID_KS0127, 
					KS0127_SET_INPUT, &video->source);

				if (get_ks_format(video, video->source, &fmt) != 0) {
					printk("get_ks_format failed\n");
				}
				else {
					mga_set_vipformat(video->mga, &fmt);
			 	}
				mgavideo_init_videoformat(video);

				if (input != KS_INPUT_YUV656)
					mgavideo_set_color(video, 
						video->input[video->source].brightness,
						video->input[video->source].hue,
						video->input[video->source].contrast,
						video->input[video->source].color);
			}

			video->switch_active = 1;
			mgavideo_enable_grabber(video);

			/* 
			 * this is need to make sure that the ks0127 is running.
			 * It does not already create interrupts after
			 * switch sources.
			 */
			start = jiffies;
			for(i=HZ/2; i; i--) {
				u8 statmask;
				u8 statval;
			
				/* Sleep nicely for +/- tick */
				sleep_on_timeout(&video->sleepq, 1);

				mga_i2c_control(video->mga,
					I2C_DRIVERID_KS0127, 
					KS0127_GET_STATUS, &status);

				/* for tuner, don't use NOVID (see ks spec) */

				statmask = (input==KS_INPUT_TUNER) ? KS_STATUS_COLOR_DET | 0x01 :
					KS_STATUS_COLOR_DET | 0x01 | KS_STATUS_SYNC;
				statval = (input==KS_INPUT_YUV656) ? KS_STATUS_COLOR_DET :
					KS_STATUS_COLOR_DET | 0x01;

				dprintk("jiffies=%ld retry=%d input = %x norm %d switch %x & %x = %x int %d\n",
					jiffies -start, retry,input,norm, status, statmask, status & statmask, video->interrupt_count);

				if ((status & statmask) != statval) {
					continue;
				} else if ((video->interrupt_count > video->interrupt_count_last)
					|| (input==KS_INPUT_YUV656) ) {
					break;
				}
			}
			if (i < 1) {
				/* had a time-out when waiting for good status/interrupt.
					now might be a good time for a reset */
				reset_ok = 1;
			}

			video->switch_active = 0;

			if ( (input == KS_INPUT_YUV656) ||
				((input == KS_INPUT_TUNER) && (video->tuner_freq == 0) ) )
				break; // won't get ints;
		}
		dprintk("mgavideo: %ld to switch setup.\n", 
			start - jiffies);
	}

	/* always set audio input */
	switch(input) {
	case KS_INPUT_TUNER:
		tuner = 1;
		break;
	case KS_INPUT_SVIDEO:
	case KS_INPUT_COMPOSITE:
	case KS_INPUT_YUV656:
	default:
		tuner = 0;
		break;
	}
	mga_set_audioinput(video->mga, tuner);
}


void
mgavideo_get_active(struct mga_video_s *video, int *input, int *norm)
{
	*input = video->source;
	*norm = video->norm;
}


void
mgavideo_set_color(struct mga_video_s *video, int brightness, int hue,
	int saturation, int contrast)
{
	if (!VALID_INPUT(video->source)) {
		dprintk( "mgavideo_set_color: no valid input\n" );
		return;
	}
	video->input[video->source].brightness = brightness;
	video->input[video->source].hue = hue;
	video->input[video->source].contrast = contrast;
	video->input[video->source].color = saturation;

	if (video->mga->options & MGAVIDEO_HAS_DECODER) {
                mga_i2c_control(video->mga,
			I2C_DRIVERID_KS0127, 
			KS0127_SET_BRIGHTNESS, &brightness);

		mga_i2c_control(video->mga,
			I2C_DRIVERID_KS0127, 
			KS0127_SET_HUE, &hue);

		mga_i2c_control(video->mga,
			I2C_DRIVERID_KS0127, 
			KS0127_SET_SATURATION, &saturation);

		mga_i2c_control(video->mga,
			I2C_DRIVERID_KS0127, 
			KS0127_SET_CONTRAST, &contrast);
	}
}


void 
mgavideo_decoder(struct mga_video_s *video, unsigned int cmd, void *arg)
{
        if (video->mga->options & MGAVIDEO_HAS_DECODER)
                mga_i2c_control(video->mga, I2C_DRIVERID_KS0127, 
			cmd, arg);
}


int
mgavideo_input(struct mga_video_s *video)
{
	return video->source;
}


int
mgavideo_get_norm(struct mga_video_s *video, int source)
{
	if (!VALID_INPUT(source))
		return -1;
	return video->input[source].norm;
}


void
mgavideo_set_norm(struct mga_video_s *video, int source, int norm)
{
	if (!VALID_INPUT(source)) {
		dprintk( "mgavideo_set_norm: wrong input %i\n", source );
		return;
	}

	if (video->input[source].norm == norm)
		return;
		
	video->input[source].norm = norm;

	if (source == video->source) {
		mgavideo_set_active(video, source, norm);
	}
}


unsigned long
mgavideo_get_freq(struct mga_video_s *video)
{
	return video->tuner_freq;
}


void
mgavideo_set_freq(struct mga_video_s *video, unsigned long freq)
{
	video->input[KS_INPUT_TUNER].freq = freq;

	if (video->source == KS_INPUT_TUNER && freq != video->tuner_freq) {
		mgavideo_set_active(video, video->source, video->norm);
	}
}


/****************************************************************************
* interrupt processing
****************************************************************************/
void 
mgavideo_vsync(struct mga_dev_s *mga)
{
	struct	mga_video_s	*video;
	video = mga->video;

	video->interrupt_count++;
}


/* Converts an image from the framebuffer memory to main memory. */
/* If no complete image was available, returns -1. */
/* If the buffer has been filled with a complete image, returns 0. */

void
mgavideo_grab_needed(struct mga_video_s *video, int state)
{
	u32 scaling;
	if (state == 0) {
		scaling = 0;
		video->ks_scaling = scaling;
		mgavideo_decoder(video, KS0127_SET_HSCALE, &scaling);
	}

	mgagrab_grab_needed( video->mga->grab, state );
}



int mgavideo_grab_image(struct mga_dev_s *mga,
	struct video_mmap *geometry, u8 *buffer, int size)
{
	mgavideo_set_grabsize(mga->video, geometry->width, geometry->height, geometry->format);
	return 0;
}


/* V4L buffers routines */
/* Design issues :
- ks0127 sends a complete image (720 x active_lines before scaling).
- it scales horizontaly to match the demanded v4l grab size ( not the preview size )
- the bes is programmed to display the demanded portion of the image
  (user can independently specify a preview subimage and a grab subimage)
- v4lgrabbing uses hardware horizontal scaling by the ks0127 and
  line dropping for vertical scaling. It can only down-scale, not upscale.
*/

void mgavideo_set_grabsize (struct mga_video_s *video,
	int width, int height, int format)
{
	struct buf_data *buf;
	struct mga_dev_s *mga;
	int	scale_width;

	mga = video->mga;
	buf = &(mga->v4lbuf);
	
	mgagrab_set_grabsize( video->mga->grab, width, height, format );

	scale_width = width;
	
	/* We can only downscale */
	if (width < 2) scale_width = 2;
	if (width > buf->subimage_w) scale_width = buf->subimage_w;

//	if (width == buf->grab_width && height == buf->grab_height) return ;

	if (buf->decoder_hscale) {
		u32 scaling;

		/* Compute the scaling factor (0.15 fixed point) */
		scale_width += buf->subimage_x;
		if (scale_width > 720)
			scale_width = 720;
		scaling = (scale_width << 15) / 720;

		/* Compute the number of pixels sent by the ks0127 */
		buf->real_width = ((720 * scaling) >> 15);

		if (buf->real_width < 2) buf->real_width = 2;
		
		if ( video->ks_scaling != scaling ) {
			video->ks_scaling = scaling;

			mgavideo_decoder(video, KS0127_SET_HSCALE, &scaling);
			buf->scale_change_stamp = mgagrab_get_stamp( video->mga->grab );
		}

		buf->grab_crop_x = (buf->subimage_x * scaling) >> 15;
	}
}


void mgavideo_set_grabsubimage (struct mga_video_s *video,
	int x, int y, int w, int h)
{
	struct buf_data *buf;
	struct mga_dev_s *mga;

	mga = video->mga;
	buf = &(mga->v4lbuf);
	
	mgagrab_set_grabsubimage( video->mga->grab, x, y, w, h );
	
	/* Recompute scaling and bes parameters */
	mgavideo_set_grabsize (video, buf->grab_width, buf->grab_height,
		buf->grab_format);
}


/* /proc routines */
#define ERR_FORMAT -1
#define ERR_OVERFLOW -2

int proc_read_uint(const char *s, unsigned int *val)
{
	unsigned int tmp = 0;
	int to_add;
	int err = 0;
	int stop = 0;
	int readhex = 0;
	int i = 0, i0;
	/* skip leading whitespaces */
	while (s[i] == ' ' || s[i] == '\t')
		i++;
	i0 = i;
	if (s[i] == '0' && s[i+1] == 'x') {
		i += 2;
		readhex = 1;
	}
	do {
		to_add = -1;
		if (s[i] == ' ' || s[i] == '\n' || s[i] == '\t' || s[i] == 0) {
			if (i == i0) err = ERR_FORMAT;
			stop = 1;
		} else if (s[i] >= '0' && s[i] <= '9') {
			to_add = s[i] - '0';
		} else if (s[i] >= 'a' && s[i] <= 'f') {
			if (!readhex)
				err = ERR_FORMAT;
			else
				to_add = s[i] - 'a' + 10;
		} else if (s[i] >= 'A' && s[i] <= 'F') {
			if (!readhex)
				err = ERR_FORMAT;
			else
				to_add = s[i] - 'A' + 10;
		} else {
			err = ERR_FORMAT;
		}
		if (to_add>=0) {
			if (readhex) {
				if (tmp & 0xF0000000)
					err = ERR_OVERFLOW;
				else
					tmp = (tmp << 4) + to_add;
			} else {
				if (tmp > 0x19999999 || (tmp == 0x19999999 && to_add > 5))
					err = ERR_OVERFLOW;
				else
					tmp = tmp * 10 + to_add;
			}
		}
		i++;
	} while (!stop && !err);
	if (!err) {
		*val = tmp;
		err = i;
	}
	return err;
}

/*
Reads a decimal or hex value
Format is [-]0x..... with any number of hex digits 0..9 a..f A..f
       or [-]....... with any number of digits 0..9
Read stops with ' ' or '\t' or '\n' or zero.
*/
int proc_read_int(const char *s, int *val)
{
	unsigned int tmp;
	int err, i = 0;

	/* skip leading whitespaces */
	while (s[i] == ' ' || s[i] == '\t')
		i++;
	
	if (s[i] == '-') {
		i++;
		err = proc_read_uint(s+i, &tmp);
		if (err<0) return err;
		if (tmp & 0x80000000) return ERR_OVERFLOW;
		*val = -tmp;
		return err + i;
	} else {
		err = proc_read_uint(s+i, &tmp);
		if (err<0) return err;
		if (tmp & 0x80000000) return ERR_OVERFLOW;
		*val = tmp;
		return err + i;
	}
	return err;
}

#ifdef	NEVER
static int video_read_proc(char *page, char **start, off_t off,
                  int count, int *eof, void *data)
{
	struct mga_video_s *video = (struct mga_video_s*) data;
	char *s = page;
	int len;

	/* Build the message (it is less than 4k, no buffer overrun) */
	s = s + sprintf(s, "IMAGE_SIZE %d %d\n",
		video->v4lbuf.image_width, mga->v4lbuf.image_height);
	s = s + sprintf(s, "REAL_WIDTH %d\n",
		video->v4lbuf.real_width);
	s = s + sprintf(s, "FIELDS %d\n",
		video->v4lbuf.fields);
	s = s + sprintf(s, "SUBIMAGE %d %d %d %d\n",
		video->v4lbuf.subimage_x, mga->v4lbuf.subimage_y,
		video->v4lbuf.subimage_w, mga->v4lbuf.subimage_h);
	s = s + sprintf(s, "GRAB_SIZE %d %d\n",
		video->v4lbuf.grab_width, mga->v4lbuf.grab_height);
	len = s - page;
	return len;
}
#endif

/****************************************************************************
* initialization
****************************************************************************/


static void
mgavideo_attach(struct mga_dev_s *mga)
{
	struct mga_video_s *video;
	int		   i;

	mga->video = video = kmalloc(sizeof(*mga->video), GFP_KERNEL);
	if (mga->video == NULL)
		return;

	memset(mga->video, 0, sizeof(*mga->video));

	mga->video->mga = mga;

	spin_lock_init(&video->lock);
	

	for(i=0; i < 4; i++) {
		video->input[i].norm = default_norm;
		video->input[i].crop[0].x = 24;
		video->input[i].crop[0].y = 4 * 2;
		video->input[i].crop[0].width = 696;
		video->input[i].crop[0].height = ((_VINHGT_NTSC) - 6) * 2;
#ifdef NEVER
		video->input[i].crop[1].x = 20; /* alt (gz, by thelmen): 14 */
		video->input[i].crop[1].y = 2 * 4; /* alt: 9*2 */
		video->input[i].crop[1].width = 694; /* alt: 705 */
		video->input[i].crop[1].height = ((_VINHGT_PAL) - 12) * 2; /* alt: 14 */
#else
		video->input[i].crop[1].x = 12;
		video->input[i].crop[1].y = 2 * 4;
		video->input[i].crop[1].width = 704;
		video->input[i].crop[1].height = 576	-4; 
#endif
	}

	init_waitqueue_head(&video->sleepq);

	video->input[3].crop[0].x = 20;
	video->input[3].crop[0].y = 4;
	video->input[3].crop[0].width = 696;
	video->input[3].crop[0].height = 472;

	video->input[3].crop[1].x = 20;
	video->input[3].crop[1].y = 0;
	video->input[3].crop[1].width = 700;
	video->input[3].crop[1].height =  576;//((_VINHGT_PAL) - 5) * 2;

	video->norm = -1;
	video->source = -1;

	mga_create_proc_entry(video->mga, "video", mgavideo_read_proc, 
		mgavideo_write_proc, video);

        if (mga->grab != NULL)
        	mgagrab_set_ongrab(mga->grab, mgavideo_grab_image);
	else
	        printk("mgavideo: grab not loaded");
}


static void
mgavideo_deattach(struct mga_video_s *video)
{
        if (video->mga->grab != NULL)
        	mgagrab_set_ongrab(video->mga->grab, NULL);
	else
	        printk("mgavideo: grab not loaded");

	mga_disable_irq(video->mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)mgavideo_vsync);
	mga_remove_proc_entry(video->mga, "video");

	video->mga->video = NULL;
	kfree(video);
}


struct id_to_sym_s
{
	int id;
	const char *sym;
};

static const char*
id_to_ascii( struct id_to_sym_s *lut, int size, int id )
{
	int i;
	for ( i=0; i<size/sizeof(*lut); i++ ) {
		if ( lut[ i ].id == id )
			return lut[ i ].sym;
	}
	return "None";
}


int mgavideo_read_proc(char *page, char **start, off_t off,
                  int count, int *eof, void *data)
{
	struct mga_video_s *video = (struct mga_video_s*)data;
	char *s = page;
	int i;
	int len;

	static struct id_to_sym_s print_input[] = {
		{ KS_INPUT_COMPOSITE, "Composite" },
		{ KS_INPUT_SVIDEO, "S-Video" },
		{ KS_INPUT_TUNER, "Tuner" },
		{ KS_INPUT_YUV656, "MJPEG" }
	};

	static struct id_to_sym_s print_norm[] = {
		{ KS_STD_NTSC, "NTSC" },
		{ KS_STD_PAL, "PAL" },
		{ KS_STD_NTSC_M, "NTSC 50Hz" },
		{ KS_STD_PAL_N, "PAL 60Hz" },
		{ KS_STD_SECAM, "SECAM" }
	};
	
	s = s + sprintf(s, "Active input: %s\n",
		id_to_ascii( print_input, sizeof(print_input), video->source ) );
	s = s + sprintf(s, "Active Standard: %s\n",
		id_to_ascii( print_norm, sizeof(print_norm), video->norm ) );

	s = s + sprintf(s, "Vsync interrupt count: %d\n", video->interrupt_count);

	s = s +sprintf(s, "\n");

	for(i=0; i < 4; i++) {
		s = s + sprintf(s, "%s input\n",
			id_to_ascii( print_input, sizeof(print_input), i ) );
		s = s + sprintf(s, "Standard: %s\n", 
			id_to_ascii( print_norm, sizeof(print_norm), video->input[i].norm ) );
		s = s + sprintf(s , "Brightness: %d Hue: %d Contrast: %d Saturation: %d\n",
			video->input[i].brightness,	
			video->input[i].hue,	
			video->input[i].contrast,	
			video->input[i].color);
		s = s + sprintf(s, "NTSC Cropping: %d,%d %dx%d\n",
			video->input[i].crop[0].x,
			video->input[i].crop[0].y,
			video->input[i].crop[0].width,
			video->input[i].crop[0].height);
		s = s + sprintf(s, "PAL Cropping: %d,%d %dx%d\n",
			video->input[i].crop[1].x,
			video->input[i].crop[1].y,
			video->input[i].crop[1].width,
			video->input[i].crop[1].height);
		s = s +sprintf(s, "\n");
	}

	len = s - page;

	return len;
}


int mgavideo_write_proc(struct file *file, const char *buf,
                          unsigned long count, void *data)
{
	struct mga_video_s *video = (struct mga_video_s *)data;

	(void)video;

	return count;		
}



#ifdef MODULE
int init_module(void)
#else
int init_mgavideo(void) //??
#endif
{
	struct mga_dev_s *mga;
	int		 i;

	printk(KERN_INFO "Matrox KS switchboard\n");

	for(i=0; (mga = mga_get_card(i)) != NULL ; i++) {
		mgavideo_attach(mga);
	}

	if(i == 0) {
		return -ENXIO;
	}
	return 0;
}


#ifdef MODULE
void cleanup_module(void)
{
	int i = 0;
	struct mga_dev_s *mga;

	for(i=0; (mga = mga_get_card(i)) != NULL; i++) {
	
		mgavideo_deattach(mga->video);

		/* need to release it 2 times */
		mga_release_card(mga);
		mga_release_card(mga);
	}
}
#endif




void
mgavideo_lock(struct mga_dev_s *mga)
{
	int tri;

	tri = KS_OUTPUT_NORMAL;
	mga_i2c_control(mga,
		I2C_DRIVERID_KS0127, 
		KS0127_SET_OUTPUT_MODE, &tri);

	/*HACK2: see HACK below; now also an enable is here so _set_active will always be entered
	with the grabber active. Without this, _set_active would detect no input change the 2nd time
	and not enable the grabber (after closing and opening the device again).
	is it a hack or a solution now? -- lvs */
	mgavideo_enable_grabber(mga->video);
	
	// XXX MOD_INC_USE_COUNT;
}


void
mgavideo_unlock(struct mga_dev_s *mga)
{
	/*HACK: mgagrab_deactivate and _activate aren't called symmetrically, leaving the
		grabber activated when the driver is closed, meaning the BES and VIN
		are continuously reprogrammed when the driver is idle (loaded but unopened).
		Problem originates in mgavideo_set_active, where at the very first invocation
		the grab-activation count is increased to 1 (first tries to decrease (fails) then
		reenables grabbing (succeeds)), but there is currently no way for the user to
		reset this to 0.
		Doing the disable explicitly:
		-- lvs
	*/
	mgavideo_disable_grabber(mga->video);
	/*end hack*/

	/* todo, probably will kill card with this
	int tri;
		tri = KS_OUTPUT_SYNC_CLK_HIZ;
		mga_i2c_control(video->mga,
			I2C_DRIVERID_KS0127, 
			KS0127_SET_OUTPUT_MODE, &tri);
	*/

	// XXX MOD_DEC_USE_COUNT;
}


EXPORT_SYMBOL(mgavideo_audio);
EXPORT_SYMBOL(mgavideo_decoder);
EXPORT_SYMBOL(mgavideo_set_active);
EXPORT_SYMBOL(mgavideo_get_active);
EXPORT_SYMBOL(mgavideo_set_color);
EXPORT_SYMBOL(mgavideo_set_norm);
EXPORT_SYMBOL(mgavideo_get_norm);
EXPORT_SYMBOL(mgavideo_input);
EXPORT_SYMBOL(mgavideo_set_freq);
EXPORT_SYMBOL(mgavideo_get_freq);
EXPORT_SYMBOL(mgavideo_grab_needed);
EXPORT_SYMBOL(mgavideo_grab_image);
EXPORT_SYMBOL(mgavideo_vsync);
EXPORT_SYMBOL(mgavideo_lock);
EXPORT_SYMBOL(mgavideo_unlock);

