/*=========================================================================

    Copyright (C) 2000 Bas Vermeulen -- bvermeul@blackstar.nl

    No Wires Needed SNWNMP

    snwnmp.c 0.3.1 2002/03/03

=========================================================================*/
#include <linux/config.h>
#ifdef __IN_PCMCIA_PACKAGE__
#include <pcmcia/k_compat.h>
#endif

#include <linux/netdevice.h>
#include <linux/wireless.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ciscode.h>
#include <pcmcia/ds.h>

#include "snwnmp.h"
#include "ring.h"

static char *version = "No Wires Needed SNWNMP driver v0.3.1";

#define MAX_ROAMTABLE_SIZE		256

struct roam_entry_t {
	unsigned char bssid[MAX_ADDR_LEN];
	char essid[IW_ESSID_MAX_SIZE + 1];
	int bsstype;
	int channel;
	int roamage;
	int quality;
	int load;
	int beaconperiod;
	int dtimperiod;
	u16 capinfo;
	unsigned char rates[8];
};

struct roam_table_t {
	struct roam_entry_t entry[MAX_ROAMTABLE_SIZE];
	int size;
};

struct nwn_key {
	unsigned char data[MAX_KEY_SIZE];
	int len;
};

struct nwn_encryption {
	int has_big_wep;
	int airlock;
	int enable;
	int restricted;
	int index;
	struct nwn_key keys[MAX_KEYS];
};

struct nwn_retry_limit {
	unsigned long lifetime;
	unsigned long short_limit;
	unsigned long long_limit;
};

struct nwn_tx_power {
	unsigned long num_txpower;
	unsigned long cur_txpower;
	unsigned long tx_power[8];
};

struct nwn_counters {
	unsigned long transmitted_fragment_count;
	unsigned long multicast_transmitted_frame_count;
	unsigned long failed_count;
	unsigned long retry_count;
	unsigned long multiple_retry_count;
	unsigned long frame_duplicate_count;
	unsigned long rts_success_count;
	unsigned long rts_failure_count;
	unsigned long ack_failure_count;
	unsigned long received_fragment_count;
	unsigned long multicast_received_frame_count;
	unsigned long fcs_error_count;
	unsigned long transmitted_frame_count;
	unsigned long wep_undecryptable_count;
};

struct nwn_capabilities {
	unsigned char supportedRates[8];
	unsigned short capinfo;
	unsigned short supportAirlock;
	unsigned short supportAdHoc;
};

struct nwn_private {
	dev_node_t node;
	struct net_device_stats stats;
	struct roam_table_t roam_table;
	struct nwn_encryption enc;
	struct nwn_capabilities capabilities;
	struct nwn_counters counter;
	struct poldhu_ring_buffer ring;
	char ssid[IW_ESSID_MAX_SIZE + 1];
	char wanted_essid[IW_ESSID_MAX_SIZE + 1];
	unsigned char bssid[MAX_ADDR_LEN];
	unsigned long bsstype;
	unsigned long channel;
	unsigned char oprates[8];
	unsigned long rts_threshold;
	unsigned long frag_threshold;
	struct nwn_tx_power tx_power;
	struct nwn_retry_limit retry;
	struct timer_list bssid_timer;
	struct timer_list roam_timer;
	spinlock_t lock;
	int is_550;
};

#ifdef CONFIG_SNWNMP_DEBUG
static int snwnmp_debug = CONFIG_SNWNMP_DEBUG;
#define DEBUG(n, args...)  if (snwnmp_debug>(n)) printk(KERN_DEBUG args)
#define PRINT(n, args...) printk(KERN_INFO args)
#else
#define DEBUG(n, args...)
#define PRINT(n, args...) printk(KERN_DEBUG args)
#endif

#define kfree_s(o,s) kfree(o)
#define SNWNMP_STRING(n) (n + ETH_HLEN + SNWNMP_HLEN)
#define SNWNMP_LONG(n)   *((long*) (n + ETH_HLEN + SNWNMP_HLEN))
#define SNWNMP_SHORT(n)  *((short*) (n + ETH_HLEN + SNWNMP_HLEN))

/*======================================================================*/
/* SNWNMP functions                                                     */
/*======================================================================*/

int snwnmp_handle_trap(struct net_device *dev, unsigned char *frame,
		       unsigned len)
{
	struct snwnmp_header *header = (struct snwnmp_header *) frame;

	switch (ntohl(header->oid)) {
	case OID_NWN_BSSCHANGETRAP:
		snwnmp_get(dev, OID_NWN_SMTCURRENTBSSID);
		break;
	case OID_3COM_RESETTRAP:
		break;
	default:
		printk(KERN_DEBUG "%s: SNWNMP TRAP (oid %x)\n",
		       dev->name, ntohl(header->oid));
	}
	return 0;
}

int snwnmp_handle_response(struct net_device *dev, unsigned char *frame,
			   unsigned len)
{
	struct snwnmp_header *header = (struct snwnmp_header *) frame;
	struct nwn_private *lp = (struct nwn_private *) dev->priv;
	int i, associated = 0;
	int print_bssid = 0;

#if defined(SNWNMP_DEBUG)
	printk(KERN_DEBUG "%s: SNWNMP response to 0x%lx received: ",
	       dev->name, ntohl(header->oid));
	for (i = 0; i < len; i++)
		printk("%02x ", frame[i]);
	printk("\n");
#endif

	/* Acquire the lock */
	spin_lock(&lp->lock);

	switch (ntohl(header->oid)) {
	case OID_NWN_SMTCURRENTSSID:
		PRINT(0, "%s: Current SSID is %s\n", dev->name, 
				SNWNMP_STRING(frame));
		memcpy(lp->ssid, SNWNMP_STRING(frame), 32);
		break;
	case OID_NWN_SMTCURRENTBSSID:
		print_bssid = 0;
		for (i = 0; i < 6; i++) {
			if (lp->bssid[i] !=
			    frame[i + ETH_HLEN + SNWNMP_HLEN]) {
				lp->bssid[i] =
				    frame[i + ETH_HLEN + SNWNMP_HLEN];
				print_bssid = 1;
			}
			if (frame[i + ETH_HLEN + SNWNMP_HLEN] != 0)
				associated = 1;
		}
		if (print_bssid) {
			printk(KERN_DEBUG "%s: Current BSSID is ",
			       dev->name);
			for (i = 0; i < 6; i++)
				printk("%02x%s",
				       frame[i + ETH_HLEN + SNWNMP_HLEN],
				       i < 5 ? ":" : "\n");
		}
		if (!associated && print_bssid) {
			snwnmp_get(dev, OID_DOT11_DOT11DISASSOCIATEREASON);
			snwnmp_get(dev, OID_DOT11_DOT11DEAUTHENTICATEREASON);
		}
		if (associated) {
			snwnmp_get(dev, OID_NWN_SMTCURRENTBSSTYPE);
			snwnmp_get(dev, OID_DOT11_DOT11CURRENTCHANNEL);
			snwnmp_get(dev, OID_NWN_SMTCAPABILITYINFO);
			snwnmp_get(dev, OID_NWN_SMTCURRENTSSID);
		} 
		break;
	case OID_DOT11_DOT11POWERMANAGEMENTMODE:
		PRINT(0, "%s: %s\n", dev->name, 
				ntohl(SNWNMP_LONG(frame)) == 1 ?
				"Active Mode" : "Power Save Mode");
		break;
	case OID_DOT11_DOT11DEAUTHENTICATEREASON:
		PRINT(0, "%s: Station deauthenticated (%d)", dev->name, ntohl(SNWNMP_LONG(frame)));
		switch (ntohl(SNWNMP_LONG(frame)))
		{
		case 1:
			printk("Unspecified reason");
			break;
		case 2:
			printk("Previous authentication no longer valid");
			break;
		case 3:
			printk("Deauthenticated because sending station is leaving (or has left) IBSS or ESS");
			break;
		case 4:
			printk("Disassociated due to inactivity");
			break;
		case 5:
			printk("Disassociated because AP is unable to handle all currently associated stations");
			break;
		case 6:
			printk("Class 2 frame received from nonauthenticated station");
			break;
		case 7:
			printk("Class 3 frame received from nonassociated station");
			break;
		case 8:
			printk("Disassociated because sending station is leaving (or has left) BSS");
			break;
		case 9:
			printk("Station requesting (re)association is not authenticated with responding station");
			break;
		default:
			printk("Reserved");
			break;
		}
		printk("\n");
		break;
	case OID_DOT11_DOT11DISASSOCIATEREASON:
		PRINT(0, "%s: Station disassociated (%d)", dev->name, ntohl(SNWNMP_LONG(frame)));
		switch (ntohl(SNWNMP_LONG(frame)))
		{
		case 1:
			printk("Unspecified reason");
			break;
		case 2:
			printk("Previous authentication no longer valid");
			break;
		case 3:
			printk("Deauthenticated because sending station is leaving (or has left) IBSS or ESS");
			break;
		case 4:
			printk("Disassociated due to inactivity");
			break;
		case 5:
			printk("Disassociated because AP is unable to handle all currently associated stations");
			break;
		case 6:
			printk("Class 2 frame received from nonauthenticated station");
			break;
		case 7:
			printk("Class 3 frame received from nonassociated station");
			break;
		case 8:
			printk("Disassociated because sending station is leaving (or has left) BSS");
			break;
		case 9:
			printk("Station requesting (re)association is not authenticated with responding station");
			break;
		default:
			printk("Reserved");
			break;
		}
		printk("\n");
		break;
	case OID_DOT11_DOT11CURRENTCHANNEL:
		PRINT(0, "%s: Current Channel (%d)\n", dev->name,
		      ntohl(SNWNMP_LONG(frame)));
		lp->channel = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_NWN_SMTCAPABILITYINFO:
		PRINT(1, "%s: Capabilities (%02X)\n", dev->name,
		      ntohs(SNWNMP_SHORT(frame)));
		lp->capabilities.capinfo = ntohs(SNWNMP_SHORT(frame));
		break;
	case OID_DOT11_DOT11RTSTHRESHOLD:
		PRINT(0, "%s: RTS Threshold (%d)\n", dev->name,
		      ntohl(SNWNMP_LONG(frame)));
		lp->rts_threshold = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11FRAGMENTATIONTHRESHOLD:
		PRINT(0, "%s: Fragmentation Threshold (%d)\n", dev->name,
		      ntohl(SNWNMP_LONG(frame)));
		lp->frag_threshold = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11SHORTRETRYLIMIT:
		lp->retry.short_limit = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11LONGRETRYLIMIT:
		lp->retry.long_limit = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11MAXTRANSMITLIFETIME:
		lp->retry.lifetime = ntohl(SNWNMP_LONG(frame));
		snwnmp_get(dev, OID_DOT11_DOT11SHORTRETRYLIMIT);
		snwnmp_get(dev, OID_DOT11_DOT11LONGRETRYLIMIT);
		break;
	case OID_DOT11_DOT11NUMBERSUPPORTEDPOWERLEVELS:
		lp->tx_power.num_txpower = ntohl(SNWNMP_LONG(frame));

		for (i = 0; i < ntohl(SNWNMP_LONG(frame)); i++)
			snwnmp_get(dev, OID_DOT11_DOT11TXPOWERLEVEL1 + i);
		break;
	case OID_DOT11_DOT11CURRENTTXPOWERLEVEL:
		lp->tx_power.cur_txpower = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TXPOWERLEVEL1:
		lp->tx_power.tx_power[0] = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TXPOWERLEVEL2:
		lp->tx_power.tx_power[1] = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TXPOWERLEVEL3:
		lp->tx_power.tx_power[2] = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TXPOWERLEVEL4:
		lp->tx_power.tx_power[3] = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TXPOWERLEVEL5:
		lp->tx_power.tx_power[4] = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TXPOWERLEVEL6:
		lp->tx_power.tx_power[5] = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TXPOWERLEVEL7:
		lp->tx_power.tx_power[6] = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TXPOWERLEVEL8:
		lp->tx_power.tx_power[7] = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11DESIREDSSID:
		if (strlen(SNWNMP_STRING(frame)) == 0)
			PRINT(0, "%s: Requested SSID (Any)\n", dev->name);
		else
			PRINT(0, "%s: Requested SSID (%s)\n",
				dev->name, SNWNMP_STRING(frame));
		break;
	case OID_DOT11_DOT11AUTHENTICATIONALGORITHMENABLE:
		PRINT(0, "%s: Current authentication algorithm (%d,%d)\n",
		      dev->name,
		      ntohl(SNWNMP_LONG(frame)),
		      ntohl(SNWNMP_LONG(frame + sizeof(long))));
		break;
	case OID_DOT11_DOT11PRIVACYINVOKED:
		PRINT(0, "%s: Current privacy setting (%d)\n",
		      dev->name, ntohl(SNWNMP_LONG(frame)));
		break;
	case OID_DOT11_DOT11EXCLUDEUNENCRYPTED:
		PRINT(0, "%s: Excluding unencrypted traffic (%d)\n",
		      dev->name, ntohl(SNWNMP_LONG(frame)));
		break;
	case OID_NWN_SMTPUBLICKEYENABLE:
		PRINT(0, "%s: Public key enable (%d)\n",
		      dev->name, ntohl(SNWNMP_LONG(frame)));
		break;
	case OID_DOT11_DOT11WEPDEFAULTKEYID:
		PRINT(0, "%s: Default key ID (%d)\n",
		      dev->name, ntohl(SNWNMP_LONG(frame)));
		break;
	case OID_DOT11_DOT11WEPDEFAULTKEY:
#if CONFIG_PCMCIA_SWALLOW_DEBUG
		if (swallow_debug > 0) {
			printk(KERN_DEBUG "%s: Default key(s) ", dev->name);
			for (i = 0; i < ntohl(header->length); i++)
				printk("%02x ",
				       frame[i + ETH_HLEN + SNWNMP_HLEN]);
			printk("\n");
		}
#endif
		break;
	case OID_NWN_SMTOPERATIONALCHANNELS:
		printk(KERN_DEBUG "%s: Operational channels:", dev->name);
		for (i = 0; i < ntohl(header->length); i++)
			printk("%02x ", frame[i + ETH_HLEN + SNWNMP_HLEN]);
		printk("\n");
		break;
	case OID_NWN_SMTQUALITYLEVEL0:
	case OID_NWN_SMTQUALITYLEVEL1:
	case OID_NWN_SMTQUALITYLEVEL2:
		PRINT(0, "%s: Quality level %d = %d\n", dev->name,
				ntohl(header->oid) - OID_NWN_SMTQUALITYLEVEL0,
				ntohl(SNWNMP_LONG(frame)));
		break;
	case OID_NWN_SMTQUALITYINDICATOR:
		PRINT(0, "%s: Quality %d\n", dev->name, ntohl(SNWNMP_LONG(frame)));
		break;
	case OID_NWN_SMTQUALITYUPPERLIMIT:
		PRINT(0, "%s: Quality upper limit %d\n", dev->name, ntohl(SNWNMP_LONG(frame)));
		break;
	case OID_NWN_SMTQUALITYLOWERLIMIT:
		PRINT(0, "%s: Quality lower limit %d\n", dev->name, ntohl(SNWNMP_LONG(frame)));
		break;
	case OID_NWN_SMTCURRENTBSSTYPE:
		lp->bsstype = ntohl(SNWNMP_LONG(frame));
	case OID_DOT11_DOT11DESIREDBSSTYPE:
		switch (ntohl(SNWNMP_LONG(frame))) {
		case 1:
			PRINT(0, "%s: Managed mode\n", dev->name);
			break;
		case 2:
			PRINT(0, "%s: Ad-Hoc mode\n", dev->name);
			break;
		case 3:
			PRINT(0, "%s: Managed mode with Airlock\n", dev->name);
			break;
		default:
			PRINT(0, "%s: Unknown mode (%ld)\n", dev->name, lp->bsstype);
			break;
		}
		break;
	case OID_DOT11_DOT11OPERATIONALRATESET:
		memcpy(lp->oprates, SNWNMP_STRING(frame), 
				ntohl(header->length));
		break;
	case OID_DOT11_DOT11SUPPORTEDDATARATESTX:
	case OID_DOT11_DOT11SUPPORTEDDATARATESRX:
		for (i=0; i < ntohl(header->length) / sizeof(long); i++)
			lp->capabilities.supportedRates[i] = 
				(unsigned char)ntohl(((long*)SNWNMP_STRING(frame))[i]);
		break;
		/* Roaming OID's */
	case OID_NWN_ROAMTABLELENGTH:
		lp->roam_table.size = ntohl(SNWNMP_LONG(frame));
		mod_timer(&lp->roam_timer, jiffies + HZ * 60);
		break;
	case OID_NWN_ROAMSSID:
		if (lp->is_550) {
			int start, end, index = 0;
			start = end = ETH_HLEN + SNWNMP_HLEN;
			while (end < len) {
				while (frame[end++]);
				memcpy(lp->roam_table.entry[index++].essid,
				       frame + start, end - start);
				start = end;
			}
		} else {
			for (i = 0; i < lp->roam_table.size; i++) {
				memcpy(lp->roam_table.entry[i].essid,
				       frame + ETH_HLEN + SNWNMP_HLEN +
				       i * IW_ESSID_MAX_SIZE,
				       IW_ESSID_MAX_SIZE);
				lp->roam_table.entry[i].
				    essid[IW_ESSID_MAX_SIZE] = '\0';
			}
		}
		break;
	case OID_NWN_ROAMBSSID:
		for (i = 0; i < lp->roam_table.size; i++) {
			memcpy(lp->roam_table.entry[i].bssid,
			       frame + ETH_HLEN + SNWNMP_HLEN + i * 6, 6);
		}
		break;
	case OID_NWN_ROAMCAPABILITYINFORMATION:
		for (i = 0; i < lp->roam_table.size; i++) {
			lp->roam_table.entry[i].capinfo =
			    ntohs(SNWNMP_SHORT(frame + i * sizeof(short)));
		}
		break;
	case OID_NWN_ROAMQUALITY:
		for (i = 0; i < lp->roam_table.size; i++) {
			lp->roam_table.entry[i].quality =
			    ntohl(SNWNMP_LONG(frame + i * sizeof(long)));
		}
		break;
	case OID_NWN_ROAMLOAD:
		for (i = 0; i < lp->roam_table.size; i++) {
			lp->roam_table.entry[i].load =
			    ntohl(SNWNMP_LONG(frame + i * sizeof(long)));
		}
		break;
	case OID_NWN_ROAMCHANNEL:
		for (i = 0; i < lp->roam_table.size; i++) {
			lp->roam_table.entry[i].channel =
			    ntohl(SNWNMP_LONG(frame + i * sizeof(long)));
		}
		break;
	case OID_NWN_ROAMBSSTYPE:
		for (i = 0; i < lp->roam_table.size; i++) {
			lp->roam_table.entry[i].bsstype =
			    ntohl(SNWNMP_LONG(frame + i * sizeof(long)));
		}
		break;
	case OID_NWN_ROAMAGE:
		for (i = 0; i < lp->roam_table.size; i++) {
			lp->roam_table.entry[i].roamage =
			    ntohl(SNWNMP_LONG(frame + i * sizeof(long)));
		}
		break;
	case OID_NWN_ROAMBEACONPERIOD:
		for (i = 0; i < lp->roam_table.size; i++) {
			lp->roam_table.entry[i].beaconperiod =
			    ntohl(SNWNMP_LONG(frame + i * sizeof(long)));
		}
		break;
	case OID_NWN_ROAMDTIMPERIOD:
		for (i = 0; i < lp->roam_table.size; i++) {
			lp->roam_table.entry[i].dtimperiod =
			    ntohl(SNWNMP_LONG(frame + i * sizeof(long)));
		}
		break;
	case OID_NWN_ROAMRATES:
		for (i = 0; i < lp->roam_table.size; i++) {
			int j;
			for (j = 0; j < 6; j++)
			{
				lp->roam_table.entry[i].rates[j] =
					SNWNMP_STRING(frame)[i*6 + j] & ~0x80;
			}
		}
		break;
	/* Various counters */
	case OID_DOT11_DOT11TRANSMITTEDFRAGMENTCOUNT:
		lp->counter.transmitted_fragment_count = ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11MULTICASTTRANSMITTEDFRAMECOUNT:
		lp->counter.multicast_transmitted_frame_count =
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11FAILEDCOUNT:
		lp->counter.failed_count =
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11RETRYCOUNT:
		lp->counter.retry_count = 
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11MULTIPLERETRYCOUNT:
		lp->counter.multiple_retry_count = 
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11FRAMEDUPLICATECOUNT:
		lp->counter.frame_duplicate_count =
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11RTSSUCCESSCOUNT:
		lp->counter.rts_success_count =
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11RTSFAILURECOUNT:
		lp->counter.rts_failure_count =
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11ACKFAILURECOUNT:
		lp->counter.ack_failure_count =
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11RECEIVEDFRAGMENTCOUNT:
		lp->counter.received_fragment_count =
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11MULTICASTRECEIVEDFRAMECOUNT:
		lp->counter.multicast_received_frame_count =
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11FCSERRORCOUNT:
		lp->counter.fcs_error_count = 
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11TRANSMITTEDFRAMECOUNT:
		lp->counter.transmitted_frame_count = 
			ntohl(SNWNMP_LONG(frame));
		break;
	case OID_DOT11_DOT11WEPUNDECRYPTABLECOUNT:
		lp->counter.wep_undecryptable_count = 
			ntohl(SNWNMP_LONG(frame));
		break;
		/* 3Com specific OID's */
	case OID_3COM_INFRASTRUCTURE:
	case OID_3COM_ADHOC:
		if (ntohl(SNWNMP_LONG(frame)))
			PRINT(0, "%s: WEP encryption\n", dev->name);
		else
			PRINT(0, "%s: Open system\n", dev->name);
		break;
	case OID_3COM_ACTIVE_PROBING:
		if (ntohl(SNWNMP_LONG(frame)))
			PRINT(0, "%s: Active Probing enabled\n", dev->name);
		else
			PRINT(0, "%s: Active Probing disabled\n", dev->name);
		break;
	default:
		printk(KERN_DEBUG
		       "%s: Received unknown oid %X with length %d\n",
		       dev->name, ntohl(header->oid),
		       ntohl(header->length));
		break;
	}
	/* Release the lock */
	spin_unlock(&lp->lock);
	return 0;
}

/*
 * This function re-sends failed SNWNMP get packages, after fixing them up.
 */
int snwnmp_handle_error(struct net_device *dev, unsigned char *frame,
			unsigned len)
{
	struct sk_buff *skb;
	struct snwnmp_header *header = (struct snwnmp_header *) frame;
	unsigned pkt_len = 0, header_len = 0;
#if defined(SNWNMP_DEBUG)
	int i;
#endif

	if ((ntohl(header->length) == 0x5a5a5a5a) ||
	    (ntohl(header->length) == 0x00000000)) {
		switch (ntohl(header->oid)) {
		case OID_NWN_SMTCURRENTSSID:
			header->length = htonl(32);
			break;
		case OID_NWN_SMTCURRENTBSSID:
			header->length = htonl(6);
			break;
		case OID_NWN_ROAMTABLELENGTH:
			header->length = htonl(4);
			break;
		case OID_NWN_ROAMSSID:
			header->length = htonl(128);
			break;
		case OID_NWN_ROAMBSSID:
			header->length = htonl(48);
			break;
		case OID_NWN_ROAMBSSTYPE:
			header->length = htonl(32);
			break;
		case OID_NWN_ROAMCAPABILITYINFORMATION:
			header->length = htonl(16);
			break;
		case OID_NWN_ROAMAGE:
			header->length = htonl(32);
			break;
		case OID_NWN_ROAMCHANNEL:
			header->length = htonl(32);
			break;
		case OID_NWN_ROAMQUALITY:
			header->length = htonl(32);
			break;
		case OID_NWN_ROAMLOAD:
			header->length = htonl(32);
			break;
		case OID_NWN_ROAMBEACONPERIOD:
			header->length = htonl(32);
			break;
		case OID_NWN_ROAMDTIMPERIOD:
			header->length = htonl(32);
			break;
		case OID_NWN_ROAMRATES:
			header->length = htonl(48);
			break;
		default:
			printk(KERN_DEBUG
			       "swallow_cs: Unsupported OID(0x%x)\n",
			       ntohl(header->oid));
			return -EOPNOTSUPP;
		}
	}

	header_len = ETH_HLEN + 12;
	pkt_len = header_len + ntohl(header->length);

#if defined(SNWNMP_DEBUG)
	printk(KERN_DEBUG "%s: Resending %d bytes.\n", dev->name, pkt_len);
#endif

	memcpy(header->h_dest, dev->dev_addr, ETH_ALEN);
	memcpy(header->h_source, dev->dev_addr, ETH_ALEN);
	header->h_proto = __constant_htons(ETH_P_SNWNMP);
	header->version = 1;
	header->operation = SNWNMP_GET;

	if ((skb = dev_alloc_skb(pkt_len + dev->hard_header_len + 15)) == NULL) {
		printk(KERN_DEBUG "%s: couldn't allocate an sk_buff of"
		       " size %d in snwnmp_handle_error()\n", dev->name,
		       pkt_len + 5);
		return -ENOMEM;
	}

	memcpy(skb_put(skb, header_len), frame, header_len);
	memset(skb_put(skb, ntohl(header->length)), 0,
	       ntohl(header->length));

	skb->dev = dev;
	skb->protocol = __constant_htons(ETH_P_SNWNMP);
	skb->nh.raw = skb->data;
	if (dev->hard_header)
		skb->nh.raw += dev->hard_header_len;

	return dev_queue_xmit(skb);
}

/*
 * Get all values for a specific OID.
 * We send a packet to find out how large the buffer should be;
 * the processing part will make sure the request gets sent out again.
 */
int snwnmp_get(struct net_device *dev, unsigned long oid)
{
	struct sk_buff *skb;
	struct snwnmp_header *header;
	unsigned char *packet;
	unsigned int pkt_len = 0;
#if defined(SNWNMP_DEBUG)
	int i;
#endif

#if defined(SNWNMP_DEBUG)
	printk(KERN_DEBUG "%s: Getting OID 0x%lx\n", dev->name, oid);
#endif

	pkt_len = ETH_HLEN + SNWNMP_HLEN;

	if ((packet = (unsigned char *)
	     kmalloc(pkt_len, GFP_ATOMIC)) == NULL) {
		printk(KERN_DEBUG "snwnmp_get: Low on memory.\n");
		return -ENOMEM;
	}
	memset(packet, 0, pkt_len);
	header = (struct snwnmp_header *) packet;

	/* Fill in the ether packet to send up to the card */
	memcpy(header->h_dest, dev->dev_addr, ETH_ALEN);
	memcpy(header->h_source, dev->dev_addr, ETH_ALEN);
	header->h_proto = __constant_htons(ETH_P_SNWNMP);
	header->version = 0x01;
	header->operation = SNWNMP_GET;
	header->oid = htonl(oid);
	header->index = 0;
	header->flags = 0;
	header->length = 0;

	/* Allocate an sk_buffer */
	if ((skb =
	     dev_alloc_skb(pkt_len + dev->hard_header_len + 15)) == NULL) {
		kfree(packet);
		printk(KERN_DEBUG "%s: couldn't allocate an sk_buff of"
		       " size %d in snwnmp_get()\n", dev->name, pkt_len);
		return -ENOMEM;
	}

	skb_reserve(skb, 2);

	memcpy(skb_put(skb, pkt_len), packet, pkt_len);

	skb->dev = dev;
	skb->protocol = __constant_htons(ETH_P_SNWNMP);
	skb->nh.raw = skb->data;
	if (dev->hard_header)
		skb->nh.raw += dev->hard_header_len;

	if (skb->len != pkt_len)
		printk(KERN_DEBUG "%s: Error! skb->len != pkt_len\n",
		       dev->name);

#if defined(SNWNMP_DEBUG)
	printk(KERN_DEBUG "%s: Sending SNWNMP get: ", dev->name);
	for (i = 0; i < pkt_len; i++)
		printk("%02x ", ((unsigned char *) packet)[i]);
	printk("\n");
#endif

	kfree(packet);

	return dev_queue_xmit(skb);
}

int snwnmp_set(struct net_device *dev, unsigned long oid,
	       unsigned char *data, unsigned len)
{
	struct sk_buff *skb;
	struct snwnmp_header *header;
	unsigned char *packet;
	unsigned int pkt_len = 0;
#if defined(SNWNMP_DEBUG)
	int i;
#endif

	pkt_len = ETH_HLEN + SNWNMP_HLEN + len;

#if defined(SNWNMP_DEBUG)
	printk(KERN_DEBUG "%s: Setting OID 0x%lx\n", dev->name, oid);
#endif

	if ((packet =
	     (unsigned char *) kmalloc(pkt_len, GFP_ATOMIC)) == NULL) {
		printk(KERN_DEBUG
		       "%s: Unable to allocate %d bytes in snwnmp_set()\n",
		       dev->name, pkt_len);
		return -ENOMEM;
	}
	memset(packet, 0, pkt_len);
	header = (struct snwnmp_header *) packet;

	/* Fill in the ethernet header */
	memcpy(header->h_dest, dev->dev_addr, ETH_ALEN);
	memcpy(header->h_source, dev->dev_addr, ETH_ALEN);
	header->h_proto = __constant_htons(ETH_P_SNWNMP);
	/* Fill in the SNWNMP header */
	header->version = 1;
	header->operation = SNWNMP_SET;
	header->oid = htonl(oid);
	header->length = htonl(len);

	/* Copy in the data */
	memcpy(packet + ETH_HLEN + SNWNMP_HLEN, data, len);

	/* Allocate an sk_buffer */
	if ((skb =
	     dev_alloc_skb(pkt_len + dev->hard_header_len + 15)) == NULL) {
		kfree(packet);
		printk(KERN_DEBUG "%s: couldn't allocate an sk_buff of"
		       " size %d in snwnmp_set()\n", dev->name, pkt_len);
		return -ENOMEM;
	}

	skb_reserve(skb, 2);

	memcpy(skb_put(skb, pkt_len), packet, pkt_len);

	skb->dev = dev;
	skb->protocol = __constant_htons(ETH_P_SNWNMP);
	skb->nh.raw = skb->data;
	if (dev->hard_header)
		skb->nh.raw += dev->hard_header_len;

	if (skb->len != pkt_len)
		printk(KERN_DEBUG "%s: Error! skb->len != pkt_len\n",
		       dev->name);

#if defined(SNWNMP_DEBUG)
	printk(KERN_DEBUG "%s: Sending SNWNMP set: ", dev->name);
	for (i = 0; i < pkt_len; i++)
		printk("%02x ", packet[i]);
	printk("\n");
#endif

	kfree(packet);

	return dev_queue_xmit(skb);
}


int snwnmp_process(struct net_device *dev, unsigned char *frame,
		   unsigned len)
{
	struct snwnmp_header *header = (struct snwnmp_header *) frame;

	if (header->h_proto != __constant_htons(ETH_P_SNWNMP)) {
		printk(KERN_DEBUG
		       "%s: Wrong ether type(0x%x) in process_snwnmp.\n",
		       dev->name, ntohs(header->h_proto));
		return -EOPNOTSUPP;
	}

	switch (header->operation) {
	case SNWNMP_ERROR:
		return snwnmp_handle_error(dev, frame, len);
	case SNWNMP_TRAP:
		return snwnmp_handle_trap(dev, frame, len);
	case SNWNMP_RESPONSE:
		return snwnmp_handle_response(dev, frame, len);
	case SNWNMP_GET:
	case SNWNMP_SET:
		printk(KERN_DEBUG
		       "%s: Unexpected operation received (%d)\n",
		       dev->name, header->operation);
		break;
	default:
		printk(KERN_DEBUG
		       "%s: Don't know how to handle operation (%d)\n",
		       dev->name, header->operation);
		break;
	}

	return -EOPNOTSUPP;
}

void snwnmp_get_roamtable(struct net_device *dev)
{
	snwnmp_get(dev, OID_NWN_ROAMTABLELENGTH);
	snwnmp_get(dev, OID_NWN_ROAMBSSID);
	snwnmp_get(dev, OID_NWN_ROAMSSID);
	snwnmp_get(dev, OID_NWN_ROAMBSSTYPE);
	snwnmp_get(dev, OID_NWN_ROAMCHANNEL);
	snwnmp_get(dev, OID_NWN_ROAMAGE);
	snwnmp_get(dev, OID_NWN_ROAMQUALITY);
	snwnmp_get(dev, OID_NWN_ROAMLOAD);
	snwnmp_get(dev, OID_NWN_ROAMBEACONPERIOD);
	snwnmp_get(dev, OID_NWN_ROAMDTIMPERIOD);
	snwnmp_get(dev, OID_NWN_ROAMCAPABILITYINFORMATION);
	snwnmp_get(dev, OID_NWN_ROAMRATES);
}

void snwnmp_get_counters(struct net_device *dev)
{
	snwnmp_get(dev, OID_DOT11_DOT11TRANSMITTEDFRAGMENTCOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11MULTICASTTRANSMITTEDFRAMECOUNT); 
	snwnmp_get(dev, OID_DOT11_DOT11FAILEDCOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11RETRYCOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11MULTIPLERETRYCOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11FRAMEDUPLICATECOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11RTSSUCCESSCOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11RTSFAILURECOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11ACKFAILURECOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11RECEIVEDFRAGMENTCOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11MULTICASTRECEIVEDFRAMECOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11FCSERRORCOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11TRANSMITTEDFRAMECOUNT);
	snwnmp_get(dev, OID_DOT11_DOT11WEPUNDECRYPTABLECOUNT);
}
