/*
 * Connection tracking state replication for Netfilter
 *
 * (C) 2002-2004, KOVACS Krisztian <hidden@sch.bme.hu>
 * (C) 2003-2004, Harald Welte <laforge@netfilter.org>
 * (C) 2004, BalaBit IT Ltd. <www.balabit.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/prefetch.h>
#include <linux/proc_fs.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/skbuff.h>
#include <linux/notifier.h>
#include <linux/sysctl.h>
#include <linux/seq_file.h>
#include <linux/percpu.h>
#include <linux/kthread.h>
#include <net/sock.h>

#include <linux/netdevice.h>
#include <linux/netlink.h>
#include <linux/netfilter_packet.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>
#include <linux/netfilter_ipv4/ip_conntrack_core.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_core.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>

#include <asm/uaccess.h>
#include <linux/unistd.h>

#include <linux/netfilter_ipv4/ct_sync.h>
#include <linux/netfilter_ipv4/ct_sync_proto.h>

#define DEBUGP printk

#if (CTSYNC_DEBUG > 0)
#define CT_SYNC_DUMP_TUPLE(x) DUMP_TUPLE(x)
#else
#define CT_SYNC_DUMP_TUPLE(x)
#endif

#define ASSERT_READ_LOCK(x)
#define ASSERT_WRITE_LOCK(x)

#include <linux/netfilter_ipv4/listhelp.h>

#define CT_SYNC_VERSION	"0.20"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("KOVACS Krisztian <hidden@sch.bme.hu>, Harald Welte <laforge@netfilter.org>");
MODULE_DESCRIPTION("Connection tracking state synchronization module");

/* module parameters */
static char *syncdev = "";
module_param(syncdev, charp, 0000);
static int l2drop = 0;
module_param(l2drop, int, 0000);
static int notrack = 1;
module_param(notrack, int, 0000);

#ifdef CONFIG_IP_NF_CONNTRACK_SYNC_MARKED
static int cmarkbit = sizeof(unsigned long) * 8 - 1;
module_param(cmarkbit, int, 0000);
#endif

/* thread wait queue heads */
static DECLARE_WAIT_QUEUE_HEAD(ct_sync_rcv_wait);

static DECLARE_WAIT_QUEUE_HEAD(ct_sync_send_wait);

static atomic_t ct_sync_initsync_active = ATOMIC_INIT(0);
static DECLARE_WAIT_QUEUE_HEAD(ct_sync_initsync_wait);

DEFINE_PER_CPU(struct ct_sync_stat, ct_sync_stats);

/* configuration data */
static struct ct_sync_cfg
{
	struct sockaddr_in addr;	/* multicast address */
	char devname[IFNAMSIZ];		/* dedicated sync network device */
	int ifindex;			/* ifindex of the network device */
	struct cts_protoh *protoh;	/* protocol handle */
} cts_cfg;

/***********************************************************************
 * FILLING CTSYNC MESSAGES WITH DATA
 ***********************************************************************/

#if 0
static int
fill_expectmsg(void *buff, __u8 event,
	       struct ip_conntrack *master,
	       struct ip_conntrack_expect *expect)
{
	struct ct_sync_msghdr *hdr = buff;
	struct ct_sync_expect *sexp = buff + sizeof(*hdr);

	CT_SYNC_ENTER();

	/* fill event header */
	hdr->type = event;
	hdr->resource = CT_SYNC_RES_EXPECT;
	hdr->len = __constant_htons(sizeof(*sexp));

	/* copy data from expectation */
	memcpy(&sexp->tuple, &expect->tuple, sizeof(sexp->tuple));
	memcpy(&sexp->mask, &expect->mask, sizeof(sexp->mask));
	sexp->seq = expect->seq;
	memcpy(&sexp->proto, &expect->proto, sizeof(sexp->proto));
	sexp->help = expect->help;
	if (expect->expectant)
		memcpy(&sexp->expectant,
			&master->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
			sizeof(sexp->expectant));
	else
		memset(&sexp->expectant, 0, sizeof(sexp->expectant));

	if (expect->sibling)
		memcpy(&sexp->sibling,
			&expect->sibling->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
			sizeof(sexp->sibling));
	else
		memset(&sexp->sibling, 0, sizeof(sexp->sibling));

	CT_SYNC_LEAVE();

	return 0;
}
#endif

static int
fill_ctmsg(void *buff, __u8 event, struct ip_conntrack *ct, __u8 flags)
{
	struct ct_sync_msghdr *hdr = buff;
	struct ct_sync_conntrack *sct = buff + sizeof(*hdr);

#if  defined(CONFIG_IP_NF_TARGET_MASQUERADE) \
  || defined(CONFIG_IP_NF_TARGET_MASQUERADE_MODULE)
	struct net_device *masq_dev;
#endif

	CT_SYNC_ENTER();

	/* fill event header */
	hdr->type = event;
	hdr->resource = CT_SYNC_RES_CONNTRACK;
	hdr->len = __constant_htons(sizeof(*sct));
	hdr->flags = flags;

	//CT_SYNC_DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);
	//CT_SYNC_DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_REPLY].tuple);

	/* copy data from conntrack struct */
	memset(sct, 0, sizeof(*sct));
	memcpy(&sct->orig, &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple, 
		sizeof(sct->orig));
	memcpy(&sct->reply, &ct->tuplehash[IP_CT_DIR_REPLY].tuple,
		sizeof(sct->reply));
	sct->status = (u32) ct->status;
	sct->expires = ct->timeout.expires - jiffies;
	if (unlikely(ct->helper != NULL)) {
		strncpy(sct->helper, ct->helper->name, sizeof(sct->helper));
		memcpy(&sct->help, &ct->help, sizeof(sct->help));
	} else
		sct->helper[0] = '\0';

	sct->expecting = ct->expecting;

	if (unlikely(ct->master != NULL)) {
		struct ip_conntrack *master_ct = master_ct(ct);
		CT_SYNC_ASSERT(master_ct);
		if (master_ct)
			memcpy(&sct->master_ct_tuple,
				&master_ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
				sizeof(sct->master_ct_tuple));
		else
			memset(&sct->master_ct_tuple, 0, sizeof(sct->master_ct_tuple));
	}

	memcpy(&sct->proto, &ct->proto, sizeof(sct->proto));

#ifdef CONFIG_IP_NF_NAT_NEEDED
	if (likely(ct->status & IPS_NAT_DONE_MASK)) {
		const struct ip_nat_info *nat = &ct->nat.info;

		memcpy(&sct->nat_seq, &nat->seq, sizeof(sct->nat_seq));
#if  defined(CONFIG_IP_NF_TARGET_MASQUERADE) \
  || defined(CONFIG_IP_NF_TARGET_MASQUERADE_MODULE)
		if (ct->nat.masq_index) {
			masq_dev = dev_get_by_index(ct->nat.masq_index);
			if (masq_dev) {
				memcpy(sct->nat_masq_iface, masq_dev->name,
				       sizeof(masq_dev->name));
				dev_put(masq_dev);
			} else {
				CT_SYNC_ERR("unable to find device for "
					    "masq_index %u\n",
					    ct->nat.masq_index);
				sct->nat_masq_iface[0] = '\0';
			}
		}
#endif
	} else
		sct->nat_initialized = 0;
#endif /* CONFIG_IP_NF_NAT_NEEDED */

#ifdef CONFIG_IP_NF_CONNTRACK_MARK
	sct->mark = (u32) ct->mark;
#endif

	CT_SYNC_LEAVE();

	return 0;
}

/***********************************************************************
 * BACKEND FOR CONNTRACK MODIFICATION
 ***********************************************************************/

/*
 * update data in the conntrack entry
 */
static inline int
_ct_sync_update_conntrack(struct ip_conntrack *ct,
			  struct ct_sync_conntrack *sct, int new)
{
	CT_SYNC_ENTER();

	if (likely(!new)) {
		/* check if the conntrack entry already has an active timer:
		 * it is possible that we try to update an active entry, which
		 * we must avoid */
		if (timer_pending(&ct->timeout)) {
			CT_SYNC_DEBUG("Trying to update an active conntrack entry!\n");
			CT_SYNC_LEAVE();
			return -1;
		}

		CT_SYNC_ASSERT(memcmp(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
				      &sct->orig, sizeof(sct->orig)) == 0);
		CT_SYNC_ASSERT(memcmp(&ct->tuplehash[IP_CT_DIR_REPLY].tuple,
				      &sct->reply, sizeof(sct->reply)) == 0);

		/* we should proceed with the update. only parts of an existing
		 * (confirmed) entry can be updated:
		 *   - status
		 *   - timeout (the timer is _inactive_)
		 *   - proto
		 *   - help (if a conntrack helper is present)
		 *   - mark
		 *   - nat.help (if there is a nat helper)
		 */
		ct->status = sct->status;
		ct->timeout.expires = sct->expires;
		memcpy(&ct->proto, &sct->proto, sizeof(ct->proto));
#ifdef CONFIG_IP_NF_CONNTRACK_MARK
		ct->mark = (unsigned long) sct->mark;
#endif
	} else {
#ifdef CONFIG_IP_NF_NAT_NEEDED
		struct ip_nat_info *nat = &ct->nat.info;
#endif

		/* this is a new entry, fill in everything */
		ct->status = sct->status;
		ct->timeout.expires = sct->expires;
		ct->expecting = 0;
#ifdef CONFIG_IP_NF_CONNTRACK_MARK
		ct->mark = (unsigned long) sct->mark;
#endif
		memcpy(&ct->proto, &sct->proto, sizeof(ct->proto));

		memcpy(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple, &sct->orig,
		       sizeof(struct ip_conntrack_tuple));
		memcpy(&ct->tuplehash[IP_CT_DIR_REPLY].tuple, &sct->reply,
		       sizeof(struct ip_conntrack_tuple));

		/* conntrack helper */
		if (unlikely(sct->helper[0] != '\0')) {
			struct ip_conntrack_helper *helper;

			sct->helper[CT_SYNC_CTHELPERSIZE - 1] = '\0';
			read_lock_bh(&ip_conntrack_lock);
			helper = __ip_conntrack_helper_find_byname(sct->helper);
			if (unlikely(!helper)) {
				CT_SYNC_ERR("Unknown conntrack helper `%s', "
					    "ignoring.\n", sct->helper);
				ct->helper = NULL;
			} else {
				ct->helper = helper;
			}
			read_unlock_bh(&ip_conntrack_lock);
		}

#ifdef CONFIG_IP_NF_NAT_NEEDED
		/* NAT */
		INIT_LIST_HEAD(&nat->bysource);

		if (likely(sct->status & IPS_NAT_DONE_MASK)) {
#if  defined(CONFIG_IP_NF_TARGET_MASQUERADE) \
  || defined(CONFIG_IP_NF_TARGET_MASQUERADE_MODULE)
			struct net_device *masq_dev;
#endif

#if  defined(CONFIG_IP_NF_TARGET_MASQUERADE) \
  || defined(CONFIG_IP_NF_TARGET_MASQUERADE_MODULE)
			if (sct->nat_masq_iface[0] != '\0') {
				masq_dev = dev_get_by_name(sct->nat_masq_iface);
				if (masq_dev) {
					ct->nat.masq_index = masq_dev->ifindex;
					dev_put(masq_dev);
				} else {
					CT_SYNC_ERR("no ifindex for dev %s\n",
						    sct->nat_masq_iface);
					ct->nat.masq_index = 0;
				}
			} else {
				/* no masquerading for this connection */
				ct->nat.masq_index = 0;
			}
#endif
		}
#endif /* CONFIG_IP_NF_NAT_NEEDED */

		/* add to hash tables */
		write_lock_bh(&ip_conntrack_lock);
		if (!__ip_conntrack_find(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple, NULL) 
		    && !__ip_conntrack_find(&ct->tuplehash[IP_CT_DIR_REPLY].tuple, NULL)) {
			/* put in conntrack hash */
			ip_conntrack_hash_insert_nolock(ct);
			atomic_inc(&ct->ct_general.use);

#ifdef CONFIG_IP_NF_NAT_NEEDED
			/* put in NAT hashes if necessary */
			if (ct->status & IPS_NAT_DONE_MASK) {
				ip_nat_hash_insert(ct);
			}
#endif

		} else {
			CT_SYNC_ERR("want to put conntrack in hash but is already there\n");
		}
		write_unlock_bh(&ip_conntrack_lock);
	} /* if (new) */

	CT_SYNC_LEAVE();

	return 0;
}

/* delete conntrack entry, but do not generate an event */
static inline int
_ct_sync_remove_conntrack(struct ip_conntrack *ct)
{
	CT_SYNC_ENTER();

	del_timer(&ct->timeout);
	write_lock_bh(&ip_conntrack_lock);
	ip_conntrack_clean_from_lists(ct);
	write_unlock_bh(&ip_conntrack_lock);
	ip_conntrack_put(ct);

	CT_SYNC_LEAVE();

	return 0;
}

static inline int
_start_ct_timer(const struct ip_conntrack_tuple_hash *h,
		unsigned int *num, unsigned int *started)
{
	struct ip_conntrack *ct = tuplehash_to_ctrack(h);

	if (DIRECTION(h))
		return 0;

	(*num)++;
	if (likely(is_confirmed(ct))) {
		if (!timer_pending(&ct->timeout)) {
			/* add five extra seconds to make sure that the entry
			 * won't timeout instantaneously */
			ct->timeout.expires += jiffies + 5 * HZ;
			add_timer(&ct->timeout);
			(*started)++;
		}
	}

	return 0;
}

static void
ct_sync_start_timers(void)
{
	int i;
	unsigned int num_started = 0;
	unsigned int num_entries = 0;

	CT_SYNC_ENTER();

	write_lock_bh(&ip_conntrack_lock);

	for (i = 0; i < ip_conntrack_htable_size; i++) {
		if (LIST_FIND(&ip_conntrack_hash[i], _start_ct_timer,
			      struct ip_conntrack_tuple_hash *, 
			      &num_entries, &num_started))
			break;
	}

	write_unlock_bh(&ip_conntrack_lock);

	CT_SYNC_INFO("started timer of %u (total %u) conntrack entries\n", 
		     num_started, num_entries);

	CT_SYNC_LEAVE();
}

static inline int
_stop_ct_timer(const struct ip_conntrack_tuple_hash *h, 
		unsigned int *num, unsigned int *stopped)
{
	struct ip_conntrack *ct = tuplehash_to_ctrack(h);

	if (DIRECTION(h))
		return 0;

	(*num)++;
	if (is_confirmed(ct)
#ifdef CONFIG_IP_NF_CONNTRACK_SYNC_MARKED
	    /* stop timer only if this is a synchronized connection */
	    && test_bit(cmarkbit, &ct->mark)
#endif
	) {
		if (del_timer(&ct->timeout))
			(*stopped)++;
	}

	return 0;
}

static void
ct_sync_stop_timers(void)
{
	int i;
	unsigned int num_entries = 0;
	unsigned int num_stopped = 0;

	CT_SYNC_ENTER();

	write_lock_bh(&ip_conntrack_lock);

	for (i = 0; i < ip_conntrack_htable_size; i++) {
		if (LIST_FIND(&ip_conntrack_hash[i], _stop_ct_timer,
			      struct ip_conntrack_tuple_hash *,
			      &num_entries, &num_stopped))
			break;
	}

	write_unlock_bh(&ip_conntrack_lock);

	CT_SYNC_INFO("stopped timer of %u (total %u) conntrack entries\n", 
			num_stopped, num_entries);

	CT_SYNC_LEAVE();
}


/***********************************************************************
 * MESSAGE PROCESSING FUNCTIONS
 ***********************************************************************/

/* process messages */
static int
ct_sync_msg_process_update_ct(struct ct_sync_msghdr *msghdr, 
			       void *data, u16 len)
{
	struct ct_sync_conntrack *sct = (struct ct_sync_conntrack *)data;
	struct ip_conntrack_tuple_hash *h;
	struct ip_conntrack *ct;
	struct ip_conntrack_tuple dummy_tuple;
	int new = 0;

	CT_SYNC_ENTER();

	if (unlikely(len < sizeof(struct ct_sync_conntrack))) {
		if (len != 0) {
			CTS_STAT_INC(rx.error);
			CT_SYNC_ERR("Illegal message size: %u\n", len);
		}
		CT_SYNC_LEAVE();
		return -1;
	}

	h = ip_conntrack_find_get(&sct->orig, NULL);
	if (h) {
		ct = tuplehash_to_ctrack(h);
	} else {
		CT_SYNC_DEBUG("Conntrack entry not found, creating.\n");
		ct = ip_conntrack_alloc(&dummy_tuple, &dummy_tuple);
		if (unlikely(IS_ERR(ct))) {
			CT_SYNC_DEBUG("Unable to create conntrack entry, ignoring message.\n");
			CT_SYNC_LEAVE();
			return -1;
		}
		new = 1;
	}

	if (unlikely(h && msghdr->flags & CTS_UPD_F_NEW)) {
		CTS_STAT_INC(rx.upd_newbutexist);
		CT_SYNC_DEBUG("received NEW conntrack, but we already know it: ");
		CT_SYNC_DUMP_TUPLE(&sct->orig);
	} else if (unlikely(!h && !(msghdr->flags & CTS_UPD_F_NEW))) {
		CTS_STAT_INC(rx.upd_nothere);
		CT_SYNC_DEBUG("received UPD conntrack, but we don't know it: ");
		CT_SYNC_DUMP_TUPLE(&sct->orig);
	}

	_ct_sync_update_conntrack(ct, sct, new);

	CT_SYNC_DEBUG("%s conntrack: ", new ? "new" : "upd");
	CT_SYNC_DUMP_TUPLE(&ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple);

	ip_conntrack_put(ct);

	CT_SYNC_LEAVE();

	return 0;
}

static int
ct_sync_msg_process_del_ct(struct ct_sync_msghdr *msghdr, 
			   void *data, u16 len)
{
	struct ip_conntrack_tuple *t = (struct ip_conntrack_tuple * ) data;
	struct ip_conntrack_tuple_hash *h;
		
	CT_SYNC_ENTER();

	if (unlikely(len < sizeof(*t))) {
		CTS_STAT_INC(rx.error);
		CT_SYNC_ERR("Illegal message size: %u\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}

	h = ip_conntrack_find_get(t, NULL);
	if (h) {
		struct ip_conntrack *ct = tuplehash_to_ctrack(h);

		_ct_sync_remove_conntrack(ct);
		CT_SYNC_DEBUG("Deleting conntrack: ");
		CT_SYNC_DUMP_TUPLE(t);
		ip_conntrack_put(ct);
	} else {
		CTS_STAT_INC(rx.del_nothere);
		CT_SYNC_DEBUG("Cannot delete nonexistent conntrack:");
		CT_SYNC_DUMP_TUPLE(t);
	}

	CT_SYNC_LEAVE();

	return 0;
}

static int
ct_sync_msg_process_updateexpect(void *data, u16 len)
{
	struct ct_sync_expect *exp = (struct ct_sync_expect *)data;

	CT_SYNC_ENTER();

	if (unlikely(len < sizeof(struct ct_sync_expect))) {
		CTS_STAT_INC(rx.error);
		CT_SYNC_ERR("Illegal message size: %u\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}

	CT_SYNC_LEAVE();

	return 0;
}

static int
ct_sync_msg_process_delexpect(void *data, u16 len)
{
	struct ct_sync_expect *sexp = (struct ct_sync_expect *)data;
	struct ip_conntrack_expect *exp;

	CT_SYNC_ENTER();

	if (unlikely(len < sizeof(struct ct_sync_expect))) {
		CTS_STAT_INC(rx.error);
		CT_SYNC_ERR("Illegal message size: %u\n", len);
		CT_SYNC_LEAVE();
		return -1;
	}
#if 0
	READ_LOCK(&ip_conntrack_lock);
	WRITE_LOCK(&ip_conntrack_expect_tuple_lock);
	exp = LIST_FIND(&ip_conntrack_expect_list, expect_cmp,
			struct ip_conntrack_expect *, exp->tuple);
	if (!exp || !exp->expectant) 
		goto unlock_out;

	if (conntrack_tuple_cmp(&exp->expectant,
			&exp->expectant->tuplehash[IP_CT_DIR_ORIGINAL].tuple))
		__unexpect_related(exp);

unlock_out:
	WRITE_INLOCK(&ip_conntrack_expect_tuple_lock);
	READ_UNLOCK(&ip_donntrack_lock);
#endif
	CT_SYNC_LEAVE();
	return 0;
}

/* process incoming message in slave mode */
static int
ct_sync_msg_process_slave(struct ct_sync_msghdr *msghdr, 
			  struct ct_sync_pkthdr *pkthdr)
{
	int ret = -1;
	char *data = (char *)msghdr + sizeof(*msghdr);
	int len = ntohs(msghdr->len);

	CT_SYNC_ENTER();

	if (pkthdr->pkttype != CT_SYNC_PKT_SYNC) {
		CT_SYNC_DEBUG("Invalid packet while being SLAVE: %d\n",
				pkthdr->pkttype);
		CT_SYNC_LEAVE();
		return -1;
	}

	switch (msghdr->resource) {
	case CT_SYNC_RES_CONNTRACK:
		switch (msghdr->type) {
		case CT_SYNC_MSG_UPDATE:
			CTS_STAT_INC(rx.upd_tot);
			ret = ct_sync_msg_process_update_ct(msghdr, data, len);
			break;
		case CT_SYNC_MSG_DELETE:
			CTS_STAT_INC(rx.del_tot);
			ret = ct_sync_msg_process_del_ct(msghdr, data, len);
			break;
		}
		break;
	case CT_SYNC_RES_EXPECT:
		switch(msghdr->type) {
		case CT_SYNC_MSG_UPDATE:
			CTS_STAT_INC(rx.exp_upd_tot);
			ret = ct_sync_msg_process_updateexpect(data, len);
			break;
		case CT_SYNC_MSG_DELETE:
			CTS_STAT_INC(rx.exp_del_tot);
			ret = ct_sync_msg_process_delexpect(data, len);
			break;
		}
		break;
	default:
		CT_SYNC_DEBUG("Invalid resource while being SLAVE: %d.\n",
				msghdr->resource); 
	}

	CT_SYNC_LEAVE();

	return ret;
}

/* process incoming message in master mode */
static int
ct_sync_msg_process_master(struct ct_sync_msghdr *msghdr,
			   struct ct_sync_pkthdr *pkthdr)
{
	int ret;
	u16 pktseq = ntohs(pkthdr->pktseq);

	CT_SYNC_ENTER();

	CT_SYNC_DEBUG("seq: %u\n", pktseq);

	switch (pkthdr->pkttype) {
		case CT_SYNC_PKT_SYNC:
			CT_SYNC_DEBUG("received SYNC while master->ignoring\n");
			ret = 0;
			break;
		default:
			CT_SYNC_DEBUG("Invalid packet while being MASTER:%d\n",
					pkthdr->pkttype);
			ret = -1;
			break;
	}

	CT_SYNC_LEAVE();

	return ret;
}

/* start initial synchronization */
static int
ct_sync_start_initsync(struct cts_protoh *cph)
{
	CT_SYNC_ENTER();

	if (cts_proto_is_master(cts_cfg.protoh) &&
	    !atomic_read(&ct_sync_initsync_active)) {
		CTS_STAT_INC(initsyncs);
		wake_up(&ct_sync_initsync_wait);
	} else
		CT_SYNC_INFO("not master, or initsync already in progress\n");

	CT_SYNC_LEAVE();

	return 0;
}

/***********************************************************************
 * SYSCTL HANDLING
 ***********************************************************************/

/* sysctl variables */
static int cts_sysctl_state;
static int cts_sysctl_maxage;
static int cts_sysctl_send_burst = 100;
static int cts_sysctl_recv_burst = 20;

static int recovery_threshold_min = 1;
static int recovery_threshold_max = 32000;

#define CTS_SYSCTL_STATE_NONE	0
#define CTS_SYSCTL_STATE_SLAVE	1
#define CTS_SYSCTL_STATE_MASTER	2

static int
cts_sysctl_change_state(int state)
{
	CT_SYNC_ENTER();

	switch (state) {
	case CTS_SYSCTL_STATE_NONE:
		/* cannot set state to NONE */
		CT_SYNC_ERR("cannot switch to state none!\n");
		CT_SYNC_LEAVE();
		return -EINVAL;
		break;

	case CTS_SYSCTL_STATE_SLAVE:
		if (cts_proto_is_master(cts_cfg.protoh)) {
			CTS_STAT_INC(master2slave);
			CT_SYNC_ERR("cannot switch active master to client!\n");
			CT_SYNC_LEAVE();
			return -EINVAL;
		} else if (cts_proto_get_state(cts_cfg.protoh) == CT_SYNC_PSTATE_NONE) {
			cts_proto_become_slave(cts_cfg.protoh);
		}
		break;

	case CTS_SYSCTL_STATE_MASTER:
		if (!cts_proto_is_master(cts_cfg.protoh)) {
			if (cts_proto_become_master(cts_cfg.protoh) < 0) {
				CT_SYNC_ERR("proto said we cannot become master\n");
				CT_SYNC_LEAVE();
				return -EINVAL;
			}
		} else {
			CT_SYNC_INFO("we already are master; staying cool\n");
		}
		break;

	default:
		/* unknown state */
		CT_SYNC_ERR("cannot switch to unknown state!\n");
		CT_SYNC_LEAVE();
		return -EINVAL;
		break;
		break;
	}

	CT_SYNC_LEAVE();

	return 0;
}

static void
cts_sysctl_update_state(void)
{
	CT_SYNC_ENTER();

	switch (cts_proto_get_state(cts_cfg.protoh)) {
	case CT_SYNC_PSTATE_NONE:
		cts_sysctl_state = CTS_SYSCTL_STATE_NONE;
		break;
	case CT_SYNC_PSTATE_SLAVE_INIT:
	case CT_SYNC_PSTATE_SLAVE_SYNSENT:
	case CT_SYNC_PSTATE_SLAVE_SYNRECV:
	case CT_SYNC_PSTATE_SLAVE_RUNNING:
		cts_sysctl_state = CTS_SYSCTL_STATE_SLAVE;
		break;
	case CT_SYNC_PSTATE_MASTER_INIT:
	case CT_SYNC_PSTATE_MASTER_ADVSENT:
	case CT_SYNC_PSTATE_MASTER_RUNNING:
		cts_sysctl_state = CTS_SYSCTL_STATE_MASTER;
		break;
	default:
		CT_SYNC_ERR("invalid state\n");
		break;
	}

	CT_SYNC_LEAVE();
}

static int
cts_sysctl_state_handler(ctl_table *table, int write,
			 struct file *filp, void *buffer,
			 size_t *lenp, loff_t *ppos)
{
	int res;
	
	if (write) {
		res = proc_dointvec(table, write, filp, buffer, lenp, ppos);
		if (res < 0)
			return res;
		return cts_sysctl_change_state(cts_sysctl_state);
	} else {
		cts_sysctl_update_state();
		return proc_dointvec(table, write, filp, buffer, lenp, ppos);
	}
}

static int
cts_sysctl_maxage_handler(ctl_table *table, int write,
			  struct file *filp, void *buffer,
			  size_t *lenp, loff_t *ppos)
{
	int res;

	if (write) {
		res = proc_dointvec(table, write, filp, buffer, lenp, ppos);
		if (res < 0)
			return res;
		return cts_proto_set_max_age(cts_sysctl_maxage);
	} else {
		cts_sysctl_maxage = cts_proto_get_max_age();
		return proc_dointvec(table, write, filp, buffer, lenp, ppos);
	}
}

/* FIXME: define this in sysctl.h */
#define NET_IPV4_NF_CONNTRACK_SYNC 20

enum {
	CTS_SYSCTL_STATE = 1,
	CTS_SYSCTL_MAXAGE = 2,
	CTS_SYSCTL_SEND_BURST = 3,
	CTS_SYSCTL_RECV_BURST = 4,
	CTS_SYSCTL_RECOVERY_THRESHOLD = 5,
};

static ctl_table ct_sync_sysctl_table[] = {
	{
		.ctl_name	= CTS_SYSCTL_STATE,
		.procname	= "state",
		.mode		= 0644,
		.data		= &cts_sysctl_state,
		.maxlen		= sizeof(int),
		.proc_handler	= &cts_sysctl_state_handler,
	},
	{
		.ctl_name	= CTS_SYSCTL_MAXAGE,
		.procname	= "maxage",
		.mode		= 0644,
		.data		= &cts_sysctl_maxage,
		.maxlen		= sizeof(int),
		.proc_handler	= &cts_sysctl_maxage_handler,
	},
	{
		.ctl_name	= CTS_SYSCTL_SEND_BURST,
		.procname	= "send_burst",
		.mode		= 0644,
		.data		= &cts_sysctl_send_burst,
		.maxlen		= sizeof(int),
		.proc_handler	= &proc_dointvec,
	},
	{
		.ctl_name	= CTS_SYSCTL_RECV_BURST,
		.procname	= "recv_burst",
		.mode		= 0644,
		.data		= &cts_sysctl_recv_burst,
		.maxlen		= sizeof(int),
		.proc_handler	= &proc_dointvec,
	},
	{
		.ctl_name	= CTS_SYSCTL_RECOVERY_THRESHOLD,
		.procname	= "recovery_threshold",
		.mode		= 0644,
		.data		= &cts_proto_recovery_threshold,
		.maxlen		= sizeof(int),
		.proc_handler	= &proc_dointvec_minmax,
		.extra1		= &recovery_threshold_min,
		.extra2		= &recovery_threshold_max,
	},
	{ .ctl_name = 0 }
};

static ctl_table ct_sync_cts_table[] = {
	{
		.ctl_name	= NET_IPV4_NF_CONNTRACK_SYNC,
		.procname	= "ct_sync",
		.mode		= 0555,
		.child		= ct_sync_sysctl_table,
	},
	{ .ctl_name = 0 }
};

static ctl_table ct_sync_netfilter_table[] = {
	{
		.ctl_name	= NET_IPV4_NETFILTER,
		.procname	= "netfilter",
		.mode		= 0555,
		.child		= ct_sync_cts_table,
	},
	{ .ctl_name = 0 }
};

static ctl_table ct_sync_net_table[] = {
	{
		.ctl_name	= NET_IPV4,
		.procname	= "ipv4",
		.mode		= 0555,
		.child		= ct_sync_netfilter_table,
	},
	{ .ctl_name = 0 }
};

static ctl_table ct_sync_root_table[] = {
	{
		.ctl_name	= CTL_NET,
		.procname	= "net",
		.mode		= 0555,
		.child		= ct_sync_net_table,
	},
	{ .ctl_name = 0 }
};

static struct ctl_table_header *ct_sync_sysctl_header;

/***********************************************************************
 * KERNEL THREADS
 ***********************************************************************/

/* sync kernel thread: sender */
static int
ct_sync_send_thread_main(void *data)
{
	unsigned long remaining;
	DECLARE_WAITQUEUE(wait, current);

	CT_SYNC_ENTER();

        for (;;) {
		int i = 0;

		/* check if we should stop */
		if (kthread_should_stop()) {
			CT_SYNC_DEBUG("send thread stopping\n");
			break;
		}

		/* if there are pending packets, dequeue and send
		 * all of them */
		while (cts_proto_send_pending(cts_cfg.protoh)) {
			cts_proto_send_dequeue(cts_cfg.protoh);

			if (++i >= cts_sysctl_send_burst)
				break;
		}

		/* send an initsync request if necessary */
		cts_proto_initsync_request(cts_cfg.protoh);

		__set_current_state(TASK_INTERRUPTIBLE);
		add_wait_queue(&ct_sync_send_wait, &wait);
		CT_SYNC_DEBUG2("falling asleep\n");
		remaining = schedule_timeout(HZ);
		remove_wait_queue(&ct_sync_send_wait, &wait);
		__set_current_state(TASK_RUNNING);

		if (remaining <= 0) {
			/* we had a timeout, try to advance buffer even if it's not full */
			CT_SYNC_DEBUG2("woken up because schedule_timeout() timed out\n");
			if (cts_proto_advance_timeout(cts_cfg.protoh) < 0)
				CT_SYNC_ERR("advancing current csb failed\n");
		} else
			CT_SYNC_DEBUG2("woken up after schedule_timeout() because we had an event\n");
        }

	CT_SYNC_LEAVE();
	
	return 0;
}

/* sync kernel thread: receiver */
static int
ct_sync_rcv_thread_main(void *data)
{
	DECLARE_WAITQUEUE(wait, current);
	
	CT_SYNC_ENTER();
	
	for (;;) {
		int pkts_enqueued = 0;
		int pkts_received = 0;
		int msgs_received = 0;
		struct ct_sync_msghdr *msghdr;
		struct ct_sync_pkthdr *pkthdr;

	  	if (kthread_should_stop())
			break;

		/* wait for a new packet to arrive */
		if (!cts_proto_recv_pending(cts_cfg.protoh)) {
			__set_current_state(TASK_INTERRUPTIBLE);
			add_wait_queue(&ct_sync_rcv_wait, &wait);

			for (;;) {
				__set_current_state(TASK_INTERRUPTIBLE);
				if (cts_proto_recv_pending(cts_cfg.protoh) ||
				    kthread_should_stop())
					break;

				CT_SYNC_DEBUG2("falling asleep\n");
				schedule();
			}

			remove_wait_queue(&ct_sync_rcv_wait, &wait);
			__set_current_state(TASK_RUNNING);
		}

		/* first process packets from the socket buffer into
		 * our protocol layer */
		while (cts_proto_recv_pending(cts_cfg.protoh)) {
			switch (cts_proto_recv_pkt(cts_cfg.protoh)) {
			case CTS_PROTO_RCV_ERROR:
				/* FIXME: unrecoverable error, should complain,
 				 * but should not exit! */
				goto error;
				break;
			case CTS_PROTO_RCV_SYNC:
				/* we'll handle this later when 
				 * dequeuing messages */
				pkts_enqueued++;
				break;
			case CTS_PROTO_RCV_INITSYNC:
				/* initiate initial synchronization */
				ct_sync_start_initsync(cts_cfg.protoh);
				break;
			default:
				/* the protocol layer handled other cases,
				   and dropped the packet */
				break;
			}

			pkts_received++;
			CTS_STAT_INC(rx.pkts_tot);
			if (pkts_received >= cts_sysctl_recv_burst)
				break;
		}

		if (pkts_enqueued > 0)
			/* process messages */
			while ((msghdr = cts_proto_recvmsg(cts_cfg.protoh, 
								&pkthdr))) {
				msgs_received++;
				CTS_STAT_INC(rx.msgs);

				if (cts_proto_is_master(cts_cfg.protoh))
					ct_sync_msg_process_master(msghdr, 
								pkthdr);
				else
					ct_sync_msg_process_slave(msghdr, 
								pkthdr);
			}

		/* give chance for other processes to run */
		if (pkts_received >= cts_sysctl_recv_burst)
			cond_resched();
	}

error:
	CT_SYNC_LEAVE();

	return 0;
}

static inline int
_send_initsync(const struct ip_conntrack_tuple_hash *h, unsigned int *num)
{
	struct ip_conntrack *ct = tuplehash_to_ctrack(h);

	if (DIRECTION(h))
		return 0;

	if (likely(is_confirmed(ct)
#ifdef CONFIG_IP_NF_CONNTRACK_SYNC_MARKED
	    /* send only if this is a synchronized connection */
	    && test_bit(cmarkbit, &ct->mark)
#endif
	    )) {
		char *buff;
		struct cts_buff *csb;
		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb, 
				CTMSG_SIZEOF(struct ct_sync_conntrack));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to store create event\n");
			return 1;
		}
		fill_ctmsg(buff, CT_SYNC_MSG_UPDATE, ct, CTS_UPD_F_INITSYNC);
		csb_use_dec(cts_cfg.protoh, csb);
		(*num)++;
	}

	return 0;
} 

static int
_send_empty_initsync(u8 flags)
{
	struct cts_buff *csb;
	struct ct_sync_msghdr *hdr;

	CT_SYNC_ENTER();

	hdr = (struct ct_sync_msghdr *) cts_proto_want_enqueue(cts_cfg.protoh, &csb, sizeof(struct ct_sync_msghdr));
	if (unlikely(!hdr)) {
		CT_SYNC_ERR("unable to store create event\n");
		return 1;
	}

	/* fill event header */
	hdr->type = CT_SYNC_MSG_UPDATE;
	hdr->resource = CT_SYNC_RES_CONNTRACK;
	hdr->len = 0;
	hdr->flags = flags;

	csb_use_dec(cts_cfg.protoh, csb);

	return 0;
}

/* sync kernel thread: recover */
static int
ct_sync_initsync_thread_main(void *data)
{
	DECLARE_WAITQUEUE(wait, current);

	CT_SYNC_ENTER();

	for (;;) {
		int i; 
		int num_sent_total;

		__set_current_state(TASK_INTERRUPTIBLE);
		add_wait_queue(&ct_sync_initsync_wait, &wait);
		schedule();
		remove_wait_queue(&ct_sync_initsync_wait, &wait);
		__set_current_state(TASK_RUNNING);

		/* somebody woke us.  either initsync or thread termination
		 * request */

		if (kthread_should_stop())
			break;

		CT_SYNC_INFO("starting initsync dump\n");
		atomic_set(&ct_sync_initsync_active, 1);

		/* Send an INITSYNC_FOLLOWS packet indicating that a full
		 * re-sync follows and all nodes waiting for re-sent
		 * packets should go to SLAVE_INIT state and forget
		 * current receiver state */
		cts_proto_initsync_follows(cts_cfg.protoh);

		/* speed: 40 conntracks per schedule, 100 schedules per
		 * second: 4000 conntracks per second.  This makes about
		 * 25 seconds for 100,000 connections.  Please note that
		 * we have four conntracks per packet, so the wire would
		 * see something like 1000 packets per second, or in
		 * other words: 1.5mbits */

		num_sent_total = 0;
		for (i = 0; i < ip_conntrack_htable_size; i++) {
			unsigned int num_sent = 0;

			if (kthread_should_stop())
				break;

			read_lock_bh(&ip_conntrack_lock);
dump_bucket_locked:
			if (LIST_FIND(&ip_conntrack_hash[i], 
				      _send_initsync,
			      	       struct ip_conntrack_tuple_hash *,
				       &num_sent)) {
				read_unlock_bh(&ip_conntrack_lock);
				break;
			}
			if (num_sent < 40 &&
		 	    i+1 < ip_conntrack_htable_size) {
				i++;
				goto dump_bucket_locked;
			}
			read_unlock_bh(&ip_conntrack_lock);

			num_sent_total += num_sent;

			CT_SYNC_DEBUG2("going to sleep, sent %u\n", num_sent);
			/* schedule until the next bucket */
			__set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ/100);
			__set_current_state(TASK_RUNNING);
			CT_SYNC_DEBUG2("woken up\n");
		}

		if (!num_sent_total) {
			/* need to send an empty PKT_SYCN/UPD_F_INITSYNC
			 * packet in order to provoke state transition
			 * in slave */
			_send_empty_initsync(CTS_UPD_F_INITSYNC);
		}

		/* signal that the initsync is over */
		_send_empty_initsync(CTS_UPD_F_INITSYNC_DONE);

		atomic_set(&ct_sync_initsync_active, 0);

		CT_SYNC_INFO("finished initsync dump\n");
	}

	CT_SYNC_LEAVE();

	return 0;
}

/***********************************************************************
 * NOTIFIER HANDLERS
 ***********************************************************************/

/* conntrack notify 'create' function */
static void 
ct_sync_create(struct ip_conntrack *ct, int new)
{
	struct cts_buff *csb;

	CT_SYNC_ENTER();

	CTS_STAT_INC(tx.upd_tot);
	if (new)
		CTS_STAT_INC(tx.upd_new);

	if (cts_proto_is_master(cts_cfg.protoh) 
		   && is_confirmed(ct)
#ifdef CONFIG_IP_NF_CONNTRACK_SYNC_MARKED
		   /* generate a message only if the conntrack is synchronized */
		   && test_bit(cmarkbit, &ct->mark)
#endif
	) {
		unsigned char *buff;
		/* This API sucks. you get a cts_buff as return parameter
		 * and have to decrement it after finishing to use it */
		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb, 
				CTMSG_SIZEOF(struct ct_sync_conntrack));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to store create event\n");
			CT_SYNC_LEAVE();
			return;
		}
		fill_ctmsg(buff, CT_SYNC_MSG_UPDATE, ct, new ? CTS_UPD_F_NEW : 0);
		csb_use_dec(cts_cfg.protoh, csb);
	}

	CT_SYNC_LEAVE();
}

/* conntrack notify 'destroy' function */
static void
ct_sync_destroy(struct ip_conntrack *ct)
{
	struct cts_buff *csb;

	CT_SYNC_ENTER();

	if (cts_proto_is_master(cts_cfg.protoh) 
		   && is_confirmed(ct)
#ifdef CONFIG_IP_NF_CONNTRACK_SYNC_MARKED
		   /* generate a message only if the conntrack is to be
		    * synchronized */
		   && test_bit(cmarkbit, &ct->mark)
#endif
	) {
		unsigned char *buff;
		struct ct_sync_msghdr *hdr;
		struct ip_conntrack_tuple *t;

		CTS_STAT_INC(tx.del_tot);

		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb,
				CTMSG_SIZEOF(*t));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to store destroy event\n");
			CT_SYNC_LEAVE();
			return;
		}
		hdr = (struct ct_sync_msghdr *) buff;
		t = (void *)hdr + sizeof(*hdr);

		/* fill event header */
		hdr->type = CT_SYNC_MSG_DELETE;
		hdr->resource = CT_SYNC_RES_CONNTRACK;
		hdr->len = __constant_htons(sizeof(*t));

		memcpy(t, &ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple, sizeof(*t));

		csb_use_dec(cts_cfg.protoh, csb);
	}

	CT_SYNC_LEAVE();
}

#if 0
/* conntrack expectation created notification */
static void
ct_sync_expect_create(struct ip_conntrack_expect *exp)
{
	struct cts_buff *csb;
	struct ip_conntrack *master_ct = exp->expectant;

	CT_SYNC_ENTER();

	if (likely(cts_proto_is_master(cts_cfg.protoh) &&
	           is_confirmed(master_ct))) {
		void *buff;

		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb,
				CTMSG_SIZEOF(struct ct_sync_expect));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to enqueue event\n");
			CT_SYNC_LEAVE();
			return;
		}
		fill_expectmsg(buff, CT_SYNC_MSG_UPDATE, master_ct, exp);
		csb_use_dec(cts_cfg.protoh, csb);
	}

	CT_SYNC_LEAVE();

	return;
}

/* conntrack expectation destroyed notification */
static void
ct_sync_expect_destroy(struct ip_conntrack_expect *exp)
{
	struct cts_buff *csb;
	struct ip_conntrack *master_ct = exp->expectant;

	CT_SYNC_ENTER();

	if (likely(cts_proto_is_master(cts_cfg.protoh) &&
	    	   is_confirmed(master_ct))) {
		void *buff;

		buff = cts_proto_want_enqueue(cts_cfg.protoh, &csb, 
				CTMSG_SIZEOF(struct ct_sync_expect));
		if (unlikely(!buff)) {
			CT_SYNC_ERR("unable to enqueue event\n");
			CT_SYNC_LEAVE();
			return;
		}
		// FIXME: implementation
		csb_use_dec(cts_cfg.protoh, csb);
	}

	CT_SYNC_LEAVE();

	return;
}
#endif

static int
ct_sync_notify(struct notifier_block *this,
	       unsigned long events, void *conntrack)
{
	struct ip_conntrack *ct = (struct ip_conntrack *) conntrack;

	/* ignore our fake conntrack entry */
	if (ct == &ip_conntrack_untracked)
		return NOTIFY_DONE;

	if (events & IPCT_DESTROY) {
		ct_sync_destroy((struct ip_conntrack *)ct);

	} else if (events & (IPCT_NEW | IPCT_RELATED)) {
		ct_sync_create((struct ip_conntrack *)ct, 1);

	} else if (events & (IPCT_STATUS |
			     IPCT_PROTOINFO |
			     IPCT_HELPER |
			     IPCT_HELPINFO |
			     IPCT_NATINFO)) {
		ct_sync_create((struct ip_conntrack *)ct, 0);
	}

	/* FIXME: make it possible to optionally synchronize volatile
	 * events */

	return NOTIFY_DONE;
}

static struct notifier_block ct_sync_notifier = {
	ct_sync_notify,
	NULL,
	0
};

/***********************************************************************
 * NOTRACK (conntrack exemptions)
 ***********************************************************************/

static unsigned int
notrack_hook(unsigned int hook,
	     struct sk_buff **pskb,
	     const struct net_device *indev,
	     const struct net_device *outdev,
	     int (*okfn)(struct sk_buff *))
{
	if ((*pskb)->nfct != NULL)
		return NF_ACCEPT;

	/* all traffic coming in or going out sync interface is not
	   to be tracked. Also, loopback traffic is ignored */

	if ((hook == NF_IP_PRE_ROUTING && indev && 
	     indev->ifindex == cts_cfg.ifindex) ||
	    (hook == NF_IP_LOCAL_OUT && outdev &&
	     outdev->ifindex == cts_cfg.ifindex) ||
	    (indev && indev == &loopback_dev) || 
	    (outdev && outdev == &loopback_dev)) {

		/* Attach fake conntrack entry */
		(*pskb)->nfct = &ip_conntrack_untracked.ct_general;
		(*pskb)->nfctinfo = IP_CT_NEW;
		nf_conntrack_get((*pskb)->nfct);
	}

	return NF_ACCEPT;
}

static unsigned int
l2drop_hook(unsigned int hook,
	    struct sk_buff **pskb,
	    const struct net_device *indev,
	    const struct net_device *outdev,
	    int (*okfn)(struct sk_buff *))
{
	/* Allow local traffic */
	if ((indev && indev == &loopback_dev) ||
	    (outdev && outdev == &loopback_dev))
		return NF_ACCEPT;

	/* in slave mode, all traffic _NOT_ on sync interface has to
	   be dropped */
	if (cts_proto_is_master(cts_cfg.protoh))
		return NF_ACCEPT;

	if ((hook == NF_PACKET_OUTPUT && outdev &&
	     outdev->ifindex != cts_cfg.ifindex) ||
	    (hook == NF_PACKET_INPUT && indev &&
	     indev->ifindex != cts_cfg.ifindex)) {
		return NF_DROP;
	}

	return NF_ACCEPT;
}

static struct nf_hook_ops cts_hook_ops[] = {
	{ .hook = notrack_hook,
	  .pf = PF_INET,
	  .hooknum = NF_IP_PRE_ROUTING,
	  .priority = NF_IP_PRI_RAW-1 },
	{ .hook = notrack_hook,
	  .pf = PF_INET,
	  .hooknum = NF_IP_LOCAL_OUT,
	  .priority = NF_IP_PRI_RAW-1 },
	{ .hook = l2drop_hook,
	  .pf = PF_PACKET,
	  .hooknum = NF_PACKET_INPUT,
	  .priority = 0 },
	{ .hook = l2drop_hook,
	  .pf = PF_PACKET,
	  .hooknum = NF_PACKET_OUTPUT,
	  .priority = 0 },
};

/***********************************************************************
 * CT_SYNC_PROTO CALLBACK
 ***********************************************************************/

/* This could be converted to notifiers ... */
static void
ct_sync_state_transition(struct cts_protoh *cph,
			 enum cts_proto_state oldstate,
			 enum cts_proto_state newstate,
			 void *data)
{
	CT_SYNC_ENTER();

	switch (newstate) {
	case CT_SYNC_PSTATE_NONE:
		break;

	case CT_SYNC_PSTATE_SLAVE_INIT:
	case CT_SYNC_PSTATE_SLAVE_SYNRECV:
	case CT_SYNC_PSTATE_SLAVE_SYNSENT:
	case CT_SYNC_PSTATE_SLAVE_RUNNING:
		if (oldstate == CT_SYNC_PSTATE_NONE) {
			/* initial slave synchronization: we don't stop
			 * conntrack timers when being initialized to
			 * slave: this could cause connections left in
			 * the conntrack hash for an indefinite amount
			 * of time... */
		} else if (oldstate >= CT_SYNC_PSTATE_MASTER_INIT) {
			/* transition from master to slave */
			ip_conntrack_unregister_notifier(&ct_sync_notifier);
			ct_sync_stop_timers();
		}
		break;

	case CT_SYNC_PSTATE_MASTER_INIT:
	case CT_SYNC_PSTATE_MASTER_ADVSENT:
	case CT_SYNC_PSTATE_MASTER_RUNNING:
		if (oldstate < CT_SYNC_PSTATE_MASTER_INIT) {
			/* transition from none/slave to master */
			ct_sync_start_timers();
			if (ip_conntrack_register_notifier(&ct_sync_notifier) < 0) {
				CT_SYNC_ERR("cannot register notifier\n");
			}
		}
		break;

	default:
		break;
	}

	CT_SYNC_LEAVE();
}

/***********************************************************************
 * STATISTICS
 ***********************************************************************/

static void *
cts_stat_seq_start(struct seq_file *seq, loff_t *pos)
{
	int cpu;

	if (*pos == 0)
		return SEQ_START_TOKEN;

	for (cpu = *pos-1; cpu < NR_CPUS; ++cpu) {
		if (!cpu_possible(cpu))
			continue;
		*pos = cpu+1;
		return &per_cpu(ct_sync_stats, cpu);
	}
	return NULL;
}

static void *
cts_stat_seq_next(struct seq_file *seq, void *v, loff_t *pos)
{
	int cpu;
	
	for (cpu = *pos; cpu < NR_CPUS; ++cpu) {
		if (!cpu_possible(cpu))
			continue;
		*pos = cpu+1;
		return &per_cpu(ct_sync_stats, cpu);
	}
	return NULL;
}

static void
cts_stat_seq_stop(struct seq_file *seq, void *v)
{
}

static void
cts_stat_seq_show_st(struct seq_file *seq, struct ct_sync_st *s)
{
	seq_printf(seq, "%lu %lu %lu %lu %lu %lu %lu "
			"%lu %lu %lu %lu %lu %lu %lu  ",
		   s->pkts_tot, s->pkts_ooseq, s->pkts_recover,
		   s->msgs, s->error,
		   s->upd_tot, s->upd_new, s->upd_initial,
		   s->upd_newbutexist, s->upd_nothere,
		   s->del_tot, s->del_nothere,
		   s->exp_upd_tot, s->exp_del_tot);
}

static int
cts_stat_seq_show(struct seq_file *seq, void *v)
{
	struct ct_sync_stat *stats = v;

	if (v == SEQ_START_TOKEN) {
		seq_printf(seq, "tx_pkts_tot tx_pkts_ooseq tx_pkts_recover "
			   "tx_msgs tx_error tx_upd_tot tx_upd_new "
			   "tx_upd_initial tx_upd_newbutexist tx_upd_nothere " 
			   "tx_del_tot tx_del_nothere tx_exp_upd_tot "
			   "tx_exp_del_tot  ");
		seq_printf(seq, "rx_pkts_tot rx_pkts_ooseq rx_pkts_recover "
			   "rx_msgs rx_error rx_upd_tot rx_upd_new "
			   "rx_upd_initial rx_upd_newbutexist rx_upd_nothere "
			   "rx_del_tot rx_del_nothere rx_exp_upd_tot "
			   "rx_exp_del_tot  ");
		seq_printf(seq, "slave2master master2slave master2slave_auto "
			   "initsyncs\n");
		return 0;
	}

	cts_stat_seq_show_st(seq, &stats->tx);
	cts_stat_seq_show_st(seq, &stats->rx);
	seq_printf(seq, "%lu %lu %lu %lu\n", 
		   stats->slave2master, stats->master2slave,
		   stats->master2slave_auto, stats->initsyncs);
	return 0;
}

static struct seq_operations cts_stat_seq_ops = {
	.start  = cts_stat_seq_start,
	.next   = cts_stat_seq_next,
	.stop   = cts_stat_seq_stop,
	.show   = cts_stat_seq_show,
};

static int cts_stat_seq_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &cts_stat_seq_ops);
}

static struct file_operations cts_stat_seq_fops = {
       .owner   = THIS_MODULE,
       .open    = cts_stat_seq_open,
       .read    = seq_read,
       .llseek  = seq_lseek,
       .release = seq_release,
};


/***********************************************************************
 * MODULE INITIALIZATION
 ***********************************************************************/

static int
kill_all(const struct ip_conntrack *i, void *data)
{
	return 1;
}

static struct task_struct *rcv_thread, *send_thread, *initsync_thread;

/* DO NOT declare this as __init!! */
static int
init_or_cleanup(int fini)
{
	int err, ret = -1;
      	struct net_device *sync_netdev;
	struct proc_dir_entry *ct_sync_stat;

	if (fini)
		goto cleanup;

	/* informational messages */
	CT_SYNC_DEBUG("ip_ct_tuple: %d bytes\n", 
			sizeof(struct ip_conntrack_tuple));
	CT_SYNC_DEBUG("ct_sync_conntrack: %d bytes\n", 
			sizeof(struct ct_sync_conntrack));
	CT_SYNC_DEBUG("ct_sync_expect: %d bytes\n", 
			sizeof(struct ct_sync_expect));
	CT_SYNC_DEBUG("ct_sync_pkthdr: %d bytes\n", 
			sizeof(struct ct_sync_pkthdr));
	CT_SYNC_DEBUG("ct_sync_msghdr: %d bytes\n", 
			sizeof(struct ct_sync_msghdr));
#ifdef CONFIG_IP_NF_CONNTRACK_SYNC_MARKED
	CT_SYNC_DEBUG("cmarkbit: %d\n", cmarkbit);
#endif

	if (strcmp(syncdev, "") == 0) {
		CT_SYNC_ERR("Required parameter syncdev is missing.\n");
		return -1;
	}

	strcpy(cts_cfg.devname, syncdev);
	cts_cfg.addr.sin_family = AF_INET;
	cts_cfg.addr.sin_port = __constant_htons(1999);
	cts_cfg.addr.sin_addr.s_addr = 
			__constant_htonl(0xe0000052); /* 224.0.0.82 */

	/* Get ifindex of the sync device, used by the notrack and l2drop hook */
	if ((sync_netdev = dev_get_by_name(cts_cfg.devname)) != NULL) {
		cts_cfg.ifindex = sync_netdev->ifindex;
		dev_put(sync_netdev);
	} else {
		CT_SYNC_ERR("Lookup for network interface '%s' failed\n", 
			    cts_cfg.devname);
		return -1;
	}

	/* Register hooks first, make sure not even the first sync packet gets
	 * tracked */
	if (notrack) {
		if (nf_register_hook(&cts_hook_ops[0]) < 0)
			goto error_none;
		if (nf_register_hook(&cts_hook_ops[1]) < 0)
			goto error_hook0;
	} else {
		printk(KERN_NOTICE "WARNING: ct_sync internal notrack disabled\n");
	}

	if (l2drop) {
		if (nf_register_hook(&cts_hook_ops[2]) < 0)
			goto error_hook1;
		if (nf_register_hook(&cts_hook_ops[3]) < 0)
			goto error_hook2;
		ip_conntrack_cleanup();
	}

	/* init protocol layer */
	cts_cfg.protoh = cts_proto_init(cts_cfg.devname, &cts_cfg.addr,
					&ct_sync_state_transition, NULL,
					&ct_sync_send_wait, &ct_sync_rcv_wait);
	if (!cts_cfg.protoh) {
		CT_SYNC_ERR("Failed to initialize protocol.\n");
		ret = -EINVAL;
		goto error_hook3;
	}

	/* register sysctl entries */
	ct_sync_sysctl_header = register_sysctl_table(ct_sync_root_table, 0);
	if (ct_sync_sysctl_header == NULL)
		goto error_proto;

	/* register statistics */
	ct_sync_stat = create_proc_entry("ct_sync", 0, proc_net_stat);
	CT_SYNC_DEBUG("proc_net_stat=%p, ct_sync_stat=%p\n", proc_net_stat, ct_sync_stat);
	if (!ct_sync_stat)
		goto error_sysctl;
	ct_sync_stat->proc_fops = &cts_stat_seq_fops;

	/* start threads */
	rcv_thread = kthread_run(ct_sync_rcv_thread_main, NULL,
				 "ct_sync_rcv");
	if (IS_ERR(rcv_thread)) {
		ret = PTR_ERR(rcv_thread);
		goto error_stat;
	}

	send_thread = kthread_run(ct_sync_send_thread_main, NULL,
				  "ct_sync_send");
	if (IS_ERR(send_thread)) {
		ret = PTR_ERR(send_thread);
		goto error_rthread;
	}

	initsync_thread = kthread_run(ct_sync_initsync_thread_main, NULL,
				      "ct_sync_initsync");
	if (IS_ERR(initsync_thread)) {
		ret = PTR_ERR(initsync_thread);
		goto error_sthread;
	}

	/* Now we have all of our infrastructure up and running */
	printk(KERN_NOTICE "netfilter conntrack_sync version %s loaded\n",
	       CT_SYNC_VERSION);
	CT_SYNC_INFO("parameters: syncdev='%s' l2drop=%u notrack=%u"
#ifdef CONFIG_IP_NF_CONNTRACK_SYNC_MARKED
		     " cmarkbit=%u"
#endif
		     "\n", syncdev, l2drop, notrack
#ifdef CONFIG_IP_NF_CONNTRACK_SYNC_MARKED
		     , cmarkbit
#endif
		     );

        return 0;

cleanup:
	CT_SYNC_DEBUG("stopping initsync thread\n");
	err = kthread_stop(initsync_thread);
	if (err < 0) {
		CT_SYNC_INFO("unable to stop initsync thread\n");
	}
error_sthread:
	CT_SYNC_DEBUG("stopping send thread\n");
	err = kthread_stop(send_thread);
	if (err < 0) {
		CT_SYNC_INFO("unable to stop send thread\n");
	}
error_rthread:
	CT_SYNC_DEBUG("stopping receive thread\n");
	err = kthread_stop(rcv_thread);
	if (err < 0) {
		CT_SYNC_INFO("unable to stop receive thread\n");
	}

	CT_SYNC_DEBUG("unregistering conntrack notifier\n");
	ip_conntrack_unregister_notifier(&ct_sync_notifier);
error_stat:
	remove_proc_entry("ct_sync_stat", proc_net_stat);
error_sysctl:
	CT_SYNC_DEBUG("removing sysctl entry\n");
	unregister_sysctl_table(ct_sync_sysctl_header);
error_proto:
	cts_proto_cleanup(cts_cfg.protoh);
error_hook3:
	if (l2drop)
		nf_unregister_hook(&cts_hook_ops[3]);
error_hook2:
	if (l2drop)
		nf_unregister_hook(&cts_hook_ops[2]);
error_hook1:
	if (notrack)
		nf_unregister_hook(&cts_hook_ops[1]);
error_hook0:
	if (notrack)
		nf_unregister_hook(&cts_hook_ops[0]);
	/* make sure all conntrack timers are running */
	ct_sync_start_timers();
error_none:
	CT_SYNC_DEBUG("returning %d\n", ret);
	return ret;
}

int __init init(void)
{
	return init_or_cleanup(0);
}

/* remove the module */
void __exit fini(void)
{
	init_or_cleanup(1);
	printk(KERN_NOTICE "netfilter conntrack_sync version %s unloaded\n",
		CT_SYNC_VERSION);
}

module_init(init);
module_exit(fini);
