/*
 * Connection tracking state replication for Netfilter
 *
 * (C) 2002-2004, KOVACS Krisztian <hidden@sch.bme.hu>
 * (C) 2003-2004, Harald Welte <laforge@netfilter.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Based on previous work from the LVS project
 * (C) Wensong Zwang <wensong@linuxvirtualserver.org>
 *
 */

#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/version.h>

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/wait.h>

#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/inetdevice.h>
#include <linux/igmp.h>
#include <linux/ip.h>

#include <asm/uaccess.h>

#include <linux/netfilter_ipv4/ct_sync.h>

#include <net/inet_sock.h>

/*
 *      Setup loopback of outgoing multicasts on a sending socket
 */
static void
set_mcast_loop(struct sock *sk, u_char loop)
{
        /* setsockopt(sock, SOL_IP, IP_MULTICAST_LOOP, &loop, sizeof(loop)); */
        lock_sock(sk);
        inet_sk(sk)->mc_loop = loop ? 1 : 0;
        release_sock(sk);
}

/*
 *      Specify TTL for outgoing multicasts on a sending socket
 */
static void 
set_mcast_ttl(struct sock *sk, u_char ttl)
{
        /* setsockopt(sock, SOL_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)); */
        lock_sock(sk);
        inet_sk(sk)->mc_ttl = ttl;
        release_sock(sk);
}


static int 
set_mcast_if(struct sock *sk, char *ifname)
{
        struct net_device *dev;

        if ((dev = __dev_get_by_name(ifname)) == NULL)
                return -ENODEV;

        if (sk->sk_bound_dev_if && dev->ifindex != sk->sk_bound_dev_if)
                return -EINVAL;

        lock_sock(sk);
        inet_sk(sk)->mc_index = dev->ifindex;
        /*  inet_sk(sk)->mc_addr  = 0; */
        release_sock(sk);

        return 0;
}

static int 
bind_mcastif_addr(struct socket *sock, char *ifname)
{
        struct net_device *dev;
        u32 addr;
        struct sockaddr_in sin;

        if ((dev = __dev_get_by_name(ifname)) == NULL)
                return -ENODEV;

        addr = inet_select_addr(dev, 0, RT_SCOPE_UNIVERSE);
        if (!addr)
                CT_SYNC_ERR("You probably need to specify IP address on "
                            "multicast interface.\n");

        CT_SYNC_DEBUG("binding socket with (%s) %u.%u.%u.%u\n",
                  ifname, NIPQUAD(addr));

        /* Now bind the socket with the address of multicast interface */
        sin.sin_family       = AF_INET;
        sin.sin_addr.s_addr  = addr;
        sin.sin_port         = 0;

        return sock->ops->bind(sock, (struct sockaddr*)&sin, sizeof(sin));
}
	
/*
 *      Join a multicast group.
 *      the group is specified by a class D multicast address 224.0.0.0/8
 *      in the in_addr structure passed in as a parameter.
 */
static int
join_mcast_group(struct sock *sk, struct in_addr *addr, char *ifname)
{
        struct ip_mreqn mreq;
        struct net_device *dev;
        int ret;

        memset(&mreq, 0, sizeof(mreq));
        memcpy(&mreq.imr_multiaddr, addr, sizeof(struct in_addr));

        if ((dev = __dev_get_by_name(ifname)) == NULL)
                return -ENODEV;
        if (sk->sk_bound_dev_if && dev->ifindex != sk->sk_bound_dev_if)
                return -EINVAL;

        mreq.imr_ifindex = dev->ifindex;

        lock_sock(sk);

        ret = ip_mc_join_group(sk, &mreq);
        release_sock(sk);

        return ret;
}

struct socket *
cts_sock_server_init(char *devname, const struct sockaddr_in *addr)
{
	struct socket *sock;

	if (sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &sock) < 0) {
		CT_SYNC_ERR("Socket creation failed.\n");
		return NULL;
	}

	if (set_mcast_if(sock->sk, devname) < 0) {
		CT_SYNC_ERR("Error setting outbound mcast interface.\n");
		goto error;
	}

	set_mcast_loop(sock->sk, 0);
	set_mcast_ttl(sock->sk, 1);

	if (bind_mcastif_addr(sock, devname) < 0) {
		CT_SYNC_ERR("Error binding to address of mcast interface.\n");
		goto error;
	}

	if (sock->ops->connect(sock, (struct sockaddr *)addr, sizeof(struct sockaddr), 0) < 0) {
		CT_SYNC_ERR("Error connecting to the multicast address.\n");
		goto error;
	}

	return sock;

  error:
	sock_release(sock);
	return NULL;
}

struct socket *
cts_sock_client_init(char *devname, const struct sockaddr_in *addr)
{
	struct socket *sock;

	if (sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &sock) < 0) {
		CT_SYNC_ERR("Error during creation of client socket.\n");
		return NULL;
	}

	sock->sk->sk_reuse = 1;

	if (sock->ops->bind(sock, (struct sockaddr *)addr, sizeof(struct sockaddr)) < 0) {
		CT_SYNC_ERR("Error binding to the multicast addr.\n");
		goto error;
	}

	if (join_mcast_group(sock->sk, (struct in_addr *)&addr->sin_addr, devname) < 0) {
		CT_SYNC_ERR("Error joining to the multicast group.\n");
		goto error;
	}

	return sock;

  error:
	sock_release(sock);
	return NULL;
}

ssize_t
cts_sock_sendmsg(struct socket *sock, struct msghdr *msg, const size_t length)
{
	ssize_t		len;
	mm_segment_t	oldfs;

	CT_SYNC_ENTER();

	CTS_STAT_INC(tx.pkts_tot);

	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_sendmsg(sock, msg, length);
	set_fs(oldfs);

	CT_SYNC_LEAVE();

	return len;
}


ssize_t
cts_sock_send(struct socket *sock, const char *buffer, const size_t length)
{
	struct msghdr	msg;
	mm_segment_t	oldfs;
	struct iovec	iov;
	ssize_t		len;

	CT_SYNC_ENTER();

	iov.iov_base	= (void *)buffer;
	iov.iov_len	= length;
	msg.msg_name	= 0;
	msg.msg_namelen	= 0;
	msg.msg_iov	= &iov;
	msg.msg_iovlen	= 1;
	msg.msg_control	= NULL;
	msg.msg_controllen = 0;
	msg.msg_flags	= MSG_DONTWAIT | MSG_NOSIGNAL;

	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_sendmsg(sock, &msg, length);
	set_fs(oldfs);

	CT_SYNC_LEAVE();

	return len;
}

ssize_t
cts_sock_recv(struct socket *sock, char *buffer, const size_t buflen)
{
	struct msghdr	msg;
	struct iovec	iov;
	ssize_t		len;
	mm_segment_t	oldfs;

	CT_SYNC_ENTER();

	CTS_STAT_INC(rx.pkts_tot);

	iov.iov_base	= buffer;
	iov.iov_len	= buflen;
	msg.msg_name	= 0;
	msg.msg_namelen = 0;
	msg.msg_iov	= &iov;
	msg.msg_iovlen	= 1;
	msg.msg_control	= NULL;
	msg.msg_controllen = 0;
	msg.msg_flags	= 0;

	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_recvmsg(sock, &msg, buflen, 0);
	set_fs(oldfs);

	CT_SYNC_LEAVE();

	return len;
}

